#ifndef __WALKASPPS_H__
#define __WALKASPPS_H__

#include "constants.h"
#include "array.cpp"
#include "hash.h"

// 0: no disjoint c-atoms or no c-atoms at all
// 1: applicable for some c-atoms
// 2: applicable for all c-atoms
extern int glDoubleFlipApplicable; 
extern double glPower;
extern int glCombBase;

int getInt(const char * line, int & pos);
int myRand(int d);
int getToken(const char * line, char * buffer, int size, int & pos);
int getAtomText(const char * line, char * buffer, int size, int & pos);

using namespace std;

class CRule;
class CCAtom;
class CWeightAtom;
class CRuleList;

/////////////////////////////////////////
// Class CAtom Definition
/////////////////////////////////////////
class CAtom
{
 public:
	typedef enum 
	{
		C_ATOM, 	// c-atom
		N_ATOM, 	// normal atom
		W_ATOM, 	// weight atom
		ATOM_TYPE_NUM 	// the number of atom types
	} TAtomType;

//	static const int m_iconstD = 100;
	static const int m_iconstC = BIGC;
	
 private:
	/* the number of c-atoms that have this atom and are satisfied.
	 * deprecated.
	 */
	int m_iSATCAtomNum;

	/* the list of c-atoms and weight atoms that this atom 
	 * appears in.
	 */
	CArray<CCAtom *> m_pCAtoms;
	int m_iCAtomLength;

	/* the list of c-atom rules that this atom appears in.
	 */
	CArray<CRule *> m_pCAtomRules;
	int m_iCAtomRuleLength;

	/* a list of unit rules that this atom appears in.
	 * We assume that the only atom in this rule is a c-atom
	 * and this atom appears in that c-atom.
	 */
	CArray<CCAtom *> m_pUnitCAtoms;
	int m_iUnitCAtomLength;

	/* a list of unit rules that this atom appears in.
	 * We assume that the only atom in this rule is a w-atom
	 * and this atom appears in that w-atom.
	 */
	CArray<CWeightAtom *> m_pUnitWAtoms;
	int m_iUnitWAtomLength;

	/* a list of rules that were chosen to apply generalized
	 * double flip.
	 */
	CArray<CRule *> m_pGDBFRules;
	
 protected:
	/* type of this atom
	 */
	TAtomType m_Type;

	/* a bad name. The integer representation of this atom.
	 * We assume that the representation starts with 1 and
	 * there are no holes in between.
	 */
	int m_iIndex;

	/* the truth value of this atom
	 */
	bool m_bValue;

	/* the text representation given by the grounder
	 */
	char m_sText[MAX_NAME_LENGTH];

	/* the break-count of this atom. It is defined as the number
	 * of propositional clauses that will become unsat if this
	 * atom is flipped.
	 */
	int m_iBreakCount;

	/* the make-count of this atom. not in use currently
	 */
	int m_iMakeCount;

	/* the timestamp of the last flip action on this atom
	 */
	unsigned int m_uiTimeStamp;
	
	/* the list of rules that this atom appears in.
	 */
	CArray<CRule *> m_pRules;
	int m_iRuleLength;

	/* whether the atom appears in a unit propositional rule, in which
	   case, it becomes a forced atom
	*/
	char m_cForced;

	void empty();
	
 public:
	CAtom(int catomLength = DEFAULT_SIZE,
	      int ruleLength = DEFAULT_SIZE);
	CAtom(const CAtom & anAtom);
	virtual ~CAtom();
	void shrinkArrays();

	TAtomType getType() const { return m_Type; }
	int getIndex() const { return m_iIndex; }
	int getRealBKCT() const;
	int getRealMKCT() const;
	int getCNFBKCT() const;
	int getCNFMKCT() const;
	int getBKCTCorrect(bool VBC = true) const;
	int getMKCTCorrect(bool VMC = true) const;
	int getBKCTWrong(bool VBC = true) const;
	int getMKCTWrong(bool VMC = true) const;
	bool getValue() const { return m_bValue; }
	const char * getText() const { return m_sText; }
	const char * getTextVal() const
	{
		if ( m_bValue )
			return "T";
		else
			return "F";
	}
	int getUnitCAtomNum() const { return m_iUnitCAtomLength; }
	int getUnitWAtomNum() const { return m_iUnitWAtomLength; }
	CCAtom * getUnitCAtom(int which) const
	{
		if ( which >= m_iUnitCAtomLength ||
		     which < 0 )
			return NULL;
		else
			return m_pUnitCAtoms[which];
	}
	CWeightAtom * getUnitWAtom(int which) const
	{
		if ( which >= m_iUnitWAtomLength ||
		     which < 0 )
			return NULL;
		else
			return m_pUnitWAtoms[which];
	}
	int getCAtomNum() const { return m_iCAtomLength; }
	CCAtom * getCAtom(int which) const
	{
		if ( which >= m_iCAtomLength ||
		     which < 0 )
			return NULL;
		else
			return m_pCAtoms[which];
	}
	CArray<CRule *> getGDBFRules() const { return m_pGDBFRules; }
	void setIndex(int id) { m_iIndex = id; }
	void setValue(bool value) { m_bValue = value; }
	void setText(const char * text) { strcpy(m_sText, text); }
	void resetTimeStamp() { m_uiTimeStamp = 0; };
	int getTimeStamp() const { return m_uiTimeStamp; };
	int addRule(CRule & rule);
	int addCAtomRule(CRule & rule);
	int addCAtom(CCAtom & catom);
	int addUnitCAtom(CCAtom & catom);
	int addUnitWAtom(CWeightAtom & watom);
	int addGDBFRule(CRule & rule);
	bool hasTheRule(CRule & rule) const
		{ return m_pRules.hasIt(&rule); }
	bool isCriticalInCAtom() const;
	bool isCriticalInCAtom(int which) const;
	bool isCriticalInWAtom() const;
	bool isCriticalInWAtom(int which) const;
	int reeval(CRuleList * pRuleList = NULL);
//	int increaseBKCT(const CArray<CAtom *> & pBody,
//			 const CArray<CAtom *> & pHead)
	int increaseBKCT()
	{
		if ( m_iBreakCount < m_iRuleLength )
			m_iBreakCount++;
		return m_iBreakCount;
	}
//	int decreaseBKCT(const CArray<CAtom *> & pBody,
//			 const CArray<CAtom *> & pHead)
	int decreaseBKCT()
	{
		if ( m_iBreakCount > 0 )
			m_iBreakCount--;
		return m_iBreakCount;
	}
//	int increaseMKCT(const CArray<CAtom *> & pBody,
//			 const CArray<CAtom *> & pHead)
	int increaseMKCT()
	{
		if ( m_iMakeCount < m_iRuleLength )
			m_iMakeCount++;
		return m_iMakeCount;
	}
//	int decreaseMKCT(const CArray<CAtom *> & pBody,
//			 const CArray<CAtom *> & pHead)
	int decreaseMKCT()
	{
		if ( m_iMakeCount > 0 )
			m_iMakeCount--;
		return m_iMakeCount;
	}
	int forced() const { return m_cForced; };
	void setForced(bool body) { body ? m_cForced = -1 : m_cForced = 1;};
	
	const CAtom & operator = (const CAtom & anAtom);
	const CAtom & operator = (const bool val)
	{
		m_bValue = val;
		return *this;
	}
	
	virtual void dump() const;
	virtual void dumpText() const;
	virtual void dumpTextWithBKCT() const;
	virtual bool getRealValue() const { return m_bValue; }
	virtual int getBKCT(bool VBC = true) const;
	virtual int getMKCT(bool VMC = true) const;
	virtual int negateSelf(CRuleList * pRuleList = NULL,
			       CAtom * flipped = NULL);
	virtual bool SAT() const { return m_bValue; }
};

/////////////////////////////////////////
// Class CCAtom Definition
/////////////////////////////////////////
/* For each catom, it appears in exactly one rule.
 * Thus, even if two c-atoms in two rules have the same
 * set of atoms and the same bounds, they are treated as
 * two different c-atom objects here.
 */
class CCAtom : public CAtom
{
 protected:
       /*
	* atomValue    in    vfunction     UV pos    XY pos
	*         0     0            0    UBAROUT   XBAROUT
	*         0     0            1    VBAROUT   YBAROUT
	*         0     1            0     UBARIN    XBARIN
	*         0     1            1     VBARIN    YBARIN
	*         1     0            0       UOUT      XOUT
	*         1     0            1       VOUT      YOUT
	*         1     1            0        UIN       XIN
	*         1     1            1        VIN       YIN
	*/
	typedef enum 
	{
		UBAROUT, VBAROUT, UBARIN, VBARIN, UOUT, VOUT, UIN, VIN,
		UVPOSNUM
	} TUVType;
	typedef enum 
	{
		XBAROUT, YBAROUT, XBARIN, YBARIN, XOUT, YOUT, XIN, YIN,
		XYPOSNUM
	} TXYType;

 protected:
	/* Indicates if this catom is a max/min catom.
	   0: not a max/min catom
	   1: a max catom
	   2: a min catom
	 */
	int m_iMaxMin;

	int m_iUpperBound;
	int m_iLowerBound;

	/* the number of atoms that currently have truth value TRUE
	 * in this c-atom.
	 */
	int m_iNumOfSatisfied;

	/* TRUE if this c-atom appears positively in a rule.
	 */
	bool m_bPositive;

	/* TRUE if this c-atom appears in a unit rule.
	 */
	bool m_bInUnitRule;

	/* The current break-count that this c-atom contributes to
	 * all FALSE/TRUE atoms in this c-atom.
	 * This is used only when this c-atom is the only c-atom in
	 * the rule.
	 */
	int m_iBKSingleCAtomCaseFalse;
	int m_iBKSingleCAtomCaseTrue;

	/* The current make-count that this c-atom contributes to
	 * all FALSE/TRUE atoms in this c-atom.
	 * This is used only when this c-atom is the only c-atom in
	 * the rule.
	 */
	int m_iMKSingleCAtomCaseFalse;
	int m_iMKSingleCAtomCaseTrue;

	/* record the values of all cases of (U, V) / (X, Y) functions.
	 * Please refer to my doctoral thesis proposal for the
	 * definitions of all (U, V) / (X, Y) functions.
	 */
	int m_iaUVValue[UVPOSNUM];
	int m_iaXYValue[XYPOSNUM];

	/* the list of atoms that appear in this c-atom.
	 */
	CArray<CAtom *> m_pAtoms;
	int m_iAtomLength;

	void empty();
	void partition(CArray<int> & indices,
		       int low, int high, int & newLow, int & newHigh);
	void quickSort(CArray<int> & indices,
		       int low, int high);
	
 public:
	CCAtom(int atomSize = DEFAULT_ATOM_SIZE_IN_CATOM,
	       int ruleSize = 1);
	CCAtom(const CCAtom & aCAtom);
	virtual ~CCAtom();
	void shrinkArrays();

	void setMaxMin(int maxmin)
		{ if ( maxmin == 0 || maxmin == 1 || maxmin == 2 )
			m_iMaxMin = maxmin; }
	int isMaxMin() const { return m_iMaxMin; }
	void maxMinResetVal() ;
	
	int getUpperBound() const { return m_iUpperBound; }
	int getLowerBound() const { return m_iLowerBound; }
	int getNumOfSatisfied() const { return m_iNumOfSatisfied; }
	virtual int getCurrentWeight() const { return getNumOfSatisfied(); };
	int getAtomLength() const { return m_iAtomLength; }
	CAtom ** getAtoms() const { return m_pAtoms.toArray(); }
	CRule * getRule() const { return m_pRules[0]; }
	void setUpperBound(int ub) { m_iUpperBound = ub; }
	void setLowerBound(int ub) { m_iLowerBound = ub; }
	void setPositive() { m_bPositive = true; }
	void setNegative() { m_bPositive = false; }
	int addAtom(CAtom & anAtom,
		    int lb = NO_LBOUND, int ub = NO_UBOUND);
	bool isPositive() const { return m_bPositive; }
	void inUnitRule() { m_bInUnitRule = true; }
	bool isInUnitRule() const { return m_bInUnitRule; }
	void needUpdate();
	const CCAtom & operator = (const CCAtom & aCAtom);

	virtual void dump() const;
	virtual void dumpText() const;
	virtual void dumpTextWithBKCT() const;
	virtual int getBKSingleCAtomCase(bool atomValue,
					 const CAtom *theOne);
	virtual int getMKSingleCAtomCase(bool atomValue,
					 const CAtom *theOne);
	virtual int getUV(bool atomValue, bool in, bool vfunction,
			  const CAtom *theOne);
	virtual int getXY(bool atomValue, bool in, bool yfunction,
			  const CAtom *theOne);
	virtual int getBKCT(bool atomValue = true) const;
	virtual int getMKCT(bool atomValue = true) const;
	virtual bool getRealValue() const
	{
		int num = 0;
		
		for ( int i = 0; i < m_iAtomLength; i++ )
			if ( m_pAtoms[i]->getRealValue() )
				num++;
			
		return ( (num <= m_iUpperBound &&
			  num >= m_iLowerBound) );
	}
	virtual bool SAT() const { return ( (m_bValue && m_bPositive) ||
					    (!m_bValue && !m_bPositive) ); }	
	virtual bool canBreak(const CAtom * atom);
	virtual bool canMake(const CAtom * atom);
	virtual int negateSelf(CRuleList * pRuleList = NULL,
			       CAtom * flipped = NULL);
	virtual bool reeval(const CAtom & flippedAtom);
	virtual void printWeight() const {};
	virtual bool noHelp(const CAtom * atom) const;
};

/////////////////////////////////////////
// Class CWeightAtom Definition
/////////////////////////////////////////
class CWeightAtom : public CCAtom
{
 protected:
	
	/* the list of weights for each atom.
	 * To do: use hash table to make the searching faster
	 */
	CArray<int> m_iWeights;

	/* the hash table that stores indices of atoms
	 */
	CHash<int> m_ihIndex;
	
	/* the sum of the weights under current truth assignment.
	 */
	int m_iCurrentWeight;
	int m_iTotalWeight; // It now denotes the max weight, which
	                    // equals m_iTotalABSWeight - m_iTotalABSNegWeight
	int m_iTotalABSNegWeight;
	int m_iTotalABSWeight;
	
	//To do: make it faster
	int getPosition(const CAtom * atom) const
	{
/* first slow implementation
  		for ( int i = 0; i < m_pAtoms.getLength(); i++ )
			if ( m_pAtoms[i] == atom )
				return i;
		return NONE;
*/
/* hash table implementation
 */
		const int * pPos = m_ihIndex.constGet(atom->getIndex());
		if ( pPos == NULL )
			return NONE;
		else
			return *pPos;
	}

	// estimate of the number of unsat clauses that do not have
	// atom a
	int f2(bool atom, int weight) const
	{
		int currentWeight = m_iCurrentWeight;
		currentWeight += (!atom ? weight : -weight);

		if ( m_iCurrentWeight > m_iUpperBound &&
		     currentWeight > m_iUpperBound )
		{
			if ( currentWeight < m_iCurrentWeight )
				return F(atom, weight);
			else
				return F(m_iCurrentWeight);
		}
		
		if ( m_iCurrentWeight < m_iLowerBound &&
		     currentWeight < m_iLowerBound )
		{
			if ( currentWeight > m_iCurrentWeight )
				return F(atom, weight);
			else
				return F(m_iCurrentWeight);
		}
	
		return 0;
	}

	// estimate of the number of unsat clauses
	int F(int currentWeight) const
	{
		int value1 = (currentWeight - m_iUpperBound) *
			m_iconstC;
		int value2 = (m_iLowerBound - currentWeight) *
			m_iconstC;

		if ( value1 < 0 && value2 < 0 )
			return 0;
		else if ( value1 < value2 )
			return value2;
		else
			return value1;
	}
	// estimate of the number of unsat clauses after atom a
	// is flipped
	int F(bool atom, int weight) const
	{
		int currentWeight = m_iCurrentWeight;
		currentWeight += (!atom ? weight : -weight);

		return F(currentWeight);
	}

	int getFactor() const
	{
//		return (int)(m_iTotalWeight/m_iAtomLength);
//		return m_iAtomLength;
		return 1;
	}
	
 public:
	CWeightAtom(int atomSize = DEFAULT_ATOM_SIZE_IN_CATOM,
	       int ruleSize = 1);
	CWeightAtom(const CWeightAtom & aWeightAtom);

	int getWeight(int index) const
	{
		return m_iWeights[index];
	}
	int getWeight(const CAtom * atom) const
	{
		int pos = getPosition(atom);
		if ( pos == NONE ) return NONE;
		return getWeight(pos);
	}
	int getCurrentWeight(void) const { return m_iCurrentWeight; }
	int getTotalWeight(void) const { return m_iTotalWeight; }
	
	int addAtom(CAtom & atom, int weight, int lb, int ub);
	const CWeightAtom & operator = (const CWeightAtom & aWeightAtom);

	virtual void dump() const;
	virtual void dumpText() const;
	virtual void dumpTextWithBKCT() const;
	virtual int getBKSingleCAtomCase(bool atomValue,
					 const CAtom *theOne);
	virtual int getMKSingleCAtomCase(bool atomValue,
					 const CAtom *theOne);
	virtual int getUV(bool atomValue, bool in, bool vfunction,
			  const CAtom *theOne);
	virtual int getXY(bool atomValue, bool in, bool yfunction,
			  const CAtom *theOne);
	virtual int getBKCT(bool atomValue = true) const;
	virtual int getMKCT(bool atomValue = true) const;
	virtual bool getRealValue() const
	{
		int num = 0;
		
		for ( int i = 0; i < m_iAtomLength; i++ )
			if ( m_pAtoms[i]->getRealValue() )
				num += m_iWeights[i];
			
		return ( (num <= m_iUpperBound &&
			  num >= m_iLowerBound) );
	}
	virtual bool SAT() const { return ( (m_bValue && m_bPositive) ||
					    (!m_bValue && !m_bPositive) ); }	
	virtual bool canBreak(const CAtom * atom);
	virtual bool canMake(const CAtom * atom);
	virtual int multiFlip(CRuleList * pRuleList = NULL,
			      CAtom * flipped = NULL);
	virtual int negateSelf(CRuleList * pRuleList = NULL,
			       CAtom * flipped = NULL);
	virtual bool reeval(const CAtom & flippedAtom);
	virtual void printWeight() const
	{
		cout<<" [ ";
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( !m_pAtoms[i]->getValue() )
				continue;
			
			cout<<m_pAtoms[i]->getText()<<" ";
		}
		cout<<" ] = "<<m_iCurrentWeight<<endl;
	}
	virtual bool noHelp(const CAtom * atom) const;
};

/////////////////////////////////////////
// Class CRule Definition
/////////////////////////////////////////
class CRule
{
 protected:
	/* the original index of the rule in the program file.
	 */
	int m_iOrigIndex;

	/* the index of the rule in the rule list. This index will
	 * change from time to time, depending on wether the rule
	 * currently is satisfied or not.
	 */
	int m_iIndex;

	/* TRUE if the rule is currently satisfied.
	 */
	bool m_bSAT;

	/* the number of true literals in the rule.
	 * The whole c-atom is considered as one literal.
	 */
	int m_iTrueLiteralNum;

	/* the number of normal true literals in the rule.
	 * By normal, we exclude the c-atoms and weight atoms.
	 */
	int m_iNormalTrueLiteralNum;

	/* all of the following three variables are deprecated
	 */
	bool m_bIsCritical;
	bool m_bBodySAT;
	bool m_bHeadSAT;

	/* the list of all atoms in this rule.
	 * This list also includes all the atoms in the c-atoms or
	 * weight atoms appearing in this rule.
	 */
	CArray<CAtom *> m_pAllAtoms;
	int m_iAtomLength;

	/* the list of all atoms in the normal part of this rule.
	 * This list is defined only when there are c/w-atoms in the
	 * rule.
	 */
	CArray<CAtom *> m_pAllNormAtoms;
	int m_iNormAtomLength;
	
	/* index of the last atom that has the largest timestamp
	 */
	int m_iLastRecentlyFlipped;

	/* the list of all c-atoms and weight atoms in this rule.
	 */
	CArray<CCAtom *> m_pAllCAtoms;
	int m_iCAtomLength;

	/* the list of atoms in the body of this rule.
	 */
	CArray<CAtom *> m_pBody;
	int m_iBodyLength;
	int m_iBodyLastFalseIndex;

	/* the list of atoms in the head of this rule.
	 */
	CArray<CAtom *> m_pHead;
	int m_iHeadLength;
	int m_iHeadLastTrueIndex;

	/* Statistics - # of times that this rule is chosen as the
	   unsat rule.
	 */
	int m_iTouchedTimes;
	int m_iTotalTouchedTimes;

	/* TRUE if this rule is chosen for generalized double flip.
	 */
	bool m_bUsedInDBF;
	
	void empty();
	int addAtom(CAtom & atom);

	// make it faster
	int findAtom(const CAtom * theOne) const
	{
		int position = -1;
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( m_pAllAtoms[i] == theOne )
			{
				position = i;
				break;
			}
		}
		return position;
	};
	
	void updateMostRecentlyFlipped(CAtom * theOne)
	{
		int position = findAtom(theOne);

		if ( position == -1 || position == 0 ||
		     position > m_iAtomLength-1 )
			return;

		if ( theOne->getTimeStamp() >
		     m_pAllAtoms[m_iLastRecentlyFlipped]->getTimeStamp() )
		{
			m_pAllAtoms[position] = m_pAllAtoms[0];
			m_pAllAtoms[0] = theOne;
			m_iLastRecentlyFlipped = 0;
		}
		else if ( m_iLastRecentlyFlipped < m_iAtomLength - 1 &&
			  theOne->getTimeStamp() ==
			  m_pAllAtoms[m_iLastRecentlyFlipped]->getTimeStamp()
			  )
		{
			m_pAllAtoms[position] =
				m_pAllAtoms[m_iLastRecentlyFlipped+1];
			m_pAllAtoms[m_iLastRecentlyFlipped+1] = theOne;
			m_iLastRecentlyFlipped++;
		}
	};

 public:
	CRule(int bodySize = DEFAULT_SIZE,
	      int headSize = DEFAULT_SIZE);
	virtual ~CRule();

	void dump() const;
	void dumpText() const;
	void dumpTextWithBKCT() const;
	
	void resetLastRecentlyFlipped() { m_iLastRecentlyFlipped = 0; };
	bool isMostRecentlyFlipped(const CAtom * theOne) const
	{
		int position = findAtom(theOne);

		return isMostRecentlyFlipped(position);
	};
	bool isMostRecentlyFlipped(int index) const
	{
		return (index >= 0 && index <= m_iLastRecentlyFlipped );
	};
	int getLargestTimeStamp() const
		{ return m_pAllAtoms[0]->getTimeStamp(); };
	
	void resetTouched() { m_iTouchedTimes = 0; }
	void touched(bool startCounting )
	{
		if ( startCounting )
		{
			m_iTouchedTimes++;
			m_iTotalTouchedTimes++;
		}
	}
	int getTouchedTimes() const { return m_iTouchedTimes; }
	int getTotalTouchedTimes() const { return m_iTotalTouchedTimes; }

	void setUsedInDBF() { m_bUsedInDBF = true; }
	void unsetUsedInDBF() { m_bUsedInDBF = false; }
	bool isUsedInDBF() const { return m_bUsedInDBF; }
	
	void setOrigIndex(int id) {m_iOrigIndex = id;}
	int getOrigIndex() const {return m_iOrigIndex;}
	
	void setIndex(int id) {m_iIndex = id;}
	int getIndex() const {return m_iIndex;}
	
	int addBodyAtom(CAtom & atom);
	int addHeadAtom(CAtom & atom);
	void shrinkArrays();

	bool isInBody(const CAtom & atom) const
	{
		for (int i = 0; i < m_pBody.getLength(); i++ )
		{
			if ( m_pBody[i] == &atom )
				return true;
		}
		return false;
	}
	bool isInHead(const CAtom & atom) const
	{
		for (int i = 0; i < m_pHead.getLength(); i++ )
		{
			if ( m_pHead[i] == &atom )
				return true;
		}
		return false;
	}
	bool isInRule(CAtom & atom) const
		{ return (isInBody(atom) || isInHead(atom)); }

	bool hasCAtom() const { return m_iCAtomLength > 0; }
	int getCAtomLength() const { return m_iCAtomLength; }
	CCAtom ** getCAtoms() const { return m_pAllCAtoms.toArray(); }
	int getAtomLength() const { return m_iAtomLength; }
	CAtom ** getAtoms() const { return m_pAllAtoms.toArray(); }
	int getNormAtomLength() const { return m_iNormAtomLength; }
	CAtom ** getNormAtoms() const { return m_pAllNormAtoms.toArray(); }
	
	void increaseTrueLiteralNum() { m_iTrueLiteralNum++; }
	void decreaseTrueLiteralNum() { m_iTrueLiteralNum--; }

	bool init();
	bool reeval();
	bool reeval(CAtom * theFlippedAtom);
	void swap(int index, bool currValue, bool isInBody);

	bool hasNoNormalTrueLiterals() const
		{ return m_iNormalTrueLiteralNum <= 0; }
	
	bool SAT() const {return m_bSAT;}
	bool isCritical(const CAtom & atom) const
		{ return (m_iTrueLiteralNum == 1 &&
			  ((isInBody(atom) && !atom.getValue()) ||
			   (isInHead(atom) && atom.getValue()))); }

	bool satNormAfterFlipping(const CAtom & atom) const
		{
			bool inBody = isInBody(atom);
			bool inHead = isInHead(atom);
			bool value = atom.getValue();
			
			return (m_iNormalTrueLiteralNum > 1 ||
				(m_iNormalTrueLiteralNum == 1 &&
				 ((!inBody || value) &&
				  (!inHead || !value))) ||
				(m_iNormalTrueLiteralNum < 1 &&
				 (inBody || inHead)));
		}
	bool satNormBeforeFlipping(const CAtom & atom) const
		{ return (m_iNormalTrueLiteralNum > 0); }
		
	bool isUnitRule() const
	{
		if ( m_iBodyLength + m_iHeadLength == 1 &&
		     m_iCAtomLength == 1)
			return true;
		else
			return false;
	}
	
	void checkUnitRule();
	void checkCAtomRule()
	{
		if ( hasCAtom() )
		{
			m_pAllNormAtoms.newArray(m_iAtomLength);
			m_iNormAtomLength = 0;
			for ( int i = 0; i < m_iAtomLength; i++ )
			{
				m_pAllAtoms[i]->addCAtomRule(*this);
				// Fixed: Bug!!! in setting up normal part
				if ( m_pAllAtoms[i]->getType() ==
				     CAtom::N_ATOM &&
				     (isInBody(*m_pAllAtoms[i]) ||
				      isInHead(*m_pAllAtoms[i])) )
				{
					m_pAllNormAtoms.add(m_pAllAtoms[i]);
					m_iNormAtomLength++;
				}
			}
			if ( 1.0/m_iCAtomLength < glPower )
			{
				glPower = 1.0/m_iCAtomLength;
				glCombBase = (int)(pow(MAX_INT, glPower))+1;
			}
		}
	}
	bool checkTrivialRule();
	bool noHelp(const CAtom * atom) const;
	
	const CRule & operator = (const CRule & rule);
};

/////////////////////////////////////////
// Class CAssignment Definition
/////////////////////////////////////////
class CAssignment
{
 protected:
	/* the list of forced atoms. They are forced to be TRUE and 
	 * were found by the grounder.
	 */
	CArray<char *> m_pForcedAtoms;
	int m_iForcedAtomLength;

	/* the initial assignment, if -DT, --data <file> option is
	 * used. This initial assignment will be used for each try.
	 * if m_bInitValues[i] == -1, then that atom will be initialized
	 * randomly.
	 * -1: random; 0: false; 1: true
	 */
	CArray<short> m_bInitValues;
	
	/* the list of all atoms in the theory.
	 */
	CArray<CAtom *> m_pAtoms;
	int m_iAtomLength;

	/* an assignment that causes the fewest unsat clauses
	 */
	CArray<bool> m_baBestValues;

	/* the list of all c-atoms and weight atoms in the theory.
	 */
	CArray<CCAtom *> m_pCAtoms;
	int m_iCAtomLength;

	/* internal auxilary variables. Other atoms that need to be
	 * flipped in double flip or permutation flip are recorded
	 * here.
	 */
	CAtom * m_pAuxAtoms[HEU_ATOM_NUM][3];
	int m_iAuxAtomNum[HEU_ATOM_NUM];

	/* index to auxilary array m_pAuxAtoms so that we don't need
	 * to spend time in finding those atoms again in flip procedure.
	 */
	int m_iChosenAtomIndex;

	/* Statistics - the number of times that double flip or 
	 * permutation flip is performed.
	 */	
	int m_iDoubleFlipNum;
	int m_iPermutationFlipNum;

	/* the number of atoms that have truth value TRUE.
	 * deprecated.
	 */
	int m_iTrueAtomNum;
	
	int m_iMaxMin;
	int m_iOldLB;
	int m_iOldUB;
	int m_iMidVal;
	CCAtom * m_pMaxMinCAtom;

	void empty();

	int hasTheAtom(int id) const;
	int addAtom(CAtom & atom);
	int addCAtom(CCAtom & catom);
	CAtom * walk(const CRule & rule, const CAtom * theAtom = NULL);
	CAtom * heuristic(const CRule & rule, const CAtom * theAtom = NULL);
	CAtom * heuristicSKC_VBC(const CRule & rule, const CAtom * theAtom);
	CAtom * heuristicSKC_DBF(const CRule & rule, const CAtom * theAtom);
	CAtom * heuristicSKC_PF(const CRule & rule, const CAtom * theAtom);
	CAtom * heuristicRNoveltyPlus_VC(const CRule & rule,
					 const CAtom * theAtom);
	CAtom * heuristicRNoveltyPlus_DBF(const CRule & rule,
					  const CAtom * theAtom);
	CAtom * findOppositeAtom_SKC(const CAtom & atom, int whichCAtom);
	CAtom * findOppositeAtom_RNoveltyPlus(const CAtom & atom,
					      int whichCAtom);
	bool getTheFixed(const CAtom & mainAtom,
			 CAtom * & colAtom,
			 CAtom * & rowAtom,
			 CAtom * & diagAtom);

	CAtom * pickAnAtom(const CRuleList & ruleList);
	CAtom * pickAnAtomFromSameRule(const CAtom * theAtom,
				       const CRule & rule);
	bool flip(CAtom * theAtom, CRuleList & ruleList);
	bool doubleFlip(CAtom * theAtom, CRuleList & ruleList);
	bool multiFlip(CAtom * theAtom, CRuleList & ruleList);
	bool permutationFlip(CAtom * theAtom, CRuleList & ruleList);
	bool generalDoubleFlip(CAtom * theAtom, CRuleList & ruleList);
 public:
	CAssignment(int atomSize = DEFAULT_SIZE,
		    int catomSize = DEFAULT_SIZE);
	virtual ~CAssignment();
	bool newAtomArray(int size);
	void shrinkArrays();
	
	void dump() const;
	
	int isMaxMin() const { return m_iMaxMin; };
	int getMiddle() const
	{
		if ( m_iMaxMin == 1 ) // max case
		{
			return m_iOldLB + (m_iOldUB - m_iOldLB) / 3;
		}
		else if ( m_iMaxMin == 2 ) // min case
		{
			return m_iOldLB + 2*(m_iOldUB - m_iOldLB) / 3;
		}
		else
			return 0;
	}
	void tryMiddleAsLB() // only used in max case
	{
		if ( m_iMaxMin != 1 && m_iMaxMin != 2 )
			return;
		m_iOldLB = m_pMaxMinCAtom->getLowerBound();
		m_iOldUB = m_pMaxMinCAtom->getUpperBound();
		m_iMidVal = getMiddle();
		m_pMaxMinCAtom->setLowerBound(m_iMidVal);
	}
	void tryMiddleAsUB()
	{
		if ( m_iMaxMin != 1 && m_iMaxMin != 2 )
			return;
		m_iOldLB = m_pMaxMinCAtom->getLowerBound();
		m_iOldUB = m_pMaxMinCAtom->getUpperBound();
		m_iMidVal = getMiddle();
		m_pMaxMinCAtom->setUpperBound(m_iMidVal);
	}
	void restoreUB()
	{
		if ( m_iMaxMin != 1 && m_iMaxMin != 2 )
			return;
		m_pMaxMinCAtom->setUpperBound(m_iOldUB);
	}
	void restoreLB()
	{
		if ( m_iMaxMin != 1 && m_iMaxMin != 2 )
			return;
		m_pMaxMinCAtom->setLowerBound(m_iOldLB);
	}

	int getOptimalValue() const
	{
		if ( m_iMaxMin == 1 ) // max case
			return m_pMaxMinCAtom->getLowerBound()-1;
		else if ( m_iMaxMin == 2 ) // min case
			return m_pMaxMinCAtom->getUpperBound()+1;
		else
			return 0;
	}

	int getJumpValue() const // should be called before current improveBound()
	{
		if ( m_iMaxMin == 1 ) // max case
			return (m_pMaxMinCAtom->getCurrentWeight() -
				m_pMaxMinCAtom->getLowerBound());
		else if ( m_iMaxMin == 2 ) // min case
			return (m_pMaxMinCAtom->getUpperBound() -
				m_pMaxMinCAtom->getCurrentWeight());
		else
			return 0;
	}
	int getNextJumpValue() const // should be called after next tryMiddle..()
	{
		if ( m_iMaxMin == 1 ) // max case
			return (getMiddle() - m_iOldLB);
		else if ( m_iMaxMin == 2 ) // min case
			return (m_iOldUB - getMiddle());
		else
			return 0;
	}
	
	// should be called when current max/min w/c-atom is satisfiable
	// return true iff the optimal bound is reached
	bool improveBound(CRuleList & ruleList)
	{
		int currW = m_pMaxMinCAtom->getCurrentWeight();
		if ( m_iMaxMin == 1 ) // max case
		{
			if ( currW >= m_pMaxMinCAtom->getUpperBound() )
				return true;
			
			m_pMaxMinCAtom->setLowerBound(currW+1);
		}
		else if ( m_iMaxMin == 2 ) // min case
		{
			if ( currW <= m_pMaxMinCAtom->getLowerBound() )
				return true;
			
			m_pMaxMinCAtom->setUpperBound(currW-1);
		}
		// TODO: This is a temporary solution. When we change the
		// bounds of max/min w/c-atoms, we also need to update
		// the whole data structure so that all variables' values
		// are set correctly.
		//m_pMaxMinCAtom->maxMinResetVal();
		flip((m_pMaxMinCAtom->getAtoms())[0], ruleList);
		flip((m_pMaxMinCAtom->getAtoms())[0], ruleList);
		return false;
	}
//	bool optimalReached() const
//	{
//		if ( m_iMaxMin == 1 ) // max case
//			return (m_pMaxMinCAtom->getLowerBound() >=
//				m_pMaxMinCAtom->getUpperBound() + 1);
//		else if ( m_iMaxMin == 2 ) // min case
//			return (m_pMaxMinCAtom->getUpperBound() <=
//				m_pMaxMinCAtom->getLowerBound() - 1);
//	}
//	bool maxMinConverged() const
//	{
//		if ( m_iMaxMin != 1 && m_iMaxMin != 2 )
//			return true;
//		return (m_pMaxMinCAtom->getUpperBound() -
//			m_pMaxMinCAtom->getLowerBound() <= 2 );
//	}
	
	bool setForcedAtom(const char * text);
	void setInitValues(const char * fileName);
	bool setAtomText(int id, const char * text);
	bool printPositiveAtoms(bool smodelsOutput,
				int modelNum,
				bool showPredicate,
				const char * showAtom = NULL) const;
	void updateBestValues();
	void printBestValues();
	
	bool parseAtoms(const char * line, CRule & rule);
	bool parseAtomsInDimacsForm(const char * line, CRule & rule);
	bool fixIt(CRuleList & ruleList);
	void reinitialize(CRuleList & ruleList);

	int getTrueAtomNum() const { return m_iTrueAtomNum; }
	int getDoubleFlipNum() const { return m_iDoubleFlipNum; }
	int getPermutationFlipNum() const { return m_iPermutationFlipNum; }

	int getAtomLength() const { return m_iAtomLength; }
	void checkWeightAtoms() const 
	{
		int id = 0;
		for ( int i = 0; i < m_iCAtomLength; i++ )
		{
			if ( m_pCAtoms[i]->getType() != CAtom::W_ATOM )
				continue;
			if ( m_pCAtoms[i]->SAT() )
				continue;
			cout<<"Weight atom ["<<id++<<"] is unsat"<<endl;
		}
	}
};

/////////////////////////////////////////
// Class CRuleList Definition
/////////////////////////////////////////
class CRuleList
{
 protected:
	/* rules are organized as follow:
	 * unsat rules followed by sat rules.
	 * m_iLastUNSATIndex shows the end of unsat rules and
	 * the beginning of sat rules.
	 * when a rule becomes sat from unsat, switch it with the last
	 * unsat rule and m_iLastUNSATIndex--;
	 * when a rule becomes unsat from sat, switch it with the first
	 * sat rule and m_iLastUNSATIndex++;
	 */
	CArray<CRule *> m_pRules;
	int m_iRuleLength;
	int m_iLastUNSATIndex;

	/* the number of rules that currently are satisfied.
	 */
	int m_iSATRuleNum;

	void empty();
	
 public:
	CRuleList(int ruleSize = DEFAULT_SIZE);
	virtual ~CRuleList();
	
	void dump() const;

	bool verify();
	
	bool newRuleArray(int size);
	int addRule(CRule & rule);
	void shrinkArrays();
	
	bool allSAT() const;
	void swap(int index, bool currValue);
	int getUNSATRuleLength() const;
	const CRule & getUNSATRule(int id) const;

	void resetLastRecentlyFlipped();
	void resetTouched();
	
	void plotTouched(FILE * pPlotFile) const;

	bool setDBFRules(int atomNum, const char * fileName);
	
	void reinitialize();

	int getRuleLength() const { return m_iRuleLength; }

	void printUnitCAtomRules() const
	{
		for ( int i = 0; i < m_iRuleLength; i++ )
		{
			if ( m_pRules[i]->isUnitRule() &&
			     (m_pRules[i]->getCAtoms())[0]->getType() ==
			     CAtom::C_ATOM &&
			     !m_pRules[i]->SAT() )
				cerr<<"Unsat unit catom rule"<<endl;
		}
	};
};

/////////////////////////////////////////
// Class CWalkASPPS Definition
/////////////////////////////////////////
class CWalkASPPS
{
 protected:
	CAssignment m_Assignment;
	CRuleList m_RuleList;
	
	bool initializeByASPPSFile(const char * filename, bool beQuiet);
	bool initializeByDimacsFile(const char * filename, bool beQuiet);

	bool solve(long seed, bool beQuiet, bool noProcessBar, bool debugMode,
		   bool isCNFProg,
		   const char * filename, bool verify, long randSeed,
		   bool showAllPositiveAtoms,
		   bool showPredicate,
		   const char * showAtom = NULL);
	
 public:
	CWalkASPPS() {};
	virtual ~CWalkASPPS() {};

	void dump() const;

	bool toDIMACS(const char * filename);
	bool toASPPS(const char * filename);
	
	bool run(bool beQuiet, bool noProcessBar, bool debugMode,
		 bool isCNFProg,
		 const char * filename, bool verify, long randSeed,
		 bool showAllPositiveAtoms,
		 bool showPredicate,
		 const char * showAtom = NULL);
	bool verifyAssignment();
};

#endif
