#include "wsatcc.h"

#ifdef __LLN_DEBUG__
static FILE * pLog;
#endif

typedef enum { HEU_SKC, HEU_RNOVELTY_PLUS, HEU_NUM }
THeuristics;
static THeuristics glHeuristics = HEU_SKC;
static int glRNP_WP_NUMERATOR = 100; 
static int glRNP_WP_DENOMINATOR = 10000; 

static char STATUS_FILE_NAME[] = "wsatcc.stat";

static bool glUserInterrupted;

static int glMaxRetryNum = 10;				/* max # of tries    */
static int glMaxFlipNum = 100000;			/* max # of flips    */
static bool glStopOnFirst = false;			/* stop on the first *
							 * solution          */
static int glNumOfSolBeforeStop = 1;			/* default: stop after
							 * finding 1 solution*/

/* the number of the unit rules */
static int glUnitRuleNum = 0;

/* By default (10:100), only the last two digits of a random integer are
 * used. Use only the last one digit is a bad idea.
 */ 
static int glWalkProbabilityNumerator = 10;
static int glWalkProbabilityDenominator = 100;
static double glNoiseProbability = 
	(double)glWalkProbabilityNumerator /
        (double)glWalkProbabilityDenominator;

/* glUseVirtualCounts = true: use vbc even double flip or permutation
 * flip is used. It is a bad name because vbc will be always used if neither
 * double flip nor permutation flip is used.
 */
static bool glUseVirtualCounts = false;		

/* glUseMakeCount = true: use the make-count. Therefore, make-count of
 * atoms will be updated on each flip. Since it is time consuming, when
 * not needed, it should be set to false.
 */
static bool glUseMakeCount = false;

/* glUseTimeStamp = true: set and update the timestamp of each atom on
 * being flipped.
 */
static bool glUseTimeStamp = false;
static int glLastTimeStamp = 0;

/* glTabuLength defines the length of the tabu. Each atom has a timestamp
 * that indicates when it was lastly flipped. All atoms in a clause that
 * have their timestamp >= T_m - glTabuLength are in the tabu, where T_m
 * is the timestamp of the most recently flipped atom(s) in the clause.
 */
static int glTabuLength = 1;

/* glPureDoubleFlip = true: do not use double flip walk option.
 */
static bool glPureDoubleFlip = true;

/* glComboWeight = true: compute combined vbc when double flip or permutation
 * flip is used.
 */
static bool glComboWeight = true;

/* glDoubleFlipApplicable:
 * 0: no disjoint c-atoms or no c-atoms at all
 * 1: applicable for some c-atoms
 * 2: applicable for all c-atoms
 */
int glDoubleFlipApplicable; 
static bool glDisjointCatoms = false;
static bool glDisjointWatoms = false;
static int glDBFlipProbabilityNumerator = 0;
static int glDBFlipProbabilityDenominator = 100;
static double glDBFlipProbability = 
	(double)glDBFlipProbabilityNumerator /
        (double)glDBFlipProbabilityDenominator;

static bool glPermutationCatoms = false;
static int glInitialMethod = 1;
static int glCombinationMethod = 3;

static bool glWriteLogs = true;

/* Statistics - the number of times that a unit clause is picked
 */
static int glNumPickUnitRule = 0;
static int glOverFlow = 0;
static int glAllBestAtomsOverflow = 0;

/* Initially, we set an atom to true with probability glInitToTrueProb
 */
static int glInitNumerator = 50;
static int glInitDenominator = 100;
static double glInitToTrueProb = ((double)glInitNumerator /
				  (double)glInitDenominator);

static bool glTieBreaking = false;

static bool glSmodelsOutput = false;

static bool glPlotTrueAtomNum = false;
static bool glPlotUnsatRuleNum = false;
static int glPlotFrequency = 500;
static bool glPlotTouchedTimes = false;
static int glPlotLowerBound = 50;
static int glPlotHighest = glPlotLowerBound;
static int glPlotTouchedTimesThreshold = -1;
static bool glPlotTouchedTimesStart = false;

static bool glUseInitValues = false;
static bool glInitValuesError = false;
static char glInitFileName[MAX_FILENAME_LENGTH];
static bool glSetEscapePara = false;
/* The percentage of flips that use different noise ratio.
 * By default, it is 25%.
 */
static int glEscapeLimit = 25;
static int glEscapeProbNumerator = 50;
static int glEscapeProbDenominator = 100;

static bool glGeneralDoubleFlip = false;
static bool glHasSetDBFRules = false;
static char glGeneralDBFRuleFile[MAX_FILENAME_LENGTH];
static CArray<CRule *> glRulesUsedInDBF;
static int glCandidateNum;

static bool glHasEmptyRules = false;
static int glPrintAssignmentAfterFlip = -1;
static int glMinUnsatClauses = MAX_INT;

static bool glUseCorrectCase = true;

static double glAvgFlipPerSecond = MAX_INT;
static int glBinarySearchTimeLimit = 5; // in seconds
static double glBinarySearchJumpTime = 1;

int glMaxCombVal = 0;
int glCombBase = 0;
double glPower = 1.0;

static bool glShowPositiveAtomsInPBFormat = false;

////////////////////////////////////////////////////
// Global Functions
////////////////////////////////////////////////////

/* generate a random number r s.t. 0 <= r <= b-1
 */
inline int myRand(int b)
{
/*	This was suggested in "man 3 rand". But it is not significantly
 *	better than the intuitive method.
 *	return (int)((double)b*random()/(RAND_MAX+1.0));
 */
	return (random()%b);
}

/* Get the next integer starting @ line[pos]. Space characters are eliminated
 * while if the first non-space character is not a digit, NONE will be
 * returned.
 */
int getInt(const char * line, int & pos)
{
	char num[16];
	int i = pos, j = 0;
	int length = strlen(line);
	bool isNegative = false;
	
	while ( i < length && isspace(line[i]) )
		i++;

	if ( i >= length )
		return NONE;
	
	if ( !isdigit(line[i]) )
	{
		if ( line[i] != '-' )
			return NONE;
		else
		{
			isNegative = true;
			i++;
		}
	}
	
	while ( i < length &&
		j < 16 &&
		isdigit(line[i]) )
	{
		num[j] = line[i];
		j++;
		i++;
	}
	if ( i >= length || j >= 16 )
		return NONE;
	
	num[j] = '\0';
	int id = atoi(num);
	pos = i;
	if ( !isNegative )
		return id;
	else
		return (0 - id);
}

int getToken(const char * line, char * buffer, int size, int & pos)
{
	int i = pos, j = 0;
	int length = strlen(line);

	while ( i < length && isspace(line[i]) )
		i++;
	
	if ( i >= length || !isalpha(line[i]) )
		return NONE;
	
	while ( i < length &&
		j < size &&
		isalpha(line[i]) )
	{
		buffer[j] = line[i];
		j++;
		i++;
	}
	if ( i >= length || j >= size )
		return NONE;
	
	buffer[j] = '\0';
	int num = strlen(buffer);
	pos = i;
	return num;
}

int getAtomText(const char * line, char * buffer, int size, int & pos)
{
	int i = pos, j = 0;
	int length = strlen(line);

	while ( i < length && isspace(line[i]) )
		i++;
	
	while ( i < length &&
		j < size )
	{
		buffer[j] = line[i];
		j++;
		i++;
	}
	
	buffer[j] = '\0';
	for ( int k = 0; k < j; k++ )
	{
		if ( buffer[k] == '\n' )
		{
			buffer[k] = ' ';
		}
	}
	
	int num = strlen(buffer);
	pos = i;
	return num;
}

/* Compute the factorial of n (n!)
 * If downto is not 1, then compute n * (n-1) * ... * downto.
 */
inline long long fact(int n, int downto = 1)
{
	if ( downto <= 0 || n < downto )
		return 1;

	long long f = downto;
	for ( int i = downto+1; i <= n; i++ )
	{
		f *= i;
		if ( f < 0 || f >= MAX_INT )
		{
			f = -1;
			break;
		}
	}

	return f;
}

/* logfact(n, k) returns log10(n) + log10(n-1) + ... + log10(downto)
 */
inline int logfact(int n, int downto = 1)
{
	int f = 0;
	int temp = downto;
	
	if ( downto <= 0 || n < downto )
		return 1;
	for ( int i = downto+1; i <= n; i++ )
	{
		temp *= i;
		if ( temp < 0 )
		{
			f = -1;
			break;
		}
		if ( temp >= MAX_INT )
		{
			f += (int)::log10(temp);
			temp = 1;
		}
	}

	return f;
}

/* floating point version of fact(...)
 */
inline double dfact(int n, int downto = 1)
{
	double f = (double)downto;

	if ( downto <= 0 || n < downto )
		return 1;
	for ( double i = (double)(downto+1); i <= (double)n; i += 1.0 )
		f *= i;
	
	return f;
}

/* compute the # of combinations of choosing k from n
 * C(n, k) = n! / (k! * (n-k)!) or
 * C(n, k) = n * (n-1) * ... * (n-k+1) / k!
 *
 * Case 0 (deprecated):
 * using stirling approximation:
 * n! = sqrt(2Pi)*n^((n+1)/2)*e^(-n)
 * C(n, k) = n^((n+1)/2) / (sqrt(2Pi)*k^((k+1)/2)*(n-k)^((n-k+1)/2))
 * Case 1:
 * using linear approximation
 * Case 2:
 * using quadratic approximation
 * Case 3:
 * using loop computation, which may cause overflow
 */
inline int combination(int n, int k)
{
	if ( n < k || n < 0 || k < 0 )
		return 0;
	else if ( glCombinationMethod == 4 )
		return BIGC*(n-k+1);
	else if ( n == k )
		return 1;
	else if ( k == 1 || k == n-1 )
		return n;
	else if ( k == 2 || k == n-2 )
		return (n*(n-1)/2);
	else
	{
		int result = 1;

		double dn = (double)n;
		double dk = (double)k;
		double dn_k = dn - dk;

		int kprim = (k < n / 2 ? k : n-k);
		int a, b;
		int numerator, denominator;
		int nsqu;

		switch (glCombinationMethod)
		{
		    case 0:
			result = (int) ( pow(dn, (n+0.5)) /
					 ( sqrt(2*PI) *
					   pow(dk, (k+0.5)) *
					   pow(dn_k, (n-k+0.5)) ) );
			break;
		    case 1:
			if ( k < n / 2 )
			{
				result = n * k;
			}
			else
			{
				result = n * (n-k);
			}
			break;
		    case 2:
			nsqu = n*n;
			
			a = (nsqu-5*n+2)/4;
			b = (-nsqu+9*n-6)/4;
			result = a*k*k + b*k + 1;
			break;
		    case 3:
			numerator = fact(n, n-kprim+1);
			denominator = fact(kprim);

			if ( numerator < 0 || denominator <= 0 )
				result = -1;
			else if ( numerator < denominator )
				result = 1;
			else
				result = numerator / denominator;
			break;
		    case 4:
			result = BIGC*(n-k+1);
			break;
		    case 5:
			result = (int)((n+0.5)*::log(n) -
				       (n-k+0.5)*::log(n-k) -
				       (k+0.5)*::log(k));
			break;			
		    default:
			if ( k < n / 2 )
			{
				result = n * k;
			}
			else
			{
				result = n * (n-k);
			}
			break;
		}

		int big = OVER_RATE * MAX_INT;
		if ( result < 0 || result > big )
		{
			glOverFlow++;

			int approx = (int)((n+0.5)*::log(n) -
					   (n-k+0.5)*::log(n-k) -
					   (k+0.5)*::log(k))+1;
			
			result = glCombBase + approx;
			
			if ( result < 0 )
			{
				result = big;
			}
		}
		
		return result;
	}
}

inline int permutation(int n, int k)
{
	if ( n < k )
		return 0;
	else
		return fact(n, (n-k+1));
}

///////////////////////////////
// class CAtom
///////////////////////////////

CAtom::CAtom(int catomLength,
	     int ruleLength)
{
	m_Type = CAtom::N_ATOM;
	m_bValue = 0;
	m_iIndex = 0;
	m_uiTimeStamp = 0;
	m_iBreakCount = 0;
	m_iMakeCount = 0;
	m_iCAtomLength = 0;
	m_iCAtomRuleLength = 0;
	m_iUnitCAtomLength = 0;
	m_iSATCAtomNum = 0;
	m_iRuleLength = 0;
	m_cForced = 0;
	
	m_pCAtoms.newArray(catomLength);
	m_pCAtomRules.newArray(ruleLength);
	m_pUnitCAtoms.newArray(catomLength);
	m_pRules.newArray(ruleLength);
}

CAtom::CAtom(const CAtom & anAtom)
{
	*this = anAtom;
}

CAtom::~CAtom()
{
}

void CAtom::empty()
{
	m_iCAtomLength = m_iRuleLength = 0;
	
	m_pCAtoms.empty();
	m_pRules.empty();
}

void CAtom::dump() const
{
	cout<<"The Atom =>\n";
	cout<<"    m_Type = "<<m_Type<<endl;
	cout<<"    m_iIndex = "<<m_iIndex<<endl;
	cout<<"    m_sText = "<<m_sText<<endl;
	cout<<"    m_bValue = "<<(m_bValue?"TRUE":"FALSE")<<endl;
	cout<<"    m_uiTimeStamp = "<<m_uiTimeStamp<<endl;
	cout<<"    m_cForced = "<<(int)m_cForced<<endl;
	cout<<"    m_iRealBreakCount = "<<m_iBreakCount<<endl;
	cout<<"    m_iRealMakeCount = "<<m_iMakeCount<<endl;
	cout<<"    m_iBreakCount = "<<getBKCT()<<endl;
	cout<<"    m_iMakeCount = "<<getMKCT()<<endl;
	cout<<"    m_iCAtomLength = "<<m_iCAtomLength<<endl;
	cout<<"    m_iCAtomRuleLength = "<<m_iCAtomRuleLength<<endl;
	cout<<"    m_iUnitCAtomLength = "<<m_iUnitCAtomLength<<endl;
	cout<<"    m_iRuleLength = "<<m_iRuleLength<<endl;
}

void CAtom::dumpText() const
{
	cout<<m_sText<<"("<<getIndex()<<":"<<getTextVal()<<")";
}

void CAtom::dumpTextWithBKCT() const
{
	cout<<m_sText<<"("<<getIndex()<<":"<<getTextVal()<<" "<<getBKCT()<<")";
}

int CAtom::addRule(CRule & rule)
{
	m_iRuleLength++;
	return  m_pRules.add(&rule);
}

int CAtom::addCAtomRule(CRule & rule)
{
	m_iCAtomRuleLength++;
	return m_pCAtomRules.add(&rule);
}

int CAtom::addCAtom(CCAtom & catom)
{
	m_iCAtomLength++;
	return m_pCAtoms.add(&catom);
}

int CAtom::addUnitCAtom(CCAtom & catom)
{
	m_iUnitCAtomLength++;
	return m_pUnitCAtoms.add(&catom);
}

int CAtom::addUnitWAtom(CWeightAtom & catom)
{
	m_iUnitWAtomLength++;
	return m_pUnitWAtoms.add(&catom);
}

int CAtom::addGDBFRule(CRule & rule)
{
	return m_pGDBFRules.add(&rule);
}

void CAtom::shrinkArrays()
{
	m_iCAtomLength = m_pCAtoms.shrinkArray();
	m_iRuleLength = m_pRules.shrinkArray();
}

inline int CAtom::getRealBKCT() const
{
	return getBKCT(false);
}

inline int CAtom::getRealMKCT() const
{
	return getMKCT(false);
}

inline int CAtom::getCNFBKCT() const
{
	return m_iBreakCount;
}

inline int CAtom::getCNFMKCT() const
{
	return m_iBreakCount;
}

inline int CAtom::getBKCT(bool VBC) const
{
	return getBKCTCorrect(VBC);
}

inline int CAtom::getMKCT(bool VBC) const
{
	return getMKCTCorrect(VBC);
}

inline int CAtom::getBKCTWrong(bool VBC) const
{
	int newBKCT = m_iBreakCount;
	int U, V;

	// if flip this atom from false to true
	if ( !getValue() )
	{
		for ( int i = 0; i < m_iCAtomRuleLength; i++ )
		{
			if ( !m_pCAtomRules[i]->hasNoNormalTrueLiterals() &&
			     !m_pCAtomRules[i]->isCritical(*this) )
				continue;
			
			if ( !VBC && m_pCAtomRules[i]->isUnitRule() &&
			     (m_pCAtomRules[i]->getCAtoms())[0]->getType()
			     == CAtom::C_ATOM )
				continue;
			
			U = V = 1;

			CCAtom ** pCAtoms = m_pCAtomRules[i]->getCAtoms();

			// special case 1: when there is only one catom
			if ( m_pCAtomRules[i]->getCAtomLength() == 1 )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[0]) )
				{
					newBKCT += pCAtoms[0]->
						getBKSingleCAtomCase(false,
								     this);
				}

				continue;
			}
			
			for ( int j = 0;
			      j < m_pCAtomRules[i]->getCAtomLength();
			      j++ )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[j]) )
				{
					V *= pCAtoms[j]->
						getUV(false, true, true,
						      this);
					U *= pCAtoms[j]->
						getUV(false, true, false,
						      this);
				}
				else
				{
					V *= pCAtoms[j]->
						getUV(false, false, true,
						      this);
					U *= pCAtoms[j]->
						getUV(false, false, false,
						      this);
				}
			}
			
			newBKCT += (V - U);
		}
	}
	else
	{
		for ( int i = 0; i < m_iCAtomRuleLength; i++ )
		{
			if ( !m_pCAtomRules[i]->hasNoNormalTrueLiterals() &&
			     !m_pCAtomRules[i]->isCritical(*this) )
				continue;
			
			if ( !VBC && m_pCAtomRules[i]->isUnitRule() &&
			     (m_pCAtomRules[i]->getCAtoms())[0]->getType()
			     == CAtom::C_ATOM )
				continue;

			U = V = 1;

			CCAtom ** pCAtoms = m_pCAtomRules[i]->getCAtoms();

			// special case 1: when there is only one catom
			if ( m_pCAtomRules[i]->getCAtomLength() == 1 )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[0]) )
				{
					newBKCT += pCAtoms[0]->
						getBKSingleCAtomCase(true,
								     this);
				}
				
				continue;
			}

			for ( int j = 0;
			      j < m_pCAtomRules[i]->getCAtomLength();
			      j++ )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[j]) )
				{
					V *= pCAtoms[j]->
						getUV(true, true, true,
						      this);
					U *= pCAtoms[j]->
						getUV(true, true, false,
						      this);
				}
				else
				{
					V *= pCAtoms[j]->
						getUV(true, false, true,
						      this);
					U *= pCAtoms[j]->
						getUV(true, false, false,
						      this);
				}
			}
			
			newBKCT += V - U;
		}
	}

	return (newBKCT >= 0 ? newBKCT : MAX_INT);
}

inline int CAtom::getMKCTWrong(bool VMC) const
{
	int newMKCT = m_iMakeCount;
	int X, Y;
	
	// if flip this atom from false to true
	if ( !getValue() )
	{
		for ( int i = 0; i < m_iCAtomRuleLength; i++ )
		{
			if ( !m_pCAtomRules[i]->hasNoNormalTrueLiterals() )
				continue;
			
			if ( !VMC && m_pCAtomRules[i]->isUnitRule() &&
			     (m_pCAtomRules[i]->getCAtoms())[0]->getType()
			     == CAtom::C_ATOM )
				continue;

			X = Y = 1;

			CCAtom ** pCAtoms = m_pCAtomRules[i]->getCAtoms();

			// special case 1: when there is only one catom
			if ( m_pCAtomRules[i]->getCAtomLength() == 1 )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[0]) )
				{
					newMKCT += pCAtoms[0]->
						getMKSingleCAtomCase(false,
								     this);
				}

				continue;
			}
			
			for ( int j = 0;
			      j < m_pCAtomRules[i]->getCAtomLength();
			      j++ )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[j]) )
				{
					Y *= pCAtoms[j]->
						getXY(false, true, true,
						      this);
					X *= pCAtoms[j]->
						getXY(false, true, false,
						      this);
				}
				else
				{
					Y *= pCAtoms[j]->
						getXY(false, false, true,
						      this);
					X *= pCAtoms[j]->
						getXY(false, false, false,
						      this);
				}

			}			
			newMKCT += (Y - X);
		}
	}
	else
	{
		for ( int i = 0; i < m_iCAtomRuleLength; i++ )
		{
			if ( !m_pCAtomRules[i]->hasNoNormalTrueLiterals() )
				continue;

			if ( !VMC && m_pCAtomRules[i]->isUnitRule() &&
			     (m_pCAtomRules[i]->getCAtoms())[0]->getType()
			     == CAtom::C_ATOM )			     
				continue;

			X = Y = 1;

			CCAtom ** pCAtoms = m_pCAtomRules[i]->getCAtoms();

			// special case 1: when there is only one catom
			if ( m_pCAtomRules[i]->getCAtomLength() == 1 )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[0]) )
				{
					newMKCT += pCAtoms[0]->
						getMKSingleCAtomCase(true,
								     this);
				}

				continue;
			}

			for ( int j = 0;
			      j < m_pCAtomRules[i]->getCAtomLength();
			      j++ )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[j]) )
				{
					Y *= pCAtoms[j]->
						getXY(true, true, true,
						      this);
					X *= pCAtoms[j]->
						getXY(true, true, false,
						      this);
				}
				else
				{
					Y *= pCAtoms[j]->
						getXY(true, false, true,
						      this);
					X *= pCAtoms[j]->
						getXY(true, false, false,
						      this);
				}

			}			
			newMKCT += Y - X;
		}
	}

	return (newMKCT >= 0 ? newMKCT : MAX_INT);
}

inline int CAtom::getBKCTCorrect(bool VBC) const
{
	int newBKCT = m_iBreakCount;
	int U, V;

	// if flip this atom from false to true
	if ( !getValue() )
	{
		for ( int i = 0; i < m_iCAtomRuleLength; i++ )
		{
			bool satNormBefore = m_pCAtomRules[i]->
				satNormBeforeFlipping(*this);
			bool satNormAfter = m_pCAtomRules[i]->
				satNormAfterFlipping(*this);
			
			// The rule will still have true literals after
			// flipping *this
			if ( satNormAfter )
				continue;
			
			if ( !VBC && m_pCAtomRules[i]->isUnitRule() &&
			     (m_pCAtomRules[i]->getCAtoms())[0]->getType()
			     == CAtom::C_ATOM )
				continue;

			U = V = 1;

			CCAtom ** pCAtoms = m_pCAtomRules[i]->getCAtoms();

			// special case 1: when there is only one catom
			if ( m_pCAtomRules[i]->getCAtomLength() == 1 )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[0]) )
				{
					newBKCT += pCAtoms[0]->
						getBKSingleCAtomCase(false,
								     this);
				}
				
				// note satNormAfter is false
				if ( satNormBefore ) 
				{
					newBKCT += pCAtoms[0]->getUV(false,
								     false,
								     false,
								     this);
				}
				
				continue;
			}
			
			for ( int j = 0;
			      j < m_pCAtomRules[i]->getCAtomLength();
			      j++ )
			{
				if ( V == 0 && U == 0 )
				{
					break;
				}
				
				if ( m_pCAtoms.hasIt(pCAtoms[j]) )
				{
					if ( V != 0 )
					{
						V *= pCAtoms[j]->
							getUV(false, true,
							      true, this);
					}
					// Note: satNormAfter == false now
					if ( !satNormBefore && U != 0 )
					{
						U *= pCAtoms[j]->
							getUV(false, true,
							      false, this);
					}
				}
				else
				{
					if ( V != 0 )
					{
						V *= pCAtoms[j]->
							getUV(false, false,
							      true, this);
					}
					// Note: satNormAfter == false now
					if ( !satNormBefore && U != 0 )
					{
						U *= pCAtoms[j]->
							getUV(false, false,
							      false, this);
					}
				}
			}
			
			// Note: satNormAfter == false now
			if ( !satNormBefore )
				newBKCT += (V - U);
			else
				newBKCT += V;
		}
	}
	else
	{
		for ( int i = 0; i < m_iCAtomRuleLength; i++ )
		{
			bool satNormBefore = m_pCAtomRules[i]->
				satNormBeforeFlipping(*this);
			bool satNormAfter = m_pCAtomRules[i]->
				satNormAfterFlipping(*this);

			if ( satNormAfter )
				continue;

			if ( !VBC && m_pCAtomRules[i]->isUnitRule() &&
			     (m_pCAtomRules[i]->getCAtoms())[0]->getType()
			     == CAtom::C_ATOM )
				continue;

			U = V = 1;

			CCAtom ** pCAtoms = m_pCAtomRules[i]->getCAtoms();

			// special case 1: when there is only one catom
			if ( m_pCAtomRules[i]->getCAtomLength() == 1 )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[0]) )
				{
					newBKCT += pCAtoms[0]->
						getBKSingleCAtomCase(true,
								     this);
				}

				// note satNormAfter is false
				if ( satNormBefore )
				{
					newBKCT += pCAtoms[0]->getUV(true,
								     false,
								     false,
								     this);
				}
				
				continue;
			}

			for ( int j = 0;
			      j < m_pCAtomRules[i]->getCAtomLength();
			      j++ )
			{
				if ( V == 0 && U == 0 )
				{
					break;
				}

				if ( m_pCAtoms.hasIt(pCAtoms[j]) )
				{
					if ( V != 0 )
					{
						V *= pCAtoms[j]->
							getUV(true, true,
							      true, this);
					}
					// Note: satNormAfter == false now
					if ( !satNormBefore && U != 0 )
					{
						U *= pCAtoms[j]->
							getUV(true, true,
							      false, this);
					}
				}
				else
				{
					if ( V != 0 )
					{
						V *= pCAtoms[j]->
							getUV(true, false,
							      true, this);
					}
					// Note: satNormAfter == false now
					if ( !satNormBefore && U != 0 )
					{
						U *= pCAtoms[j]->
							getUV(true, false,
							      false, this);
					}
				}
			}
			
			// Note: satNormAfter == false now
			if ( !satNormBefore )
				newBKCT += (V - U);
			else
				newBKCT += V;
		}
	}

	return (newBKCT >= 0 ? newBKCT : MAX_INT);
}

inline int CAtom::getMKCTCorrect(bool VMC) const
{
	int newMKCT = m_iMakeCount;
	int X, Y;
	
	// if flip this atom from false to true
	if ( !getValue() )
	{
		for ( int i = 0; i < m_iCAtomRuleLength; i++ )
		{
			bool satNormBefore = m_pCAtomRules[i]->
				satNormBeforeFlipping(*this);
			bool satNormAfter = m_pCAtomRules[i]->
				satNormAfterFlipping(*this);

			if ( satNormBefore )
				continue;
			
			if ( !VMC && m_pCAtomRules[i]->isUnitRule() &&
			     (m_pCAtomRules[i]->getCAtoms())[0]->getType()
			     == CAtom::C_ATOM )
				continue;

			X = Y = 1;

			CCAtom ** pCAtoms = m_pCAtomRules[i]->getCAtoms();

			// special case 1: when there is only one catom
			if ( m_pCAtomRules[i]->getCAtomLength() == 1 )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[0]) )
				{
					newMKCT += pCAtoms[0]->
						getMKSingleCAtomCase(false,
								     this);
				}

				// note satNormBefore is false
				if ( satNormAfter )
				{
					newMKCT += pCAtoms[0]->getXY(false,
								     false,
								     false,
								     this);
				}

				continue;
			}
			
			for ( int j = 0;
			      j < m_pCAtomRules[i]->getCAtomLength();
			      j++ )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[j]) )
				{
					Y *= pCAtoms[j]->
						getXY(false, true, true,
						      this);
					// Note: satNormBefore == false now
					if ( !satNormAfter )
					{
						X *= pCAtoms[j]->
							getXY(false, true,
							      false, this);
					}
				}
				else
				{
					Y *= pCAtoms[j]->
						getXY(false, false, true,
						      this);
					// Note: satNormBefore == false now
					if ( !satNormAfter )
					{
						X *= pCAtoms[j]->
							getXY(false, false,
							      false, this);
					}
				}

			}

                        // Note: satNormBefore == false now
			if ( !satNormAfter )			
				newMKCT += (Y - X);
			else
				newMKCT += Y;
		}
	}
	else
	{
		for ( int i = 0; i < m_iCAtomRuleLength; i++ )
		{
			bool satNormBefore = m_pCAtomRules[i]->
				satNormBeforeFlipping(*this);
			bool satNormAfter = m_pCAtomRules[i]->
				satNormAfterFlipping(*this);

			if ( satNormBefore )
				continue;

			if ( !VMC && m_pCAtomRules[i]->isUnitRule() &&
			     (m_pCAtomRules[i]->getCAtoms())[0]->getType()
			     == CAtom::C_ATOM )			     
				continue;

			X = Y = 1;

			CCAtom ** pCAtoms = m_pCAtomRules[i]->getCAtoms();

			// special case 1: when there is only one catom
			if ( m_pCAtomRules[i]->getCAtomLength() == 1 )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[0]) )
				{
					newMKCT += pCAtoms[0]->
						getMKSingleCAtomCase(true,
								     this);
				}

				// note satNormBefore is false
				if ( satNormAfter )
				{
					newMKCT += pCAtoms[0]->getXY(true,
								     false,
								     false,
								     this);
				}

				continue;
			}

			for ( int j = 0;
			      j < m_pCAtomRules[i]->getCAtomLength();
			      j++ )
			{
				if ( m_pCAtoms.hasIt(pCAtoms[j]) )
				{
					Y *= pCAtoms[j]->
						getXY(true, true, true,
						      this);
					// Note: satNormBefore == false now
					if ( !satNormAfter )
					{
						X *= pCAtoms[j]->
							getXY(true, true,
							      false, this);
					}
				}
				else
				{
					Y *= pCAtoms[j]->
						getXY(true, false, true,
						      this);
					// Note: satNormBefore == false now
					if ( !satNormAfter )
					{
						X *= pCAtoms[j]->
							getXY(true, false,
							      false, this);
					}
				}

			}

			// Note: satNormBefore == false now
			if ( !satNormAfter )
				newMKCT += Y - X;
			else
				newMKCT += Y;
		}
	}

	return (newMKCT >= 0 ? newMKCT : MAX_INT);
}

int CAtom::negateSelf(CRuleList * pRuleList, CAtom * flipped)
{
	m_bValue = !m_bValue;

	if ( glUseTimeStamp )
	{
		glLastTimeStamp++;
		m_uiTimeStamp = glLastTimeStamp;
	}
	
	reeval(pRuleList);
	
	return (m_bValue?1:0);
}

int CAtom::reeval(CRuleList * pRuleList)
{
	bool currStatus;

	int oldSATCAtomNum = m_iSATCAtomNum;
	int changedSATCAtomNum;
	
	m_iSATCAtomNum = 0;
	for ( int i = 0; i < m_iCAtomLength; i++ )
	{
		m_pCAtoms[i]->reeval(*this);
		if ( m_pCAtoms[i]->SAT() )
			m_iSATCAtomNum++;
		m_pCAtoms[i]->needUpdate();
	}
	changedSATCAtomNum = oldSATCAtomNum - m_iSATCAtomNum;
	
	for ( int i = 0; i < m_iRuleLength; i++ )
	{
		if ( pRuleList == NULL )
			currStatus = m_pRules[i]->reeval();
		else
			currStatus = m_pRules[i]->reeval(this);
		
		if ( pRuleList != NULL )
		     pRuleList->swap(m_pRules[i]->getIndex(),
				     currStatus);
	}
	
	return changedSATCAtomNum;
}

inline bool CAtom::isCriticalInCAtom() const
{
	if ( m_iUnitCAtomLength <= 0 || m_iUnitCAtomLength > 1 )
		return false;

	return m_pUnitCAtoms[0]->canBreak(this);
}

inline bool CAtom::isCriticalInCAtom(int which) const
{
	if ( which >= m_iUnitCAtomLength || which < 0 )
		return false;

	return m_pUnitCAtoms[which]->canBreak(this);
}

inline bool CAtom::isCriticalInWAtom() const
{
	if ( m_iUnitWAtomLength <= 0 || m_iUnitWAtomLength > 1 )
		return false;

	return m_pUnitWAtoms[0]->canBreak(this);
}

inline bool CAtom::isCriticalInWAtom(int which) const
{
	if ( which >= m_iUnitWAtomLength || which < 0 )
		return false;

	return m_pUnitWAtoms[which]->canBreak(this);
}

const CAtom & CAtom::operator = (const CAtom & anAtom)
{
	empty();
	
	m_Type = anAtom.m_Type;
	m_iIndex = anAtom.m_iIndex;
	m_bValue = anAtom.m_bValue;
	strcpy(m_sText, anAtom.m_sText);
	m_iBreakCount = anAtom.m_iBreakCount;
	m_iMakeCount = anAtom.m_iMakeCount;
	
	m_iCAtomLength = anAtom.m_iCAtomLength;
	m_iCAtomRuleLength = anAtom.m_iCAtomRuleLength;
	m_iUnitCAtomLength = anAtom.m_iUnitCAtomLength;
	m_iSATCAtomNum = anAtom.m_iSATCAtomNum;
	m_iRuleLength = anAtom.m_iRuleLength;
	m_cForced = anAtom.m_cForced;
	
	m_pCAtoms = anAtom.m_pCAtoms;
	m_pCAtomRules = anAtom.m_pCAtomRules;
	m_pUnitCAtoms = anAtom.m_pUnitCAtoms;
	m_pRules = anAtom.m_pRules;
	
	return *this;
}

/////////////////////////////////
// class CCAtom
/////////////////////////////////
void CCAtom::partition(CArray<int> & indices,
		       int low, int high,
		       int & newLow, int & newHigh)
{
	int pivotValue = m_pAtoms[indices[low]]->getBKCT();
	int oldHigh = high;
	
	while ( low < high )
	{
		while ( low <= high &&
			m_pAtoms[indices[low]]->getBKCT() < pivotValue )
			low++;
		while ( high >= low &&
			m_pAtoms[indices[high]]->getBKCT() >= pivotValue )
			high--;
		if ( low > high )
			break;

		int tmp = indices[low];
		indices[low] = indices[high];
		indices[high] = tmp;
	}
	newHigh = low - 1;

	high = oldHigh;
	while ( low < high )
	{
		while ( low <= high &&
			m_pAtoms[indices[low]]->getBKCT() <= pivotValue )
			low++;
		while ( high >= low &&
			m_pAtoms[indices[high]]->getBKCT() > pivotValue )
			high--;
		if ( low > high )
			break;

		int tmp = indices[low];
		indices[low] = indices[high];
		indices[high] = tmp;
	}
	newLow = high + 1;
}

void CCAtom::quickSort(CArray<int> & indices,
		       int low, int high)
{
	int newLow, newHigh;

	if ( high > low )
	{
		partition(indices, low, high, newLow, newHigh);
		quickSort(indices, low, newHigh);
		quickSort(indices, newLow, high);
	}
}

CCAtom::CCAtom(int atomSize,
	       int ruleSize) 
	: CAtom(0, ruleSize)
{
	m_Type = CAtom::C_ATOM;
	m_iUpperBound = NO_UBOUND;
	m_iLowerBound = NO_LBOUND;
	m_iNumOfSatisfied = 0;
	m_bValue = true;
	m_bInUnitRule = false;

	needUpdate();
	
	m_iAtomLength = 0;
	m_pAtoms.newArray(atomSize);
}
			
CCAtom::CCAtom(const CCAtom & aCAtom)
{
	*this = aCAtom;
}

CCAtom::~CCAtom()
{
}

void CCAtom::empty()
{
	m_iAtomLength = 0;
	m_pAtoms.empty();
	CAtom::empty();
}

void CCAtom::dump() const
{
	cout<<"The C-Atom =>\n";
	cout<<"  ";
	CAtom::dump();
	cout<<"    m_iUpperBound = "<<m_iUpperBound<<endl;
	cout<<"    m_iLowerBound = "<<m_iLowerBound<<endl;
	cout<<"    m_iNumOfSatisfied = "<<m_iNumOfSatisfied<<endl;
	cout<<"    m_iAtomLength = "<<m_iAtomLength<<endl;
	cout<<"    m_bInUnitRule = "<<(m_bInUnitRule ? "Yes" : "No")<<endl;
}

void CCAtom::dumpText() const
{
	cout<<m_iLowerBound<<" { ";
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		m_pAtoms[i]->dumpText();
		cout<<" ";
	}
	cout<<" } "<<m_iUpperBound;
}

void CCAtom::dumpTextWithBKCT() const
{
	cout<<m_iLowerBound<<" { ";
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		m_pAtoms[i]->dumpTextWithBKCT();
		cout<<" ";
	}
	cout<<" } "<<m_iUpperBound;
}

int CCAtom::addAtom(CAtom & atom, int lb, int ub)
{
	m_pAtoms.add(&atom);
	m_iAtomLength++;

	if ( atom.getValue() )
		m_iNumOfSatisfied++;

	if ( lb <= m_iNumOfSatisfied &&
	     ub >= m_iNumOfSatisfied )
		m_bValue = true;
	else
		m_bValue = false;
	
	return m_iAtomLength-1;
}

void CCAtom::shrinkArrays()
{
	m_iAtomLength = m_pAtoms.shrinkArray();
	m_iRuleLength = m_pRules.shrinkArray();
}

inline void CCAtom::needUpdate()
{
	m_iBKSingleCAtomCaseFalse = NONE;
	m_iBKSingleCAtomCaseTrue = NONE;
	for ( unsigned char i = 0; i < UVPOSNUM; i++ )
		m_iaUVValue[i] = NONE;
	
	if ( glUseMakeCount )
	{
		m_iMKSingleCAtomCaseFalse = NONE;
		m_iMKSingleCAtomCaseTrue = NONE;
		for ( unsigned char i = 0; i < XYPOSNUM; i++ )
			m_iaXYValue[i] = NONE;
	}
}

inline int CCAtom::getBKSingleCAtomCase(bool atomValue,
					const CAtom *theOne)
{
	if ( atomValue == false &&
	     m_iBKSingleCAtomCaseFalse == NONE )
	{
		int n = getUpperBound();
		int posNum = getNumOfSatisfied();

		m_iBKSingleCAtomCaseFalse = combination(posNum+1, n+1)-
			combination(posNum, n+1);
	}

	if ( atomValue == true &&
	     m_iBKSingleCAtomCaseTrue == NONE )
	{
		int k = getAtomLength();
		int m = getLowerBound();
		int posNum = getNumOfSatisfied();
		int negNum = k - posNum;

		m_iBKSingleCAtomCaseTrue = combination(negNum+1, k-m+1)-
			combination(negNum,k-m+1);
	}
	
	if ( atomValue == false )
	{
		return m_iBKSingleCAtomCaseFalse;
	}
	else
	{
		return m_iBKSingleCAtomCaseTrue;
	}
}

inline int CCAtom::getMKSingleCAtomCase(bool atomValue,
					const CAtom *theOne)
{
	if ( atomValue == false &&
	     m_iMKSingleCAtomCaseFalse == NONE )
	{
		int k = getAtomLength();
		int m = getLowerBound();
		int negNum = k - getNumOfSatisfied();

		m_iMKSingleCAtomCaseFalse = combination(negNum, k-m+1)-
			combination(negNum-1, k-m+1);
	}

	if ( atomValue == true &&
	     m_iMKSingleCAtomCaseTrue == NONE )
	{
		int n = getUpperBound();
		int posNum = getNumOfSatisfied();

		m_iMKSingleCAtomCaseTrue = combination(posNum, n+1)-
			combination(posNum-1, n+1);
	}
	
	if ( atomValue == false )
	{
		return m_iMKSingleCAtomCaseFalse;
	}
	else
	{
		return m_iMKSingleCAtomCaseTrue;
	}
}

/*
 * atomValue    in    vfunction     UV pos
 *         0     0            0    UBAROUT
 *         0     0            1    VBAROUT
 *         0     1            0     UBARIN
 *         0     1            1     VBARIN
 *         1     0            0       UOUT
 *         1     0            1       VOUT
 *         1     1            0        UIN
 *         1     1            1        VIN
 */
/*
 * t: # of clauses that have exactly one true literal and the underlying
 *    variable of the literal is *theOne
 * f^2: # of clauses that are currently unsat and do not contain *theOne
 * V/Vbar = t+f^2
 * U/Ubar = f^2
 */
inline int CCAtom::getUV(bool atomValue, bool in, bool vfunction,
			 const CAtom *theOne)
{
	TUVType pos = TUVType((atomValue<<2) + (in<<1) + vfunction);

	if ( m_iaUVValue[pos] == NONE )
	{
		int k = getAtomLength();
		int n = getUpperBound();
		int m = getLowerBound();
		int posNum = getNumOfSatisfied();
		int negNum = k - posNum;

		switch ( pos )
		{
		    case VIN:
			m_iaUVValue[pos] = (combination(negNum+1, k-m+1) +
					    combination(posNum-1, n+1));
			break;
		    case VOUT:
			m_iaUVValue[pos] = m_iaXYValue[pos] =
				(combination(negNum, k-m+1) +
				 combination(posNum, n+1));
			break;
		    case UIN:
			m_iaUVValue[pos] = m_iaXYValue[pos] =
				(combination(negNum, k-m+1) +
				 combination(posNum-1, n+1));
			break;
		    case UOUT:
			m_iaUVValue[pos] = m_iaXYValue[pos] =
				(combination(negNum, k-m+1) +
				 combination(posNum, n+1));
			break;
		    case VBARIN:
			m_iaUVValue[pos] = (combination(posNum+1, n+1) +
					    combination(negNum-1, k-m+1));
			break;
		    case VBAROUT:
			m_iaUVValue[pos] = m_iaXYValue[pos] =
				(combination(posNum, n+1) +
				 combination(negNum, k-m+1));
			break;
		    case UBARIN:
			m_iaUVValue[pos] = m_iaXYValue[pos] =
				(combination(posNum, n+1) +
				 combination(negNum-1, k-m+1));
			break;
		    case UBAROUT:
			m_iaUVValue[pos] = m_iaXYValue[pos] =
				(combination(posNum, n+1) +
				 combination(negNum, k-m+1));
			break;
		    default:
			return 0;
		}		
	}

	return m_iaUVValue[pos];
}

/*
 * atomValue    in    vfunction     XY pos
 *         0     0            0    XBAROUT
 *         0     0            1    YBAROUT
 *         0     1            0     XBARIN
 *         0     1            1     YBARIN
 *         1     0            0       XOUT
 *         1     0            1       YOUT
 *         1     1            0        XIN
 *         1     1            1        YIN
 */
inline int CCAtom::getXY(bool atomValue, bool in, bool yfunction,
			 const CAtom *theOne)
{
	TXYType pos = TXYType((atomValue<<2) + (in<<1) + yfunction);

	if ( m_iaXYValue[pos] == NONE )
	{
		int k = getAtomLength();
		int n = getUpperBound();
		int m = getLowerBound();
		int posNum = getNumOfSatisfied();
		int negNum = k - posNum;

		switch ( pos )
		{
		    case YIN:
			m_iaXYValue[pos] = (combination(negNum, k-m+1) +
					    combination(posNum, n+1));
			break;
		    case YOUT:
			m_iaXYValue[pos] = m_iaUVValue[pos] =
				(combination(negNum, k-m+1) +
				 combination(posNum, n+1));
			break;
		    case XIN:
			m_iaXYValue[pos] = m_iaUVValue[pos] =
				(combination(negNum, k-m+1) +
				 combination(posNum-1, n+1));
			break;
		    case XOUT:
			m_iaXYValue[pos] = m_iaUVValue[pos] =
				(combination(negNum, k-m+1) +
				 combination(posNum, n+1));
			break;
		    case YBARIN:
			m_iaXYValue[pos] = (combination(posNum, n+1) +
					    combination(negNum, k-m+1));
			break;
		    case YBAROUT:
			m_iaXYValue[pos] = m_iaUVValue[pos] =
				(combination(posNum, n+1) +
				 combination(negNum, k-m+1));
			break;
		    case XBARIN:
			m_iaXYValue[pos] = m_iaUVValue[pos] =
				(combination(posNum, n+1) +
				 combination(negNum-1, k-m+1));
			break;
		    case XBAROUT:
			m_iaXYValue[pos] = m_iaUVValue[pos] =
				(combination(posNum, n+1) +
				 combination(negNum, k-m+1));
			break;
		    default:
			return 0;
		}
	}

	return m_iaXYValue[pos];
}

// The following two functions are not correct. They are not used at
// this moment. However, we need to fix them if we want to follow
// wsat(pb) algorithm for c-atoms.
inline int CCAtom::getBKCT(bool VBC) const
{
	int flippedCount = 0;
	int needFlipFalse;
	int needFlipTrue;
	int sumBKCT = 0;
	
	if ( m_bValue )
	{
		needFlipFalse = m_iNumOfSatisfied - m_iLowerBound + 1;
		needFlipTrue = m_iUpperBound - m_iNumOfSatisfied + 1;
	}
	else
	{
		needFlipTrue = m_iLowerBound - m_iNumOfSatisfied;
		needFlipFalse = m_iNumOfSatisfied - m_iUpperBound;
	}
	
	if ( (needFlipFalse > 0 && needFlipFalse < needFlipTrue) ||
	     needFlipTrue <= 0 )
	{
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( m_pAtoms[i]->getValue() )
			{
				sumBKCT += m_pAtoms[i]->getBKCT();
				flippedCount++;
				if ( flippedCount == needFlipFalse )
					break;
			}
		}
	}
	else
	{
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( !m_pAtoms[i]->getValue() )
			{
				sumBKCT += m_pAtoms[i]->getBKCT();
				flippedCount++;
				if ( flippedCount == needFlipTrue )
					break;
			}
		}
	}
	
	return (flippedCount > 0 ? sumBKCT / flippedCount : 1);
}

inline int CCAtom::getMKCT(bool VMC) const
{
	int flippedCount = 0;
	int needFlipFalse;
	int needFlipTrue;
	int sumMKCT = 0;
	
	if ( m_bValue )
	{
		needFlipFalse = m_iNumOfSatisfied - m_iLowerBound + 1;
		needFlipTrue = m_iUpperBound - m_iNumOfSatisfied + 1;
	}
	else
	{
		needFlipTrue = m_iLowerBound - m_iNumOfSatisfied;
		needFlipFalse = m_iNumOfSatisfied - m_iUpperBound;
	}
	
	if ( (needFlipFalse > 0 && needFlipFalse < needFlipTrue) ||
	     needFlipTrue <= 0 )
	{
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( m_pAtoms[i]->getValue() )
			{
				sumMKCT += m_pAtoms[i]->getMKCT();
				flippedCount++;
				if ( flippedCount == needFlipFalse )
					break;
			}
		}
	}
	else
	{
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( !m_pAtoms[i]->getValue() )
			{
				sumMKCT += m_pAtoms[i]->getMKCT();
				flippedCount++;
				if ( flippedCount == needFlipTrue )
					break;
			}
		}
	}
	
	return (flippedCount > 0 ? sumMKCT / flippedCount : 1);
}

bool CCAtom::canBreak(const CAtom * atom)
{
	if ( m_bPositive && m_bValue )
	{
		if ( m_iNumOfSatisfied == m_iUpperBound &&
		     !atom->getValue() )
			return true;
		if ( m_iNumOfSatisfied == m_iLowerBound &&
		     atom->getValue() )
			return true;
	}
	else if ( !m_bPositive && !m_bValue )
	{
		if ( m_iNumOfSatisfied == m_iUpperBound+1 &&
		     atom->getValue() )
			return true;
		if ( m_iNumOfSatisfied == m_iLowerBound-1 &&
		     !atom->getValue() )
			return true;
	}
	return false;
}

bool CCAtom::canMake(const CAtom * atom)
{
	if ( m_bPositive && !m_bValue )
	{
		if ( m_iNumOfSatisfied == m_iUpperBound+1 &&
		     atom->getValue() )
			return true;
		if ( m_iNumOfSatisfied == m_iLowerBound-1 &&
		     !atom->getValue() )
			return true;
	}
	else if ( !m_bPositive && m_bValue )
	{
		if ( m_iNumOfSatisfied == m_iUpperBound &&
		     !atom->getValue() )
			return true;
		if ( m_iNumOfSatisfied == m_iLowerBound &&
		     atom->getValue() )
			return true;
	}
	return false;
}

int CCAtom::negateSelf(CRuleList * pRuleList, CAtom * flipped)
{
	int deltaTrueAtom = 0;
	
	if ( SAT() )
	{
		int bestFlipToTrue = -1, bestFlipToFalse = -1;
		int bkctFlipToTrue, bkctFlipToFalse;

		bkctFlipToTrue = bkctFlipToFalse = MAX_INT;
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( m_pAtoms[i]->getValue() &&
			     m_pAtoms[i]->getRealBKCT() < bkctFlipToTrue &&
			     !m_pAtoms[i]->forced() )
			{
				bestFlipToTrue = i;
				bkctFlipToTrue = m_pAtoms[i]->getRealBKCT();
			}
			
			if ( !m_pAtoms[i]->getValue() &&
			     m_pAtoms[i]->getRealBKCT() < bkctFlipToFalse &&
			     !m_pAtoms[i]->forced() )
			{
				bestFlipToFalse = i;
				bkctFlipToFalse = m_pAtoms[i]->getRealBKCT();
			}
		}

		if ( bestFlipToTrue != -1 && bestFlipToFalse != -1 )
		{
			m_pAtoms[bestFlipToTrue]->negateSelf(pRuleList);
			m_pAtoms[bestFlipToFalse]->negateSelf(pRuleList);
		}
		
		return 0;
	}

	int direction;
	CArray<int> sortedID(m_iAtomLength);
	sortedID.setLength(m_iAtomLength);
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		sortedID[i] = i;
	}
		
	quickSort(sortedID, 0, m_iAtomLength-1);
		
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		if ( m_pAtoms[sortedID[i]] == flipped )
			continue;
			
		if ( m_iNumOfSatisfied < m_iLowerBound )
			direction = 1;
		else
			direction = 0;
		
		if ( (direction == 1 &&
		      m_pAtoms[sortedID[i]]->getValue()) ||
		     (direction == 0 &&
		      !m_pAtoms[sortedID[i]]->getValue()) )
			continue;
		
		m_pAtoms[sortedID[i]]->negateSelf(pRuleList);
		if ( m_pAtoms[sortedID[i]]->getValue() )
		{
			deltaTrueAtom++;
		}
		else
		{
			deltaTrueAtom--;
		}
		
		if ( m_bValue )
		{
			break;
		}
	}

	return deltaTrueAtom;
}

bool CCAtom::reeval(const CAtom & flippedAtom)
{
	bool oldValue = m_bValue;

	if ( flippedAtom.getValue() )
		m_iNumOfSatisfied++;
	else
		m_iNumOfSatisfied--;

	if ( m_iNumOfSatisfied <= m_iUpperBound &&
	     m_iNumOfSatisfied >= m_iLowerBound )
		m_bValue = true;
	else
		m_bValue = false;

	// if the value of this c-atom is changed, we need to update
	// the # of true literals in the corresponding clause.
	if ( (m_bPositive && oldValue && !m_bValue) ||
	     (!m_bPositive && !oldValue && m_bValue) )
		for ( int i = 0; i < m_iRuleLength; i++ )
			m_pRules[i]->decreaseTrueLiteralNum();

	if ( (m_bPositive && !oldValue && m_bValue) ||
	     (!m_bPositive && oldValue && !m_bValue) )
		for ( int i = 0; i < m_iRuleLength; i++ )
			m_pRules[i]->increaseTrueLiteralNum();

	return m_bValue;
}

// Return TRUE if and only if flipping the atom will not improve the status
// of satisfiability of the catom
// In case when the catom literal is UNSAT, that means flipping *atom will
// not bring the catom literal closer to SAT
// In case when the catom literal is SAT, that means flipping *atom will
// make the literal UNSAT
bool CCAtom::noHelp(const CAtom * atom) const
{
	bool found = false;
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		if ( m_pAtoms[i] == atom )
		{
			found = true;
			break;
		}
	}
	if ( !found )
		return true;
	
	if ( m_bPositive && !m_bValue )
	{
		// increase m_iNumOfSatisfied when it is already over UB
		if ( m_iNumOfSatisfied > m_iUpperBound &&
		     !atom->getValue() )
			return true;
		if ( m_iNumOfSatisfied < m_iLowerBound &&
		     atom->getValue() )
			return true;
	}

	return false;
}

const CCAtom & CCAtom::operator = (const CCAtom & catom)
{
	m_Type = catom.m_Type;
	m_bValue = catom.m_bValue;
	strcpy(m_sText, catom.m_sText);
	m_iBreakCount = catom.m_iBreakCount;
	m_iMakeCount = catom.m_iMakeCount;
	
	m_iUpperBound = catom.m_iUpperBound;
	m_iLowerBound = catom.m_iLowerBound;
	m_iNumOfSatisfied = catom.m_iNumOfSatisfied;
	m_bPositive = catom.m_bPositive;
	m_bInUnitRule = catom.m_bInUnitRule;
	
	m_iAtomLength = catom.m_iAtomLength;
	m_iRuleLength = catom.m_iRuleLength;
	
	m_pAtoms = catom.m_pAtoms;
	m_pRules = catom.m_pRules;

	return *this;
}

//////////////////////////
// class CWeightAtom
//////////////////////////
CWeightAtom::CWeightAtom(int atomSize,
			 int ruleSize) 
	: CCAtom(atomSize, ruleSize)
{
	m_Type = CAtom::W_ATOM;
	m_iWeights.newArray(atomSize);
	m_iCurrentWeight = 0;
	m_iTotalWeight = 0;
	m_iTotalABSNegWeight = 0;
	m_iTotalABSWeight = 0;
}
			
CWeightAtom::CWeightAtom(const CWeightAtom & aWeightAtom)
{
	*this = aWeightAtom;
}

void CWeightAtom::dump() const
{
	cout<<"The Weight-Atom =>\n";
	cout<<"  ";
	CCAtom::dump();
	cout<<"    m_iTotalWeight = "<<m_iTotalWeight<<endl;
	cout<<"    m_iTotalABSNegWeight = "<<m_iTotalABSNegWeight<<endl;
	cout<<"    m_iTotalABSWeight = "<<m_iTotalABSWeight<<endl;
	cout<<"    m_iCurrentWeight = "<<m_iCurrentWeight<<endl;
	for ( int i = 0; i < m_iAtomLength; i++ )
		cout<<"      weight( "<<m_pAtoms[i]->getText()
		    <<"= "<<m_pAtoms[i]->getTextVal()
		    <<" ) = "<<m_iWeights[i]<<endl;
}

void CWeightAtom::dumpText() const
{
	cout<<m_iLowerBound<<" [ ";
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		m_pAtoms[i]->dumpText();
		cout<<" = "<<m_iWeights[i]<<" ";
	}
	cout<<" ] "<<m_iUpperBound;
}

void CWeightAtom::dumpTextWithBKCT() const
{
	cout<<m_iLowerBound<<" [ ";
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		cout<<"<";
		m_pAtoms[i]->dumpTextWithBKCT();
		cout<<" = "<<m_iWeights[i]<<"> ";
	}
	cout<<" ] "<<m_iUpperBound;
}

inline int CWeightAtom::getBKSingleCAtomCase(bool atomValue,
					const CAtom *theOne)
{
	int pos = getPosition(theOne); // pos shouldn't be NONE
	int weight = getWeight(pos);
	int result;
	
	// when weight is negative:
	// l <= W <= u =>
	// m_iTotalABSNegWeight+l <= |W| <= m_iTotalABSNegWeight+u
	// where |W| is the set in which all weights are changed to their
	// absolute values
	// m_iTotalWeight should also take m_iTotalABSNegWeight+its value
	if ( weight < 0 )
	{
		atomValue = !atomValue;
		weight = -weight;
	}
	
	if ( atomValue == false )
	{
		int n = getUpperBound() + m_iTotalABSNegWeight;
		int posNum = getCurrentWeight() + m_iTotalABSNegWeight;

		result = (combination(posNum+weight, n+1)-
			  combination(posNum, n+1));
	}
	else
	{
		int k = m_iTotalABSWeight;
		int m = getLowerBound() + m_iTotalABSNegWeight;
		int posNum = getCurrentWeight() + m_iTotalABSNegWeight;
		int negNum = k - posNum;

		result = (combination(negNum+weight, k-m+1)-
			  combination(negNum, k-m+1));
	}

	return result/getFactor();
}

inline int CWeightAtom::getMKSingleCAtomCase(bool atomValue,
					const CAtom *theOne)
{
	int pos = getPosition(theOne); // pos shouldn't be NONE
	int weight = getWeight(pos);
	int result;
	
	if ( weight < 0 )
	{
		atomValue = !atomValue;
		weight = -weight;
	}

	if ( atomValue == false )
	{
		int k = m_iTotalABSWeight;
		int m = getLowerBound() + m_iTotalABSNegWeight;
		int posNum = getCurrentWeight() + m_iTotalABSNegWeight;
		int negNum = k - posNum;

		result = (combination(negNum, k-m+1)-
			  combination(negNum-weight, k-m+1));
	}
	else
	{
		int n = getUpperBound() + m_iTotalABSNegWeight;
		int posNum = getCurrentWeight() + m_iTotalABSNegWeight;

		result = (combination(posNum, n+1)-
			  combination(posNum-weight, n+1));
	}

	return result/getFactor();
}

/*
 * atomValue    in    vfunction     UV pos
 *         0     0            0    UBAROUT
 *         0     0            1    VBAROUT
 *         0     1            0     UBARIN
 *         0     1            1     VBARIN
 *         1     0            0       UOUT
 *         1     0            1       VOUT
 *         1     1            0        UIN
 *         1     1            1        VIN
 */
/*
 * t: # of clauses that have exactly one true literal and the underlying
 *    variable of the literal is *theOne
 * f^2: # of clauses that are currently unsat and do not contain *theOne
 * V/Vbar = t+f^2
 * U/Ubar = f^2
 */
inline int CWeightAtom::getUV(bool atomValue, bool in, bool vfunction,
			      const CAtom *theOne)
{
	int weight = 0;
	if ( in )
	{
		int wpos = getPosition(theOne);
		weight = getWeight(wpos);		
	}
	if ( weight < 0 )
	{
		atomValue = !atomValue;
		weight = -weight;
	}
	
	TUVType pos = TUVType((atomValue<<2) + (in<<1) + vfunction);

	// Lengning --- changed on March 20, 2006
	// int k = getTotalWeight();
	// int n = getUpperBound();
	// int m = getLowerBound();
	// int posNum = getCurrentWeight();
	int k = m_iTotalABSWeight;
	int n = getUpperBound() + m_iTotalABSNegWeight;
	int m = getLowerBound() + m_iTotalABSNegWeight;
	int posNum = getCurrentWeight() + m_iTotalABSNegWeight;
	int negNum = k - posNum;
	
	switch ( pos )
	{
	    case VIN:
		m_iaUVValue[pos] = (combination(negNum+weight, k-m+1)+
				    combination(posNum-weight, n+1));
		break;
	    case VOUT:
		m_iaUVValue[pos] = m_iaXYValue[pos] =
			(combination(negNum, k-m+1) +
			 combination(posNum, n+1));
		break;
	    case UIN:
		m_iaUVValue[pos] = m_iaXYValue[pos] =
			(combination(negNum, k-m+1) +
			 combination(posNum-weight, n+1));
		break;
	    case UOUT:
		m_iaUVValue[pos] = m_iaXYValue[pos] =
			(combination(negNum, k-m+1) +
			 combination(posNum, n+1));
		break;
	    case VBARIN:
		m_iaUVValue[pos] = (combination(posNum+weight, n+1)+
				    combination(negNum-weight, k-m+1));
		break;
	    case VBAROUT:
		m_iaUVValue[pos] = m_iaXYValue[pos] =
			(combination(posNum, n+1) +
			 combination(negNum, k-m+1));
		break;
	    case UBARIN:
		m_iaUVValue[pos] = m_iaXYValue[pos] =
			(combination(posNum, n+1) +
			 combination(negNum-weight, k-m+1));
		break;
	    case UBAROUT:
		m_iaUVValue[pos] = m_iaXYValue[pos] =
			(combination(posNum, n+1) +
			 combination(negNum, k-m+1));
		break;
	    default:
		return 0;
	}

	return m_iaUVValue[pos]/getFactor();
}

/*
 * atomValue    in    vfunction     XY pos
 *         0     0            0    XBAROUT
 *         0     0            1    YBAROUT
 *         0     1            0     XBARIN
 *         0     1            1     YBARIN
 *         1     0            0       XOUT
 *         1     0            1       YOUT
 *         1     1            0        XIN
 *         1     1            1        YIN
 */
inline int CWeightAtom::getXY(bool atomValue, bool in, bool yfunction,
			      const CAtom *theOne)
{
	TXYType pos = TXYType((atomValue<<2) + (in<<1) + yfunction);
	int weight = 0;
	if ( in )
	{
		int wpos = getPosition(theOne);
		weight = getWeight(wpos);		
	}

	// Lengning --- changed on March 20, 2006
	// int k = getTotalWeight();
	// int n = getUpperBound();
	// int m = getLowerBound();
	// int posNum = getCurrentWeight();
	int k = m_iTotalABSWeight;
	int n = getUpperBound() + m_iTotalABSNegWeight;
	int m = getLowerBound() + m_iTotalABSNegWeight;
	int posNum = getCurrentWeight() + m_iTotalABSNegWeight;
	int negNum = k - posNum;
	
	switch ( pos )
	{
	    case YIN:
		m_iaXYValue[pos] = (combination(negNum, k-m+1) +
				    combination(posNum, n+1));
		break;
	    case YOUT:
		m_iaXYValue[pos] = m_iaUVValue[pos] =
			(combination(negNum, k-m+1) +
			 combination(posNum, n+1));
		break;
	    case XIN:
		m_iaXYValue[pos] = m_iaUVValue[pos] =
			(combination(negNum, k-m+1) +
			 combination(posNum-weight, n+1));
		break;
	    case XOUT:
		m_iaXYValue[pos] = m_iaUVValue[pos] =
			(combination(negNum, k-m+1) +
			 combination(posNum, n+1));
		break;
	    case YBARIN:
		m_iaXYValue[pos] = (combination(posNum, n+1) +
				    combination(negNum, k-m+1));
		break;
	    case YBAROUT:
		m_iaXYValue[pos] = m_iaUVValue[pos] =
			(combination(posNum, n+1) +
			 combination(negNum, k-m+1));
		break;
	    case XBARIN:
		m_iaXYValue[pos] = m_iaUVValue[pos] =
			(combination(posNum, n+1) +
			 combination(negNum-weight, k-m+1));
		break;
	    case XBAROUT:
		m_iaXYValue[pos] = m_iaUVValue[pos] =
			(combination(posNum, n+1) +
			 combination(negNum, k-m+1));
		break;
	    default:
		return 0;
	}

	return m_iaXYValue[pos]/getFactor();
}

int CWeightAtom::addAtom(CAtom & atom, int weight, int lb, int ub)
{
	int * pIndex = new int;
	*pIndex = m_pAtoms.add(&atom);
	m_iWeights.add(weight);
	m_ihIndex.insert(atom.getIndex(), pIndex);
	
	if ( weight < 0 )
	{
		m_iTotalABSNegWeight += -weight;
		m_iTotalABSWeight += -weight;
	}
	else
	{
		m_iTotalWeight += weight;
		m_iTotalABSWeight += weight;
	}
	
	m_iAtomLength++;

	if ( atom.getValue() )
	{
		m_iNumOfSatisfied++;
		m_iCurrentWeight += weight;
	}

	if ( lb <= m_iCurrentWeight &&
	     ub >= m_iCurrentWeight )
		m_bValue = true;
	else
		m_bValue = false;
	
	return m_iAtomLength-1;
}

const CWeightAtom & CWeightAtom::operator = (const CWeightAtom & weightatom)
{
	(CCAtom)(*this) = (CCAtom)weightatom;

	m_iWeights = weightatom.m_iWeights;
	m_iCurrentWeight = weightatom.m_iCurrentWeight;
	m_iTotalWeight = weightatom.m_iTotalWeight;
	m_iTotalABSNegWeight = weightatom.m_iTotalABSNegWeight;
	m_iTotalABSWeight = weightatom.m_iTotalABSWeight;
	m_ihIndex = weightatom.m_ihIndex;
	
	return *this;
}

// The following two functions are not correct. They are not used at
// this moment. However, we need to fix them if we want to follow
// wsat(pb) algorithm for c-atoms.
inline int CWeightAtom::getBKCT(bool VBC) const
{
	return 0;
}

inline int CWeightAtom::getMKCT(bool VMC) const
{
	return 0;
}

bool CWeightAtom::canBreak(const CAtom * atom)
{
	if ( m_bPositive && m_bValue )
	{
		int weight = getWeight(atom);
		bool sat = atom->getValue();
		
		if ( sat && weight > 0 ) 
		{
			if ( m_iCurrentWeight - weight < m_iLowerBound )
				return true;
		} else if ( !sat && weight < 0 )
		{
			if ( m_iCurrentWeight + weight < m_iLowerBound )
				return true;
		} else if ( sat && weight < 0 )
		{
			if ( m_iCurrentWeight - weight > m_iUpperBound )
				return true;
		} else if ( !sat && weight > 0 )
		{
			if ( m_iCurrentWeight + weight > m_iUpperBound )
				return true;
		}
	}
	else if ( !m_bPositive && !m_bValue )
	{
		int weight = getWeight(atom);
		bool sat = atom->getValue();
		
		if ( sat && weight > 0 ) 
		{
			if ( m_iCurrentWeight - weight < m_iUpperBound )
				return true;
		} else if ( !sat && weight < 0 )
		{
			if ( m_iCurrentWeight + weight < m_iUpperBound )
				return true;
		} else if ( sat && weight < 0 )
		{
			if ( m_iCurrentWeight - weight > m_iLowerBound )
				return true;
		} else if ( !sat && weight > 0 )
		{
			if ( m_iCurrentWeight + weight > m_iLowerBound )
				return true;
		}
	}
	
	return false;
}

bool CWeightAtom::canMake(const CAtom * atom)
{
	if ( m_bPositive && !m_bValue )
	{
		int weight = getWeight(atom);
		bool sat = atom->getValue();
		
		if ( sat && weight > 0 ) 
		{
			if ( m_iCurrentWeight - weight <= m_iUpperBound )
				return true;
		} else if ( !sat && weight < 0 )
		{
			if ( m_iCurrentWeight - (-weight) <= m_iUpperBound )
				return true;
		} else if ( sat && weight < 0 )
		{
			if ( m_iCurrentWeight + (-weight) >= m_iLowerBound )
				return true;
		} else if ( !sat && weight > 0 )
		{
			if ( m_iCurrentWeight + weight >= m_iLowerBound )
				return true;
		}
	}
	else if ( !m_bPositive && m_bValue )
	{
		int weight = getWeight(atom);
		bool sat = atom->getValue();
		
		if ( sat && weight > 0 ) 
		{
			if ( m_iCurrentWeight - weight < m_iLowerBound )
				return true;
		} else if ( !sat && weight < 0 )
		{
			if ( m_iCurrentWeight - (-weight) < m_iLowerBound )
				return true;
		} else if ( sat && weight < 0 )
		{
			if ( m_iCurrentWeight + (-weight) > m_iUpperBound )
				return true;
		} else if ( !sat && weight > 0 )
		{
			if ( m_iCurrentWeight + weight > m_iUpperBound )
				return true;
		}
	}
	
	return false;
}

int CWeightAtom::multiFlip(CRuleList * pRuleList, CAtom * toBeFlipped)
{
	if ( toBeFlipped == NULL || !m_bValue )
		return 0;
	
	int deltaTrueAtom = 0;
	int pos = getPosition(toBeFlipped);
	// for directions, if it is 1, then we need to flip atoms to increase
	// weight; if it is 0, then we need to flip atoms to decrease weight.
	int direction;
	CArray<int> sortedID(m_iAtomLength);
	int maxTimeStamp = 0;
	int weightChanged = (toBeFlipped->getValue()?
			     -m_iWeights[pos] :
			     m_iWeights[pos]);
	int newWeight = m_iCurrentWeight + weightChanged;
	

	if ( newWeight <= m_iUpperBound &&
	     newWeight >= m_iLowerBound )
		return 0;
	
	sortedID.setLength(m_iAtomLength);
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		sortedID[i] = i;
		if ( m_pAtoms[i]->getTimeStamp() > maxTimeStamp )
			maxTimeStamp = m_pAtoms[i]->getTimeStamp();
	}
		
	quickSort(sortedID, 0, m_iAtomLength-1);
		
	if ( newWeight < m_iLowerBound )
		direction = 1;
	else
		direction = 0;

	int sumWeight = 0;
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		if ( m_pAtoms[sortedID[i]] == toBeFlipped )
			continue;
		if ( m_pAtoms[sortedID[i]]->getTimeStamp() >=
		     maxTimeStamp )
			continue;
		
		if ( (direction == 1 &&
		      m_pAtoms[sortedID[i]]->getValue() &&
		      m_iWeights[sortedID[i]] > 0) ||
		     (direction == 1 &&
		      !m_pAtoms[sortedID[i]]->getValue() &&
		      m_iWeights[sortedID[i]] < 0 ) ||
		     (direction == 0 &&
		      !m_pAtoms[sortedID[i]]->getValue() &&
		      m_iWeights[sortedID[i]] > 0 ) ||
		     (direction == 0 &&
		      m_pAtoms[sortedID[i]]->getValue() &&
		      m_iWeights[sortedID[i]] < 0 ) )
			continue;
		
		m_pAtoms[sortedID[i]]->negateSelf(pRuleList);
		sumWeight += abs(m_iWeights[sortedID[i]]);
		if ( m_pAtoms[sortedID[i]]->getValue() )
		{
			deltaTrueAtom++;
		}
		else
		{
			deltaTrueAtom--;
		}
		
		if ( sumWeight >= abs(weightChanged) )
		{
			break;
		}
	}
	
	return deltaTrueAtom;
}

// note return the number of atoms whose values have been changed !(To do)
int CWeightAtom::negateSelf(CRuleList * pRuleList, CAtom * flipped)
{
	int deltaTrueAtom = 0;
	
        // use a randomly selected start position
	int startingPos = myRand(m_iAtomLength);

	// for safety, if we have flipped all atoms and still cannot negate it
	// we can escape from an infinite loop
	int safeGuard = 0;

	// for directions, if it is 1, then we need to flip atoms to increase
	// weight; if it is 0, then we need to flip atoms to decrease weight.
	int direction;

	// if it is currently satisfied, then find the first n necessary atoms
	// to flip w.r.t. the chosen startPos.
	if ( getValue() )
	{
		if ( m_iCurrentWeight - m_iLowerBound >
		     m_iUpperBound - m_iCurrentWeight )
			direction = 1;
		else
			direction = 0;
		
		while ( m_iCurrentWeight <= m_iUpperBound &&
			m_iCurrentWeight >= m_iLowerBound )
		{
			safeGuard = 0;
			while ( (direction == 1 &&
				 m_pAtoms[startingPos]->getValue() &&
				 m_iWeights[startingPos] > 0) ||
				(direction == 1 &&
				 !m_pAtoms[startingPos]->getValue() &&
				 m_iWeights[startingPos] < 0 ) ||
				(direction == 0 &&
				 !m_pAtoms[startingPos]->getValue() &&
				 m_iWeights[startingPos] > 0 ) ||
				(direction == 0 &&
				 m_pAtoms[startingPos]->getValue() &&
				 m_iWeights[startingPos] < 0 ) )
			{
				startingPos = (startingPos+1)%
					m_iAtomLength;
				safeGuard++;
				if ( safeGuard > m_iAtomLength )
					return deltaTrueAtom;
			}
			
			m_pAtoms[startingPos]->negateSelf(pRuleList);
			if ( m_pAtoms[startingPos]->getValue() )
			{
				deltaTrueAtom++;
			}
			else
			{
				deltaTrueAtom--;
			}

			if ( !m_bValue )
				break;
		}
	}
	else
	{
		CArray<int> sortedID(m_iAtomLength);
		int maxTimeStamp = 0;
		sortedID.setLength(m_iAtomLength);
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			sortedID[i] = i;
			if ( m_pAtoms[i]->getTimeStamp() > maxTimeStamp )
				maxTimeStamp = m_pAtoms[i]->getTimeStamp();
		}
		
		quickSort(sortedID, 0, m_iAtomLength-1);
		
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( m_pAtoms[sortedID[i]] == flipped )
				continue;
			if ( m_pAtoms[sortedID[i]]->getTimeStamp() >=
			     maxTimeStamp )
				continue;
			
			if ( m_iCurrentWeight < m_iLowerBound )
				direction = 1;
			else
				direction = 0;

			if ( (direction == 1 &&
			      m_pAtoms[sortedID[i]]->getValue() &&
			      m_iWeights[sortedID[i]] > 0) ||
			     (direction == 1 &&
			      !m_pAtoms[sortedID[i]]->getValue() &&
			      m_iWeights[sortedID[i]] < 0 ) ||
			     (direction == 0 &&
			      !m_pAtoms[sortedID[i]]->getValue() &&
			      m_iWeights[sortedID[i]] > 0 ) ||
			     (direction == 0 &&
			      m_pAtoms[sortedID[i]]->getValue() &&
			      m_iWeights[sortedID[i]] < 0 ) )
				continue;

			m_pAtoms[sortedID[i]]->negateSelf(pRuleList);
			if ( m_pAtoms[sortedID[i]]->getValue() )
			{
				deltaTrueAtom++;
			}
			else
			{
				deltaTrueAtom--;
			}
			
			if ( m_bValue )
			{
				break;
			}
		}
	}
	
	return deltaTrueAtom;
}

// When this function is called, the value of "flippedAtom" has been changed.
bool CWeightAtom::reeval(const CAtom & flippedAtom)
{
	bool oldValue = m_bValue;
	int weight = getWeight(&flippedAtom);
	
	if ( flippedAtom.getValue() )
	{
		m_iNumOfSatisfied++;
		m_iCurrentWeight += weight;
	}
	else
	{
		m_iNumOfSatisfied--;
		m_iCurrentWeight -= weight;
	}

	if ( m_iCurrentWeight <= m_iUpperBound &&
	     m_iCurrentWeight >= m_iLowerBound )
		m_bValue = true;
	else
		m_bValue = false;

	// if the value of this c-atom is changed, we need to update
	// the # of true literals in the corresponding clause.
	if ( (m_bPositive && oldValue && !m_bValue) ||
	     (!m_bPositive && !oldValue && m_bValue) )
		for ( int i = 0; i < m_iRuleLength; i++ )
			m_pRules[i]->decreaseTrueLiteralNum();

	if ( (m_bPositive && !oldValue && m_bValue) ||
	     (!m_bPositive && oldValue && !m_bValue) )
		for ( int i = 0; i < m_iRuleLength; i++ )
			m_pRules[i]->increaseTrueLiteralNum();

	return m_bValue;
}

// Return TRUE if and only if flipping the atom will not improve the status
// of satisfiability of the catom
// In case when the catom literal is UNSAT, that means flipping *atom will
// not bring the catom literal closer to SAT
// In case when the catom literal is SAT, that means flipping *atom will
// make the literal UNSAT
bool CWeightAtom::noHelp(const CAtom * atom) const
{
	bool found = false;
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		if ( m_pAtoms[i] == atom )
		{
			found = true;
			break;
		}
	}
	if ( !found )
		return true;
	
	if ( m_bPositive && !m_bValue )
	{
		int weight = getWeight(atom);
		bool sat = atom->getValue();
		
		if ( sat && weight > 0 ) 
		{
			if ( m_iCurrentWeight < m_iLowerBound )
				return true;
		} else if ( !sat && weight < 0 )
		{
			if ( m_iCurrentWeight < m_iLowerBound )
				return true;
		} else if ( sat && weight < 0 )
		{
			if ( m_iCurrentWeight > m_iUpperBound )
				return true;
		} else if ( !sat && weight > 0 )
		{
			if ( m_iCurrentWeight > m_iUpperBound )
				return true;
		}
	}

	return false;
}

//////////////////////////
// class CRule
//////////////////////////

CRule::CRule(int bodySize,
	     int headSize)
{
	m_bBodySAT = true;
	m_bHeadSAT = false;
	m_bSAT = false;
	m_iTrueLiteralNum = -1;
	m_iNormalTrueLiteralNum = -1;

	m_iLastRecentlyFlipped = 0;
	
	m_bIsCritical = false;

	m_iAtomLength = 0;
	m_pAllAtoms.newArray(bodySize+headSize);
	m_iNormAtomLength = 0;
	m_pAllNormAtoms.newArray(0);
	m_iCAtomLength = 0;
	m_pAllCAtoms.newArray(1);

	m_iTotalTouchedTimes = 0;
	
	m_iBodyLength = 0;
	m_iHeadLength = 0;
	
	m_pBody.newArray(bodySize);
	m_pHead.newArray(headSize);
}

CRule::~CRule()
{
}

void CRule::empty()
{
	m_iAtomLength = 0;
	m_pAllAtoms.empty();
	m_iNormAtomLength = 0;
	m_pAllNormAtoms.empty();
	
	m_iBodyLength = m_iHeadLength = 0;
	
	m_pBody.empty();
	m_pHead.empty();
}

void CRule::dump() const
{
	cout<<"The Rule =>\n";
	cout<<"    m_iIndex = "<<m_iIndex<<endl;
	cout<<"    m_iOrigIndex = "<<m_iOrigIndex<<endl;
	cout<<"    m_iTrueLiteralNum = "<<m_iTrueLiteralNum<<endl;
	cout<<"    m_iNormalTrueLiteralNum = "<<m_iNormalTrueLiteralNum<<endl;
	cout<<"    m_iLastRecentlyFlipped = "<<m_iLastRecentlyFlipped<<endl;
	cout<<"    m_iAtomLength = "<<m_iAtomLength<<endl;
	cout<<"    m_iNormAtomLength = "<<m_iNormAtomLength<<endl;
	cout<<"    m_iCAtomLength = "<<m_iCAtomLength<<endl;
	cout<<"    m_iBodyLength = "<<m_iBodyLength<<endl;
	cout<<"    m_iHeadLength = "<<m_iHeadLength<<endl;
	cout<<"    m_bSAT = "<<m_bSAT<<endl;
	cout<<"    m_iTouchedTimes = "<<m_iTouchedTimes<<endl;
	cout<<"    m_iTotalTouchedTimes = "<<m_iTotalTouchedTimes<<endl;
	cout<<"    m_bUsedInDBF = "
	    <<(m_bUsedInDBF ? "true" : "false")<<endl;
}

void CRule::dumpText() const
{
	for ( int i = 0; i < m_iBodyLength; i++ )
	{
		m_pBody[i]->dumpText();
		cout<<" ";
	}

	cout<<" -> ";
	
	for ( int i = 0; i < m_iHeadLength; i++ )
	{
		m_pHead[i]->dumpText();
	}
	cout<<endl;
}

void CRule::dumpTextWithBKCT() const
{
	for ( int i = 0; i < m_iBodyLength; i++ )
	{
		m_pBody[i]->dumpTextWithBKCT();
		cout<<" , ";
	}

	cout<<" -> ";
	
	for ( int i = 0; i < m_iHeadLength; i++ )
	{
		m_pHead[i]->dumpTextWithBKCT();
		cout<<" | ";
	}
	cout<<endl;
}

int CRule::addAtom(CAtom & atom)
{
	if ( atom.getType() == CAtom::N_ATOM )
	{
		m_pAllAtoms.add(&atom);
		m_iAtomLength++;
	}
	else if ( atom.getType() == CAtom::C_ATOM ||
		  atom.getType() == CAtom::W_ATOM )
	{
		// never add catom as a normal atom 'cause we are using
		// virtual break count now

		m_pAllCAtoms.add((CCAtom *)&atom);
		m_iCAtomLength++;

		CAtom ** pCAtoms = ((CCAtom &)atom).getAtoms();
		int length = ((CCAtom &)atom).getAtomLength();
		
		for ( int i = 0; i < length; i++ )
		{
			// Note: be aware of the case when the same
			// atom occurs both as a regular literal and
			// inside some c/w-atoms (esp. when it occurs
			// both in the head and in the body
			if ( !m_pAllAtoms.hasIt(pCAtoms[i]) )
			{
				m_pAllAtoms.add(pCAtoms[i]);
				m_iAtomLength++;
			}
		}
	}

	return m_iAtomLength-1;
}

int CRule::addBodyAtom(CAtom & atom)
{
	if ( atom.getType() == CAtom::C_ATOM ||
	     atom.getType() == CAtom::W_ATOM )
		((CCAtom &)atom).setNegative();
	
	addAtom(atom);
	
	m_pBody.add(&atom);
	m_iBodyLength++;

	if ( m_bBodySAT )
	{
		m_bBodySAT = m_bBodySAT && atom.getValue();
		m_bSAT = !m_bBodySAT || m_bHeadSAT;
	}
	
	return m_iBodyLength-1;
}

int CRule::addHeadAtom(CAtom & atom)
{
	if ( atom.getType() == CAtom::C_ATOM ||
	     atom.getType() == CAtom::W_ATOM )
		((CCAtom &)atom).setPositive();

	addAtom(atom);
	
	m_pHead.add(&atom);
	m_iHeadLength++;

	if ( !m_bHeadSAT && !m_bSAT )
	{
		m_bHeadSAT = m_bHeadSAT || atom.getValue();
		m_bSAT = !m_bBodySAT || m_bHeadSAT;
	}
	
	return m_iHeadLength-1;
}

void CRule::shrinkArrays()
{
	m_iAtomLength = m_pAllAtoms.shrinkArray();
	m_iBodyLength = m_pBody.shrinkArray();
	m_iHeadLength = m_pHead.shrinkArray();
}

bool CRule::reeval()
{
	for ( int i = 0; i < m_iBodyLength; i++ )
		if ( !m_pBody[i]->getRealValue() )
		{
			return true;
		}
	
	for ( int i = 0; i < m_iHeadLength; i++ )
		if ( m_pHead[i]->getRealValue() )
		{
			return true;
		}

	return false;
}

bool CRule::init()
{
	m_iTrueLiteralNum = 0;
	m_iNormalTrueLiteralNum = 0;
	CAtom * trueLit = NULL;
	
	for ( int i = 0; i < m_iBodyLength; i++ )
	{
		if ( !m_pBody[i]->getValue() )
		{
			trueLit = m_pBody[i];
			m_iTrueLiteralNum++;
			if ( m_pBody[i]->getType() == CAtom::N_ATOM )
				m_iNormalTrueLiteralNum++;
		}
	}
	
	for ( int i = 0; i < m_iHeadLength; i++ )
	{
		if ( m_pHead[i]->getValue() )
		{
			trueLit = m_pHead[i];
			m_iTrueLiteralNum++;
			if ( m_pHead[i]->getType() == CAtom::N_ATOM )
				m_iNormalTrueLiteralNum++;
		}
	}

	if ( m_iTrueLiteralNum == 0 )
	{
		if ( glUseMakeCount )
		{
			for ( int i = 0; i < m_iBodyLength; i++ )
				m_pBody[i]->increaseMKCT();
			for ( int i = 0; i < m_iHeadLength; i++ )
				m_pHead[i]->increaseMKCT();
		}
	}
	else if ( m_iTrueLiteralNum == 1 )
	{
		trueLit->increaseBKCT();
	}

	/* check if it is a unit rule */
	if ( m_pBody.getLength() + m_pHead.getLength() == 1 )
	{
		if ( (m_pBody.getLength() == 1 && m_pBody[0]->forced() == 1) ||
		     (m_pHead.getLength() == 1 && m_pHead[0]->forced() == -1) )
		{
			cout<<"Unsatisfiable because of two contradict unit"
			    <<" rules\n";
			exit(1);
		}
		
		if ( m_pBody.getLength() == 1 )
			m_pBody[0]->setForced(true);
		else
			m_pHead[0]->setForced(false);
		glUnitRuleNum++;
	}
	
	if ( m_iTrueLiteralNum > 0 )
		m_bSAT = true;
	else
		m_bSAT = false;
	
	return m_bSAT;
}

// Note: If the rule doesn't contain c-atoms,then the breakcount of the
//       flipped atom will be updated.  Otherwise, only the # of true
//       literals will be updated and the "virtual" break count will be
//       computed in the atom's getBKCT() function.
// Note: m_iTrueLiteralNum is only useful in rules that don't have c-atoms;
//       m_iNormalTrueLiteralNum is only useful in rules that have c-atoms.
bool CRule::reeval(CAtom * theFlippedAtom)
{
	if ( theFlippedAtom == NULL )
		return reeval();

	// We want to keep the atoms that have the largest timestamp
	// in the front so that rnovelty+ algorithm can refer to them
	// easily.
	if ( glUseTimeStamp )
	{
		if ( theFlippedAtom->getTimeStamp() >=
		     m_pAllAtoms[m_iLastRecentlyFlipped]->getTimeStamp() )
		{
			updateMostRecentlyFlipped(theFlippedAtom);
		}
	}
	
	bool flippedBodyAtom = false;
	
	if ( m_pBody.hasIt(theFlippedAtom) )
		flippedBodyAtom = true;
	else if ( m_pHead.hasIt(theFlippedAtom) )
		flippedBodyAtom = false;
	else
	{
		// since neither the body nor the head has this atom,
		// flipping it won't cause the # of normal true literals
		// to change.
		m_bSAT = m_iTrueLiteralNum > 0;
		return m_bSAT;
	}

	bool currValue = theFlippedAtom->getValue();

	if ( ((flippedBodyAtom && currValue) ||
	      (!flippedBodyAtom && !currValue)) )
	{
		m_iTrueLiteralNum--;
		if ( m_iCAtomLength != 0 )
		{
			m_iNormalTrueLiteralNum--;
			m_bSAT = m_iTrueLiteralNum > 0;
			return m_bSAT;
		}

		// number changed from 1 to 0
		if ( m_iTrueLiteralNum == 0 )
		{
			theFlippedAtom->decreaseBKCT();
			if ( glUseMakeCount )
			{
				for ( int i = 0; i < m_iBodyLength; i++ )
					m_pBody[i]->increaseMKCT();
				for ( int i = 0; i < m_iHeadLength; i++ )
					m_pHead[i]->increaseMKCT();
			}
		}
		// number changed from 2 to 1
		else if ( m_iTrueLiteralNum == 1 )
		{
			bool notFound = true;
			CAtom * ptr;
			
			for ( int i = 0; i < m_iBodyLength &&
				         notFound; i++ )
			{
				ptr = m_pBody[i];
				if ( !ptr->getValue() )
				{
					ptr->increaseBKCT();
					notFound = false;
				}
			}
			
			for ( int i = 0; i < m_iHeadLength &&
				         notFound; i++ )
			{
				ptr = m_pHead[i];
				if ( ptr->getValue() )
				{
					ptr->increaseBKCT();
					notFound = false;
				}
			}
		}
	}
	else if ( ((flippedBodyAtom && !currValue) ||
		   (!flippedBodyAtom && currValue)) )
	{
		m_iTrueLiteralNum++;
		if ( m_iCAtomLength != 0 )
		{
			m_iNormalTrueLiteralNum++;
			m_bSAT = m_iTrueLiteralNum > 0;
			return m_bSAT;
		}

		// number changed from 0 to 1
		if ( m_iTrueLiteralNum == 1 )
		{
			theFlippedAtom->increaseBKCT();
			if ( glUseMakeCount )
			{
				for ( int i = 0; i < m_iBodyLength; i++ )
					m_pBody[i]->decreaseMKCT();
				for ( int i = 0; i < m_iHeadLength; i++ )
					m_pHead[i]->decreaseMKCT();
			}
		}
		// number changed from 1 to 2
		else if ( m_iTrueLiteralNum == 2 )
		{	
			bool notFound = true;
			CAtom * ptr;
			
			for ( int i = 0; i < m_iBodyLength &&
				         notFound; i++ )
			{
				ptr = m_pBody[i];
				if ( (ptr->getIndex() !=
				     theFlippedAtom->getIndex()) &&
				     !ptr->getValue() )
				{
					ptr->decreaseBKCT();
					notFound = false;
				}
			}
			
			for ( int i = 0; i < m_iHeadLength &&
				         notFound; i++ )
			{
				ptr = m_pHead[i];
				if ( (ptr->getIndex() !=
				     theFlippedAtom->getIndex()) &&
				     ptr->getValue() )
				{
					ptr->decreaseBKCT();
					notFound = false;
				}
			}
		}
	}

	if ( m_iTrueLiteralNum > 0 )
		m_bSAT = true;
	else
		m_bSAT = false;

	return m_bSAT;
}

void CRule::swap(int index, bool currValue, bool isInBody)
{
	int length = (isInBody ? m_iBodyLength : m_iHeadLength);
	int lastIndex = (isInBody ? m_iBodyLastFalseIndex : 
				    m_iHeadLastTrueIndex);
	CArray<CAtom *> & pList = (isInBody ? m_pBody : m_pHead);
	bool value = (isInBody ? !currValue : currValue);

	if ( index < 0 || index > length )
		return;

	CAtom * temp;
	if ( value ) // the literal just changed from false to true
	{
		temp = pList[lastIndex+1];
		pList.put(lastIndex+1, pList[index]);
		pList.put(index, temp);
		if ( isInBody )
			m_iBodyLastFalseIndex++;
		else
			m_iHeadLastTrueIndex++;
	}
	else // the literal just changed from true to false
	{
		temp = pList[lastIndex];
		pList.put(lastIndex, pList[index]);
		pList.put(index, temp);
		if ( isInBody )
			m_iBodyLastFalseIndex--;
		else
			m_iHeadLastTrueIndex--;
	}
}

void CRule::checkUnitRule()	
{
	if ( m_iBodyLength + m_iHeadLength == 1 &&
	     m_iCAtomLength == 1 )
	{
		m_pAllCAtoms[0]->inUnitRule();
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( glDisjointCatoms &&
			     m_pAllCAtoms[0]->getType() == CAtom::C_ATOM )
			{
				m_pAllAtoms[i]->addUnitCAtom(*m_pAllCAtoms[0]);
				if ( m_pAllAtoms[i]->getUnitCAtomNum() != 1 )
				{
					if ( glDoubleFlipApplicable == 2 ||
					     glDoubleFlipApplicable == 1 )
						glDoubleFlipApplicable = 1;
					else
						glDoubleFlipApplicable = 0;
				}
				else
				{
					if ( glDoubleFlipApplicable == 0 ||
					     glDoubleFlipApplicable == 2 )
						glDoubleFlipApplicable = 2;
					else
						glDoubleFlipApplicable = 1;
				}
			}
			else if ( glDisjointWatoms &&
				  m_pAllCAtoms[0]->getType() == CAtom::W_ATOM )
			{
				m_pAllAtoms[i]->addUnitWAtom(*((CWeightAtom *)
							     (m_pAllCAtoms[0])));
				if ( m_pAllAtoms[i]->getUnitWAtomNum() != 1 )
				{
					if ( glDoubleFlipApplicable == 2 ||
					     glDoubleFlipApplicable == 1 )
						glDoubleFlipApplicable = 1;
					else
						glDoubleFlipApplicable = 0;
				}
				else
				{
					if ( glDoubleFlipApplicable == 0 ||
					     glDoubleFlipApplicable == 2 )
						glDoubleFlipApplicable = 2;
					else
						glDoubleFlipApplicable = 1;
				}
			}
		}
	}
}

bool CRule::checkTrivialRule()
{
	for ( int i = 0; i < m_iBodyLength; i++ )
	{
		if ( m_pBody[i]->getType() != CAtom::N_ATOM )
			continue;
		for ( int j = 0; j < m_iHeadLength; j++ )
		{
			if ( m_pHead[j]->getType() != CAtom::N_ATOM )
				continue;
			if ( m_pBody[i] == m_pHead[j] )
				return true;
		}
	}
	return false;
}

bool CRule::noHelp(const CAtom * theAtom) const
{
	if ( isInHead(*theAtom) && !theAtom->getValue() )
		return false;
	if ( isInBody(*theAtom) && theAtom->getValue() )
		return false;

	for ( int i = 0; i < m_iCAtomLength; i++ )
	{
		if ( !m_pAllCAtoms[i]->noHelp(theAtom) )
			return false;
	}

	return true;
}

const CRule & CRule::operator = (const CRule & rule)
{
	m_iOrigIndex = rule.m_iOrigIndex;
	m_iIndex = rule.m_iIndex;
	m_bSAT = rule.m_bSAT;
	m_iTrueLiteralNum = rule.m_iTrueLiteralNum;
	m_iNormalTrueLiteralNum = rule.m_iNormalTrueLiteralNum;
	m_bIsCritical = rule.m_bIsCritical;
	m_bBodySAT = rule.m_bBodySAT;
	m_bHeadSAT = rule.m_bHeadSAT;
	
	m_iAtomLength = rule.m_iAtomLength;
	m_pAllAtoms = rule.m_pAllAtoms;
	m_iNormAtomLength = rule.m_iNormAtomLength;
	m_pAllNormAtoms = rule.m_pAllNormAtoms;
	m_iCAtomLength = rule.m_iCAtomLength;
	m_pAllCAtoms = rule.m_pAllCAtoms;
	
	m_iBodyLength = rule.m_iBodyLength;
	m_iHeadLength = rule.m_iHeadLength;

	m_pBody = rule.m_pBody;
	m_pHead = rule.m_pHead;

	m_iTouchedTimes = rule.m_iTouchedTimes;
	m_iTotalTouchedTimes = rule.m_iTotalTouchedTimes;
	m_bUsedInDBF = rule.m_bUsedInDBF;
	
	return *this;
}

//////////////////////////////
// class CAssignment
//////////////////////////////

CAssignment::CAssignment(int atomSize,
			 int catomSize)
{
	m_iForcedAtomLength = 0;
	m_pForcedAtoms.newArray();
	
	m_iAtomLength = 0;
	m_iCAtomLength = 0;

	m_iDoubleFlipNum = 0;
	m_iPermutationFlipNum = 0;

	m_iMaxMin = 0;
	
	m_pAtoms.newArray(atomSize);
	m_pCAtoms.newArray(catomSize);
	m_baBestValues.newArray(atomSize);
}

CAssignment::~CAssignment()
{
}

void CAssignment::empty()
{
	for ( int i = 0; i < m_iForcedAtomLength; i++ )
		delete m_pForcedAtoms[i];
	
	for ( int i = 0; i < m_iCAtomLength; i++ )
		delete m_pCAtoms[i];

	for ( int i = 0; i < m_iAtomLength; i++ )
		delete m_pAtoms[i];
	
	m_iAtomLength = m_iCAtomLength = 0;

	m_pAtoms.empty();
	m_pCAtoms.empty();
	m_baBestValues.empty();
}

bool CAssignment::newAtomArray(int size)
{
	if ( size > 0 )
	{
#ifdef GENERAL_INPUT		
		m_iAtomLength = 0;
		m_pAtoms.newArray(size);
#else
		m_iAtomLength = size;
		m_pAtoms.newArray(size);
		m_pAtoms.setLength(size);
		m_baBestValues.newArray(size);
		m_baBestValues.setLength(size);
#endif		

		for ( int i = 0; i < m_pAtoms.getSize(); i++ )
			m_pAtoms.put(i, NULL);
		return true;
	}
	else
	{
		m_iAtomLength = 0;
		m_pAtoms.empty();
		m_baBestValues.empty();
		return false;
	}
}

void CAssignment::dump() const
{
	int posAtomNum = 0;
	cout<<"The Assignment =>\n";
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		cout<<"  ";
		m_pAtoms[i]->dump();
		if ( m_pAtoms[i]->getValue() )
			posAtomNum++;
	}

	for ( int i = 0; i < m_iCAtomLength; i++ )
	{
		cout<<"  ";
		m_pCAtoms[i]->dump();
	}

	cout<<"Total Positive Atom Num: "<<posAtomNum<<endl;
}

int CAssignment::hasTheAtom(int id) const
{
#ifdef GENERAL_INPUT	
	int pos;
	
	for ( pos = m_iAtomLength - 1; pos >= 0; pos-- )
		if ( m_pAtoms[pos]->getIndex() == id )
			break;

	if ( pos >= 0 )
		return pos;
	else
		return NONE;
#else
	if ( id <= 0 || id > m_iAtomLength )
		return NONE;
	if ( m_pAtoms[id-1] == NULL )
		return NONE;
	if ( m_pAtoms[id-1]->getIndex() != id )
		return NONE;
	return id-1;
#endif	
}

int CAssignment::addAtom(CAtom & atom)
{
#ifdef GENERAL_INPUT	
	m_iAtomLength++;
	m_baBestValues.add(false);
	return m_pAtoms.add(&atom);

#else
	// be careful: don't need to change m_iAtomLength.
	//             it must be set outside this class.
	int id = atom.getIndex() - 1;

	m_pAtoms.put(id, &atom);
	m_baBestValues.put(id, false);
	return id;
#endif	
}

int CAssignment::addCAtom(CCAtom & catom)
{
	m_iCAtomLength++;
	return m_pCAtoms.add(&catom);
}

void CAssignment::shrinkArrays()
{
	m_iAtomLength = m_pAtoms.shrinkArray();
	m_iCAtomLength = m_pCAtoms.shrinkArray();
	m_baBestValues.shrinkArray();
	
	for ( int i = 0; i < m_iAtomLength; i++ )
		m_pAtoms[i]->shrinkArrays();
	for ( int i = 0; i < m_iCAtomLength; i++ )
		m_pCAtoms[i]->shrinkArrays();
}

// By Deborah East -
// parse the given line which should have the form of
// a b c , d e f means
// a and b and c -> d or e or f
// any of a,b, ... can be replace with a cardinality atom
// { n m a b c d e f }, which means
// at least n and at most m of a, b, c, d, e, f is true
//
// weight atom is of the form:
// [n m a =w_a b =w_b c =w_c], which means
// the sum of weights of "true" atoms should be between
// n and m.
bool CAssignment::parseAtoms(const char * line, CRule & rule)
{
	int i = 0;
	int length = strlen(line);
	bool isBody = true;
	bool isEmptyRule = true;
	int isMaxMin = 0;
	
	while ( i < length )
	{
		while ( i < length &&
			isspace(line[i]) )
			i++;
		if ( i >= length )
			break;
		
		if ( line[i] != 'm' &&
		     line[i] != '{' && line[i] !=',' && line[i] != '[' )
		{
			int id = getInt(line, i);
			
			if ( id != NONE )
			{
				int pos = hasTheAtom(id);
				if ( pos == NONE )
				{
					CAtom * pNewAtom = new CAtom;
					if ( pNewAtom == NULL )
						exit(2);
					pNewAtom->setIndex(id);
					pos = addAtom(*pNewAtom);
				}
				m_pAtoms[pos]->addRule(rule);
				if ( isBody )
					rule.addBodyAtom(*m_pAtoms[pos]);
				else
					rule.addHeadAtom(*m_pAtoms[pos]);
				isEmptyRule = false;
			}
			else
				i++;
		}
		else if ( line[i] == 'm' ) //here!!!
		{
			if ( strstr(line, "max") != NULL )
				isMaxMin = 1;
			else if ( strstr(line, "min") != NULL )
				isMaxMin = 2;
			i += 4; // skip "max:" or "min:"
			isBody = false;
		}
		else if ( line[i] == '{' )
		{
			i++;

			CCAtom * pNewCAtom = new CCAtom;
			if ( pNewCAtom == NULL )
				exit(2);
			
			addCAtom(*pNewCAtom);
			
			int lb = getInt(line, i);
			if ( lb != NONE )
				pNewCAtom->setLowerBound(lb);
			int ub = getInt(line, i);
			if ( ub != NONE )
				pNewCAtom->setUpperBound(ub);
			
			pNewCAtom->setIndex(m_iCAtomLength);
			pNewCAtom->addRule(rule);
			
			while ( i < length && line[i] != '}' )
			{
				int id = getInt(line, i);
				int pos;
				
				if ( id != NONE )
					pos = hasTheAtom(id);
				else
				{
					while ( i < length &&
						isspace(line[i]) )
						i++;
					continue;
				}
				
				if ( pos == NONE )
				{
					CAtom * pNewAtom = new CAtom;
					if ( pNewAtom == NULL )
						exit(2);
					pNewAtom->setIndex(id);
					
					pos = addAtom(*pNewAtom);
				}
				if ( !m_pAtoms[pos]->hasTheRule(rule)
				     )
					m_pAtoms[pos]->addRule(rule);
				
				m_pAtoms[pos]->addCAtom(*pNewCAtom);
				pNewCAtom->addAtom(*m_pAtoms[pos],
						   lb, ub);
			}

			if ( isBody )
				rule.addBodyAtom(*pNewCAtom);
			else
				rule.addHeadAtom(*pNewCAtom);

			if ( i >= length )
				return false;

			isEmptyRule = false;

			if ( isMaxMin != 0 )
			{
				pNewCAtom->setMaxMin(isMaxMin);
				m_iMaxMin = isMaxMin;
				m_pMaxMinCAtom = pNewCAtom;
			}

			i++;
		}
		else if ( line[i] == '[' )
		{
			i++;

			CWeightAtom * pNewWeightAtom = new CWeightAtom;
			if ( pNewWeightAtom == NULL )
				exit(2);

			addCAtom(*pNewWeightAtom);

			int lb = getInt(line, i);
			if ( lb != NONE )
				pNewWeightAtom->setLowerBound(lb);
			int ub = getInt(line, i);
			if ( ub != NONE )
				pNewWeightAtom->setUpperBound(ub);

			pNewWeightAtom->setIndex(m_iCAtomLength);
			pNewWeightAtom->addRule(rule);

			while ( i < length && line[i] != ']' )
			{
				int id = getInt(line, i);
				int pos;

				if ( id != NONE )
					pos = hasTheAtom(id);
				else
				{
					while ( i < length &&
						isspace(line[i]) )
						i++;
					continue;
				}

				if ( pos == NONE )
				{
					CAtom * pNewAtom = new CAtom;
					if ( pNewAtom == NULL )
						exit(2);
					pNewAtom->setIndex(id);
					
					pos = addAtom(*pNewAtom);
				}
				if ( !m_pAtoms[pos]->hasTheRule(rule) )
					m_pAtoms[pos]->addRule(rule);

				m_pAtoms[pos]->addCAtom(*pNewWeightAtom);

				while ( i < length &&
					( isspace(line[i]) ||
					  line[i] == '=' ) )
					i++;
				
				int weight = getInt(line, i);
				pNewWeightAtom->addAtom(*m_pAtoms[pos], weight,
							lb, ub);
			}
			
			if ( isBody )
				rule.addBodyAtom(*pNewWeightAtom);
			else
				rule.addHeadAtom(*pNewWeightAtom);

			if ( i >= length )
				return false;

			isEmptyRule = false;

			if ( isMaxMin != 0 )
			{
				pNewWeightAtom->setMaxMin(isMaxMin);
				m_iMaxMin = isMaxMin;
				m_pMaxMinCAtom = pNewWeightAtom;
			}

			i++;
		}
		else if ( line[i] == ',' )
		{
			isBody = false;
			i++;
		}
	}

	if ( !isEmptyRule )
	{
		rule.checkUnitRule();
		rule.checkCAtomRule();
	}
	else
	{		
		glHasEmptyRules = true;
		return false;
	}

	if ( isMaxMin != 0 )
		return true;
	else
		return !rule.checkTrivialRule();
}

// the difference b/t grounded ASPPS form and Dimacs (or cnf) form
// is that in Dimacs form, negative atoms are expressed by negative
// integer indeces and there are no cardinality atoms in Dimacs form.

// Bug: if the input clause has a and not a at the same time! (fixed)
bool CAssignment::parseAtomsInDimacsForm(const char * line, CRule & rule)
{
	int i = 0;
	int length = strlen(line);
	bool isBody;
	bool isEmptyRule = true;
	bool isBlankLine = true;
	
	while ( i < length && line[i] != '0' )
	{
		while ( i < length &&
			isspace(line[i]) )
			i++;
		if ( i >= length )
			break;
		
		if ( line[i] != '0' )
		{
			isBlankLine = false;
			int id = getInt(line, i);
			if ( id != NONE )
			{
				if ( id < 0 )
				{
					isBody = true;
					id = 0 - id;
				}
				else
					isBody = false;

				int pos = hasTheAtom(id);
				if ( pos == NONE )
				{
					CAtom * pNewAtom = new CAtom;
					if ( pNewAtom == NULL )
						exit(2);
					pNewAtom->setIndex(id);
					pos = addAtom(*pNewAtom);
				}
				m_pAtoms[pos]->addRule(rule);
				if ( isBody )
					rule.addBodyAtom(*m_pAtoms[pos]);
				else
					rule.addHeadAtom(*m_pAtoms[pos]);
				isEmptyRule = false;
			}
			else
				i++;
		}
	}
	
	if ( !isBlankLine && isEmptyRule )
	{
		glHasEmptyRules = true;
		return false;
	}
	
	return !rule.checkTrivialRule();
}

CAtom * CAssignment::walk(const CRule & rule, const CAtom * theAtom)
{
	CAtom ** allAtoms = rule.getAtoms();
	int numOfAtomsInRule = rule.getAtomLength();

	CAtom * result = NULL;
	if ( numOfAtomsInRule == 0 )
	{
		return result;
	}

	int tmprnd = 0;
	for ( int i = 0; i < 10; i++ )
	{
		tmprnd = myRand(numOfAtomsInRule);
		if ( !allAtoms[tmprnd]->forced() )
			break;
	}

	int rnd = -1;

	if ( theAtom == NULL )
		rnd = tmprnd;
	else
	{
		int i = tmprnd-1;
		int allTheSame = 0;
		while ( allTheSame < numOfAtomsInRule )
		{
			i = (i+1) % numOfAtomsInRule;
			// same reason as the one in findOppositeAtom function
			if ( !glGeneralDoubleFlip &&
			     allAtoms[i]->getValue() == theAtom->getValue() )
			{
				allTheSame++;
				continue;
			}
			else if ( glGeneralDoubleFlip &&
				  allAtoms[i]->getValue() !=
				  theAtom->getValue() )
			{
				allTheSame++;
				continue;
			}
			if ( allAtoms[i]->forced() )
				continue;
			
			rnd = i;
			break;
		}
		if ( rnd == -1 )
		{
			return NULL;
		}
	}
	
	if ( glDisjointCatoms && glPureDoubleFlip )
	{
		m_pAuxAtoms[0][0] = findOppositeAtom_SKC(*allAtoms[rnd], 0);
		if ( m_pAuxAtoms[0][0] == NULL )
		{
			m_iAuxAtomNum[0] = 0;
		}
	        else
		{
			m_iAuxAtomNum[0] = 1;
		}
	}
	else if ( glDisjointCatoms && !glPureDoubleFlip )
	{
		m_iAuxAtomNum[0] = -1;
	}
	
	if ( glPermutationCatoms )
	{
		if ( !getTheFixed(*allAtoms[rnd],
				  m_pAuxAtoms[0][0],
				  m_pAuxAtoms[0][1],
				  m_pAuxAtoms[0][2]) )
		{
			m_iAuxAtomNum[0] = 0;
		}
		else
		{
			m_iAuxAtomNum[0] = 3;
		}
	}
	m_iChosenAtomIndex = 0;
	return allAtoms[rnd];
}

CAtom * CAssignment::heuristic(const CRule & rule,
			       const CAtom * theAtom)
{
	if ( !glDisjointCatoms && !glGeneralDoubleFlip
	     && !glPermutationCatoms )
	{
		if ( glHeuristics == HEU_SKC )
		{
			return heuristicSKC_VBC(rule, theAtom);
		}
		else if ( glHeuristics == HEU_RNOVELTY_PLUS )
		{
			return heuristicRNoveltyPlus_VC(rule, theAtom);
		}
		else
		{
			cerr<<"Error: cannot determine which heuristic function to use.\n";
			return NULL;
		}
	}
	else if ( glDisjointCatoms || glGeneralDoubleFlip )
	{
		if ( glHeuristics == HEU_SKC )
		{
			return heuristicSKC_DBF(rule, theAtom);
		}
		else if ( glHeuristics == HEU_RNOVELTY_PLUS )
		{
			return heuristicRNoveltyPlus_DBF(rule, theAtom);
		}
		else
		{
			cerr<<"Error: cannot determine which heuristic function to use.\n";
			return NULL;
		}
	}
	else if ( glPermutationCatoms )
	{
		return heuristicSKC_PF(rule, theAtom);
	}
	else
	{
		cerr<<"Error: cannot determine which heuristic function to use.\n";
		return NULL;
	}
}

CAtom * CAssignment::heuristicSKC_VBC(const CRule & rule,
				      const CAtom * theAtom)
{
	CAtom ** allAtoms = rule.getAtoms();
	int numOfAtomsInRule = rule.getAtomLength();
	
	CAtom * result = NULL;
	if ( numOfAtomsInRule == 0 )
	{
		return result;
	}

	int weight = OVER_RATE * MAX_INT + 1;
	int bestAtoms[HEU_ATOM_NUM], numOfBestAtoms = 0;
	int offset = myRand(numOfAtomsInRule);
	
	for ( int j = 0;
	      j < numOfAtomsInRule && numOfBestAtoms < HEU_ATOM_NUM;
	      j++ )
	{
		int i = (j+offset) % numOfAtomsInRule;
		if ( //rule.noHelp(allAtoms[i]) ||
		     allAtoms[i] == theAtom || allAtoms[i]->forced() ||
		     (glUseTimeStamp && allAtoms[i]->getTimeStamp() >=
		      glLastTimeStamp - glTabuLength) )
			continue;
		
		int n = allAtoms[i]->getBKCT();
		
		if ( n <= weight )
		{
			if ( n < weight )
			{
				numOfBestAtoms = 0;
				weight = n;
			}
			
			bestAtoms[numOfBestAtoms] = i;
			numOfBestAtoms++;
		}
	}

	int tmpRandom;

	if ( (weight > 0 && numOfBestAtoms > 0 &&
	      (tmpRandom  = myRand(glWalkProbabilityDenominator))
	       < glWalkProbabilityNumerator) ||
	     numOfBestAtoms <= 0 )
	{
		return walk(rule);
	}
	else
	{
		if ( numOfBestAtoms <= 0 )
		{
			int rnd = myRand(numOfAtomsInRule);
			
			m_iChosenAtomIndex = 0;
			return allAtoms[rnd];
		}
		else if ( numOfBestAtoms == 1 )
		{
			if ( weight >= OVER_RATE * MAX_INT )
				glAllBestAtomsOverflow++;
			
			m_iChosenAtomIndex = 0;
			return allAtoms[bestAtoms[0]];
		}
		else
		{
			if ( weight >= OVER_RATE * MAX_INT )
				glAllBestAtomsOverflow++;

			m_iChosenAtomIndex = myRand(numOfBestAtoms);
					
			return allAtoms[bestAtoms[m_iChosenAtomIndex]];
		}
	}
}

CAtom * CAssignment::heuristicSKC_DBF(const CRule & rule,
				      const CAtom * theAtom)
{
	CAtom ** allAtoms = rule.getAtoms();
	int numOfAtomsInRule = rule.getAtomLength();
	
	CAtom * result = NULL;
	if ( numOfAtomsInRule == 0 )
	{
		return result;
	}

	int weight = OVER_RATE * MAX_INT + 1;
	int bestAtoms[HEU_ATOM_NUM], numOfBestAtoms = 0;
	int offset = myRand(numOfAtomsInRule);
	
	for ( int j = 0;
	      j < numOfAtomsInRule && numOfBestAtoms < HEU_ATOM_NUM;
	      j++ )
	{
		int i = (j+offset) % numOfAtomsInRule;
		if ( allAtoms[i] == theAtom || allAtoms[i]->forced() )
			continue;

		if ( glDisjointCatoms &&
		     allAtoms[i]->getUnitCAtom(0) != NULL &&
		     allAtoms[i]->getUnitCAtom(0)->getType() ==
		     CAtom::C_ATOM &&
		     !allAtoms[i]->getUnitCAtom(0)->SAT() )
		{
			CCAtom * pcatom = allAtoms[i]->getUnitCAtom(0);
			bool isPos = pcatom->isPositive();
			int lb = pcatom->getLowerBound();
			int ub = pcatom->getUpperBound();
			int sn = pcatom->getNumOfSatisfied();

			if ( isPos )
			{
				if ( sn > ub && !allAtoms[i]->getValue() )
				{
					continue;
				}
				if ( sn < lb && allAtoms[i]->getValue() )
				{
					continue;
				}
			}
			else
			{
				if ( sn > ub && allAtoms[i]->getValue() )
				{
					continue;
				}
				if ( sn < lb && !allAtoms[i]->getValue() )
				{
					continue;
				}
			}
		}
		
		int n, index;
		CAtom * theOppositeAtom = NULL;

		if ( glDisjointCatoms &&
		     allAtoms[i]->getUnitCAtom(0) != NULL &&
		     allAtoms[i]->getUnitCAtom(0)->getType() ==
		     CAtom::C_ATOM &&
		     allAtoms[i]->getUnitCAtom(0)->SAT() )
		{
			if ( !glUseVirtualCounts )
				n = allAtoms[i]->getRealBKCT();
			else
				n = allAtoms[i]->getBKCT();
			
			// if the c-atoms are disjoint, then each atom should
			// appear in at most one c-atom.  so we call the
			// heuristic function with the index to the only
			// c-atom it might appear in.
		        theOppositeAtom = findOppositeAtom_SKC(*allAtoms[i],
							       0);
			if ( theOppositeAtom != NULL &&
			     glComboWeight )
			{
				if ( !glUseVirtualCounts )
					n += theOppositeAtom->getRealBKCT();
				else
					n += theOppositeAtom->getBKCT();
			}
		}
		else if ( glDisjointCatoms )
		{
			if ( !glUseVirtualCounts )
				n = allAtoms[i]->getRealBKCT();
			else
				n = allAtoms[i]->getBKCT();

			theOppositeAtom = NULL;
		}
		else
		{
			// by default, it is glGeneralDoubleFlip
			n = allAtoms[i]->getBKCT();
		}
		
		if ( n <= weight )
		{
			if ( n < weight )
			{
				numOfBestAtoms = 0;
				weight = n;
			}
			
			bestAtoms[numOfBestAtoms] = i;
			index = numOfBestAtoms;
			numOfBestAtoms++;
			
			if ( glDisjointCatoms )
			{
				if ( theOppositeAtom == NULL )
				{
					m_iAuxAtomNum[index] = 0;
				}
				else
				{
					m_iAuxAtomNum[index] = 1;
					m_pAuxAtoms[index][0] =
						theOppositeAtom;
				}
			}
		}
	}

	if ( (weight > 0 && numOfBestAtoms > 0 &&
	      myRand(glWalkProbabilityDenominator)
	      < glWalkProbabilityNumerator) ||
	     numOfBestAtoms <= 0 )
	{
		return walk(rule);
	}
	else
	{
		if ( numOfBestAtoms <= 0 )
		{
			int rnd = myRand(numOfAtomsInRule);
			
			if ( glDisjointCatoms &&
			     allAtoms[rnd]->getUnitCAtom(0)->getType() ==
			     CAtom::C_ATOM )
			{
				m_pAuxAtoms[0][0] = findOppositeAtom_SKC
					(*allAtoms[rnd], 0);
				if ( m_pAuxAtoms[0][0] == NULL )
				{
					m_iAuxAtomNum[0] = 0;
				}
				else
				{
					m_iAuxAtomNum[0] = 1;
				}
			}

			m_iChosenAtomIndex = 0;
			return allAtoms[rnd];
		}
		else if ( numOfBestAtoms == 1 )
		{
			if ( weight >= OVER_RATE * MAX_INT )
				glAllBestAtomsOverflow++;
			
			m_iChosenAtomIndex = 0;
			return allAtoms[bestAtoms[0]];
		}
		else
		{
			if ( weight >= OVER_RATE * MAX_INT )
				glAllBestAtomsOverflow++;

			m_iChosenAtomIndex = myRand(numOfBestAtoms);
			return allAtoms[bestAtoms[m_iChosenAtomIndex]];
		}
	}
}

CAtom * CAssignment::heuristicSKC_PF(const CRule & rule,
				     const CAtom * theAtom)
{
	CAtom ** allAtoms = rule.getAtoms();
	int numOfAtomsInRule = rule.getAtomLength();
	
	CAtom * result = NULL;
	if ( numOfAtomsInRule == 0 )
	{
		return result;
	}

	int weight = OVER_RATE * MAX_INT + 1;
	int bestAtoms[HEU_ATOM_NUM], numOfBestAtoms = 0;
	int offset = myRand(numOfAtomsInRule);
	
	for ( int j = 0;
	      j < numOfAtomsInRule && numOfBestAtoms < HEU_ATOM_NUM;
	      j++ )
	{
		int i = (j+offset) % numOfAtomsInRule;
		if ( allAtoms[i] == theAtom || allAtoms[i]->forced() )
			continue;
		
		int n, index;
		CAtom * theColAtom;
		CAtom * theDiagAtom;
		CAtom * theRowAtom;
		bool permGetTheFixedSucceeded = false;
		
		if ( !glUseVirtualCounts )
			n = allAtoms[i]->getRealBKCT();
		else
			n = allAtoms[i]->getBKCT(false);
		
		permGetTheFixedSucceeded = getTheFixed(*allAtoms[i],
						       theColAtom,
						       theRowAtom,
						       theDiagAtom);
		if ( permGetTheFixedSucceeded &&
		     glComboWeight )
		{
			if ( !glUseVirtualCounts )
			{
				n += theColAtom->getRealBKCT();
				n += theRowAtom->getRealBKCT();
				n += theDiagAtom->getRealBKCT();
			}
			else
			{
				n += theColAtom->getBKCT(false);
				n += theRowAtom->getBKCT(false);
				n += theDiagAtom->getBKCT(false);
			}
		}
		
		if ( n <= weight )
		{
			if ( n < weight )
			{
				numOfBestAtoms = 0;
				weight = n;
			}
			
			bestAtoms[numOfBestAtoms] = i;
			index = numOfBestAtoms;
			numOfBestAtoms++;
			
			if ( !permGetTheFixedSucceeded )
			{
				m_iAuxAtomNum[index] = 0;
			}
			else
			{
				m_iAuxAtomNum[index] = 3;
				m_pAuxAtoms[index][0] = theColAtom;
				m_pAuxAtoms[index][1] = theRowAtom;
				m_pAuxAtoms[index][2] = theDiagAtom;
			}
		}
	}

	if ( (weight > 0 && numOfBestAtoms > 0 &&
	      myRand(glWalkProbabilityDenominator)
	      < glWalkProbabilityNumerator) ||
	     numOfBestAtoms <= 0 )
	{
		return walk(rule);
	}
	else
	{
		if ( numOfBestAtoms <= 0 )
		{
			int rnd = myRand(numOfAtomsInRule);
			
			if ( !getTheFixed(*allAtoms[rnd],
					  m_pAuxAtoms[0][0],
					  m_pAuxAtoms[0][1],
					  m_pAuxAtoms[0][2]) )
			{
				m_iAuxAtomNum[0] = 0;
			}
			else
			{
				m_iAuxAtomNum[0] = 3;
			}

			m_iChosenAtomIndex = 0;
			return allAtoms[rnd];
		}
		else if ( numOfBestAtoms == 1 )
		{
			if ( weight >= OVER_RATE * MAX_INT )
				glAllBestAtomsOverflow++;
			
			m_iChosenAtomIndex = 0;
			return allAtoms[bestAtoms[0]];
		}
		else
		{
			if ( weight >= OVER_RATE * MAX_INT )
				glAllBestAtomsOverflow++;

			m_iChosenAtomIndex = myRand(numOfBestAtoms);
			return allAtoms[bestAtoms[m_iChosenAtomIndex]];
		}
	}
}

CAtom * CAssignment::heuristicRNoveltyPlus_VC(const CRule & rule,
					      const CAtom * theAtom)
{
	CAtom ** allAtoms = rule.getAtoms();
	int numOfAtomsInRule = rule.getAtomLength();
	if ( myRand(glRNP_WP_DENOMINATOR) < glRNP_WP_NUMERATOR )
		return walk(rule);
	
	CAtom * result = NULL;
	if ( numOfAtomsInRule == 0 )
	{
		return result;
	}

	int bestWeight = OVER_RATE * MAX_INT + 1;
	int secondBestWeight = OVER_RATE * MAX_INT + 1;
	int bestAtoms[HEU_ATOM_NUM];
	int secondBestAtoms[HEU_ATOM_NUM];
	int numOfBestAtoms = 0;
	int numOfSecondBestAtoms = 0;
	int offset = myRand(numOfAtomsInRule);
	
	for ( int j = 0;
	      j < numOfAtomsInRule &&
		      numOfBestAtoms < HEU_ATOM_NUM;
	      j++ )
	{
		int i = (j+offset) % numOfAtomsInRule;
		if (//rule.noHelp(allAtoms[i]) ||
		    allAtoms[i] == theAtom || allAtoms[i]->forced() )
			continue;
		
		int n;

		// the less variable n is, the better the atom is
		n = allAtoms[i]->getBKCT() - allAtoms[i]->getMKCT();
		
		if ( n <= bestWeight )
		{
			if ( n < bestWeight )
			{
				memcpy(secondBestAtoms, bestAtoms,
				       HEU_ATOM_NUM*sizeof(int));
				numOfSecondBestAtoms = numOfBestAtoms;
				secondBestWeight = bestWeight;
				
				numOfBestAtoms = 0;
				bestWeight = n;
			}
			
			bestAtoms[numOfBestAtoms] = i;
			numOfBestAtoms++;
		}
		else if ( n <= secondBestWeight )
		{
			if ( n < secondBestWeight )
			{
				numOfSecondBestAtoms = 0;
				secondBestWeight = n;
			}
			if ( numOfSecondBestAtoms < HEU_ATOM_NUM )
			{
				secondBestAtoms[numOfSecondBestAtoms] = i;
				numOfSecondBestAtoms++;
			}
		}
	}

	if ( numOfBestAtoms > 0 )
	{
		int offset = myRand(numOfBestAtoms);
		int largestTS = rule.getLargestTimeStamp();
		
		for ( int i = 0; i < numOfBestAtoms; i++ )
		{
			int j = (i+offset) % numOfBestAtoms;
			if ( allAtoms[bestAtoms[j]]->getTimeStamp() <
			     largestTS - glTabuLength )
			     return allAtoms[bestAtoms[j]];
		}
	}
	else // in this case, numOfSecondBestAtoms == 0 too
		return allAtoms[myRand(numOfAtomsInRule)];

	if ( numOfSecondBestAtoms == 0 )
	{
		return allAtoms[myRand(numOfAtomsInRule)];
	}

	// from this point on, numOfBestAtoms > 0 and numOfSecondBestAtoms > 0
	/* from walksat
	int diff = secondBestWeight - bestWeight;
	if ( glWalkProbabilityNumerator * 2 <
	     glWalkProbabilityDenominator && diff > 1 )
		return allAtoms[bestAtoms[myRand(numOfBestAtoms)]];
	if ( glWalkProbabilityNumerator * 2 <
	     glWalkProbabilityDenominator && diff == 1 )
	{
		if ( myRand(glWalkProbabilityDenominator) <
		     2 * glWalkProbabilityNumerator )
			return allAtoms[secondBestAtoms
						[myRand
						 (numOfSecondBestAtoms)]];
		else
			return allAtoms[bestAtoms[myRand(numOfBestAtoms)]];
	}
	if ( diff == 1 )
		return allAtoms[secondBestAtoms[myRand(numOfSecondBestAtoms)]];
	if ( myRand(glWalkProbabilityDenominator) <
	     2*(glWalkProbabilityNumerator-(glWalkProbabilityDenominator/2)) )
		return allAtoms[secondBestAtoms[myRand(numOfSecondBestAtoms)]];

	return allAtoms[bestAtoms[myRand(numOfBestAtoms)]];
	*/
	
	if ( secondBestWeight - bestWeight > 1 )
	{
		int p1Denominator = glWalkProbabilityDenominator;
		int p1Numerator = 2*p1Denominator -
			2*glWalkProbabilityNumerator;
		if ( p1Numerator > p1Denominator )
			p1Numerator = p1Denominator;

		if ( myRand(p1Denominator) < p1Numerator )
			return allAtoms[bestAtoms[myRand(numOfBestAtoms)]];
		else
			return allAtoms[secondBestAtoms
					[myRand(numOfSecondBestAtoms)]];
	}
	else
	{
		
		int p2Denominator = glWalkProbabilityDenominator;
		int p2Numerator = p2Denominator - 2*glWalkProbabilityNumerator;
		if ( p2Numerator < 0 )
			p2Numerator = 0;

		if ( myRand(p2Denominator) < p2Numerator )
			return allAtoms[bestAtoms[myRand(numOfBestAtoms)]];
		else
			return allAtoms[secondBestAtoms
					[myRand(numOfSecondBestAtoms)]];
	}
}

CAtom * CAssignment::heuristicRNoveltyPlus_DBF(const CRule & rule,
					       const CAtom * theAtom)
{
	CAtom ** allAtoms = rule.getAtoms();
	int numOfAtomsInRule = rule.getAtomLength();
	if ( myRand(glRNP_WP_DENOMINATOR) < glRNP_WP_NUMERATOR )
		return walk(rule);
	
	CAtom * result = NULL;
	if ( numOfAtomsInRule == 0 )
	{
		return result;
	}

	int bestWeight = OVER_RATE * MAX_INT + 1;
	int secondBestWeight = OVER_RATE * MAX_INT + 1;
	int bestAtom = -1;
	int secondBestAtom = -1;
	int offset = myRand(numOfAtomsInRule);
	
	for ( int j = 0; j < numOfAtomsInRule; j++ )
	{
		int i = (j+offset) % numOfAtomsInRule;
		if ( allAtoms[i] == theAtom || allAtoms[i]->forced() )
			continue;

		if ( glDisjointCatoms &&
		     allAtoms[i]->getUnitCAtom(0) != NULL &&
		     allAtoms[i]->getUnitCAtom(0)->getType() ==
		     CAtom::C_ATOM &&
		     !allAtoms[i]->getUnitCAtom(0)->SAT() )
		{
			CCAtom * pcatom = allAtoms[i]->getUnitCAtom(0);
			bool isPos = pcatom->isPositive();
			int lb = pcatom->getLowerBound();
			int ub = pcatom->getUpperBound();
			int sn = pcatom->getNumOfSatisfied();

			if ( isPos )
			{
				if ( sn > ub && !allAtoms[i]->getValue() )
				{
					continue;
				}
				if ( sn < lb && allAtoms[i]->getValue() )
				{
					continue;
				}
			}
			else
			{
				if ( sn > ub && allAtoms[i]->getValue() )
				{
					continue;
				}
				if ( sn < lb && !allAtoms[i]->getValue() )
				{
					continue;
				}
			}
		}
		
		int n;
		CAtom * theOppositeAtom = NULL;

		// by default, it is glGeneralDoubleFlip
		n = allAtoms[i]->getBKCT() - allAtoms[i]->getMKCT();
		if ( glDisjointCatoms &&
		     allAtoms[i]->getUnitCAtom(0) != NULL &&
		     allAtoms[i]->getUnitCAtom(0)->getType() ==
		     CAtom::C_ATOM &&
		     allAtoms[i]->getUnitCAtom(0)->SAT() )
		{
			if ( !glUseVirtualCounts )
				n = allAtoms[i]->getRealBKCT() -
					allAtoms[i]->getRealMKCT();
			else
				n = allAtoms[i]->getBKCT() -
					allAtoms[i]->getMKCT();
			
			// if the c-atoms are disjoint, then each atom should
			// appear in at most one c-atom.  so we call the
			// heuristic function with the index to the only
			// c-atom it might appear in.
		        theOppositeAtom = findOppositeAtom_RNoveltyPlus
				(*allAtoms[i], 0);
			if ( theOppositeAtom != NULL &&
			     glComboWeight )
			{
				if ( !glUseVirtualCounts )
					n += theOppositeAtom->getRealBKCT();
				else
					n += theOppositeAtom->getBKCT();
			}
		}
		else if ( glDisjointCatoms )
		{
			if ( !glUseVirtualCounts )
				n = allAtoms[i]->getRealBKCT() -
					allAtoms[i]->getRealMKCT();
			else
				n = allAtoms[i]->getBKCT() -
					allAtoms[i]->getMKCT();

			theOppositeAtom = NULL;
		}
		
		if ( n < bestWeight )
		{
			secondBestAtom = bestAtom;
			secondBestWeight = bestWeight;
			bestAtom = i;
			
			if ( glDisjointCatoms )
			{
				m_iAuxAtomNum[1] = m_iAuxAtomNum[0];
				if ( m_iAuxAtomNum[0] == 1 )
				{
					m_pAuxAtoms[1][0] = m_pAuxAtoms[0][0];
				}			
				if ( theOppositeAtom == NULL )
				{
					m_iAuxAtomNum[0] = 0;
				}
				else
				{
					m_iAuxAtomNum[0] = 1;
					m_pAuxAtoms[0][0] =
						theOppositeAtom;
				}
			}
		}
		else if ( n < secondBestWeight )
		{
			secondBestWeight = n;
			secondBestAtom = i;
			if ( glDisjointCatoms )
			{
				if ( theOppositeAtom == NULL )
				{
					m_iAuxAtomNum[1] = 0;
				}
				else
				{
					m_iAuxAtomNum[1] = 1;
					m_pAuxAtoms[1][0] =
						theOppositeAtom;
				}
			}			
		}
	}

	if ( bestAtom > 0 )
	{
		int largestTS = rule.getLargestTimeStamp();
		
		if ( allAtoms[bestAtom]->getTimeStamp() <
		     largestTS - glTabuLength )
		{
			m_iChosenAtomIndex = 0;
			return allAtoms[bestAtom];
		}
	}
	else // in this case, numOfSecondBestAtoms == 0 too
		return walk(rule);

	if ( secondBestAtom < 0 )
	{
		return walk(rule);
	}

	// from this point on, numOfBestAtoms > 0 and numOfSecondBestAtoms > 0
	/* from walksat
	int diff = secondBestWeight - bestWeight;
	if ( glWalkProbabilityNumerator * 2 <
	     glWalkProbabilityDenominator && diff > 1 )
		return allAtoms[bestAtoms[myRand(numOfBestAtoms)]];
	if ( glWalkProbabilityNumerator * 2 <
	     glWalkProbabilityDenominator && diff == 1 )
	{
		if ( myRand(glWalkProbabilityDenominator) <
		     2 * glWalkProbabilityNumerator )
			return allAtoms[secondBestAtoms
						[myRand
						 (numOfSecondBestAtoms)]];
		else
			return allAtoms[bestAtoms[myRand(numOfBestAtoms)]];
	}
	if ( diff == 1 )
		return allAtoms[secondBestAtoms[myRand(numOfSecondBestAtoms)]];
	if ( myRand(glWalkProbabilityDenominator) <
	     2*(glWalkProbabilityNumerator-(glWalkProbabilityDenominator/2)) )
		return allAtoms[secondBestAtoms[myRand(numOfSecondBestAtoms)]];

	return allAtoms[bestAtoms[myRand(numOfBestAtoms)]];
	*/
	
	if ( secondBestWeight - bestWeight > 1 )
	{
		int p1Denominator = glWalkProbabilityDenominator;
		int p1Numerator = 2*p1Denominator -
			2*glWalkProbabilityNumerator;
		if ( p1Numerator > p1Denominator )
			p1Numerator = p1Denominator;

		if ( myRand(p1Denominator) < p1Numerator )
		{
			m_iChosenAtomIndex = 0;
			return allAtoms[bestAtom];
		}
		else
		{
			m_iChosenAtomIndex = 1;
			return allAtoms[secondBestAtom];
		}
	}
	else
	{
		
		int p2Denominator = glWalkProbabilityDenominator;
		int p2Numerator = p2Denominator - 2*glWalkProbabilityNumerator;
		if ( p2Numerator < 0 )
			p2Numerator = 0;

		if ( myRand(p2Denominator) < p2Numerator )
		{
			m_iChosenAtomIndex = 0;
			return allAtoms[bestAtom];
		}
		else
		{
			m_iChosenAtomIndex = 1;
			return allAtoms[secondBestAtom];
		}
	}
}

CAtom * CAssignment::findOppositeAtom_SKC(const CAtom & atom, int which)
{
	const CCAtom * pCAtom = (glDisjointCatoms ?
				 atom.getUnitCAtom(which) :
				 atom.getCAtom(which));

	if ( pCAtom == NULL )
	{
		return NULL;
	}

	if ( !pCAtom->SAT() && !glGeneralDoubleFlip )
	{
		return NULL;
	}
	
	CAtom ** allAtoms = pCAtom->getAtoms();
	int numOfAtomsInCAtom = pCAtom->getAtomLength();
	int rnd = -1;
	
	if ( numOfAtomsInCAtom == 0 )
	{
		return NULL;
	}

	int weight = OVER_RATE * MAX_INT + 1;
	int bestAtoms[HEU_ATOM_NUM], numOfBestAtoms = 0;
	int offset = myRand(numOfAtomsInCAtom);
	
	for ( int j = 0;
	      j < numOfAtomsInCAtom && numOfBestAtoms < HEU_ATOM_NUM;
	      j++ )
	{
		int i = (j+offset) % numOfAtomsInCAtom;

		if ( allAtoms[i]->forced() )
			continue;
		
		if ( !glGeneralDoubleFlip &&
		     allAtoms[i]->getValue() == atom.getValue() )
			continue;
		// with glGeneralDoubleFlip set, the first chosen atom
		// will be flipped first.  So when we want to choose
		// an opposite atom, we need to select from the atoms
		// that have values the same as its current value
		else if ( glGeneralDoubleFlip &&
			  allAtoms[i]->getValue() != atom.getValue() )
			continue;

		if ( rnd == -1 )
			rnd = i;
		
		int n;

		if ( !glUseVirtualCounts )
			n = allAtoms[i]->getRealBKCT();
		else
			n = allAtoms[i]->getBKCT();
		
		if ( n <= weight )
		{
			if ( n < weight )
			{
				numOfBestAtoms = 0;
				weight = n;
			}
			
			bestAtoms[numOfBestAtoms++] = i;
		}
	}

	if ( numOfBestAtoms <= 0 )
	{
		if ( rnd != -1 )
			return allAtoms[rnd];
		else
			return NULL;
	}
	else
	{
		if ( numOfBestAtoms == 1 )
		{
			if ( weight >= OVER_RATE * MAX_INT )
				glAllBestAtomsOverflow++;

			return allAtoms[bestAtoms[0]];
		}
		else
		{
			if ( weight >= OVER_RATE * MAX_INT )
				glAllBestAtomsOverflow++;

			return allAtoms[bestAtoms[myRand(numOfBestAtoms)]];
		}
	}
}

CAtom * CAssignment::findOppositeAtom_RNoveltyPlus(const CAtom & atom,
						   int which)
{
	const CCAtom * pCAtom = (glDisjointCatoms ?
				 atom.getUnitCAtom(which) :
				 atom.getCAtom(which));

	if ( pCAtom == NULL )
	{
		return NULL;
	}

	if ( !pCAtom->SAT() && !glGeneralDoubleFlip )
	{
		return NULL;
	}
	
	CAtom ** allAtoms = pCAtom->getAtoms();
	int numOfAtomsInCAtom = pCAtom->getAtomLength();
	
	if ( numOfAtomsInCAtom == 0 )
	{
		return NULL;
	}

	int rnd = -1;
	
	int weight = OVER_RATE * MAX_INT + 1;
	int bestAtoms[HEU_ATOM_NUM], numOfBestAtoms = 0;
	int offset = myRand(numOfAtomsInCAtom);
	
	for ( int j = 0;
	      j < numOfAtomsInCAtom && numOfBestAtoms < HEU_ATOM_NUM;
	      j++ )
	{
		int i = (j+offset) % numOfAtomsInCAtom;

		if ( allAtoms[i]->forced() )
			continue;
		
		if ( !glGeneralDoubleFlip &&
		     allAtoms[i]->getValue() == atom.getValue() )
			continue;
		// with glGeneralDoubleFlip set, the first chosen atom
		// will be flipped first.  So when we want to choose
		// an opposite atom, we need to select from the atoms
		// that have values the same as its current value
		else if ( glGeneralDoubleFlip &&
			  allAtoms[i]->getValue() != atom.getValue() )
			continue;

		if ( rnd == -1 )
			rnd = i;
		
		int n;

		if ( !glUseVirtualCounts )
			n = allAtoms[i]->getRealBKCT() -
				allAtoms[i]->getRealMKCT();
		else
			n = allAtoms[i]->getBKCT() -
				allAtoms[i]->getMKCT();
		
		if ( n <= weight )
		{
			if ( n < weight )
			{
				numOfBestAtoms = 0;
				weight = n;
			}
			
			bestAtoms[numOfBestAtoms++] = i;
		}
	}

	if ( numOfBestAtoms <= 0 )
	{
		if ( rnd != -1 )
			return allAtoms[rnd];
		else
			return NULL;
	}
	else
	{
		if ( numOfBestAtoms == 1 )
		{
			if ( weight >= OVER_RATE * MAX_INT )
				glAllBestAtomsOverflow++;

			return allAtoms[bestAtoms[0]];
		}
		else
		{
			if ( weight >= OVER_RATE * MAX_INT )
				glAllBestAtomsOverflow++;

			return allAtoms[bestAtoms[myRand(numOfBestAtoms)]];
		}
	}
}

/*
Case #1:
  
  ---------|-------|-------|----------
 (main)  T -> F    |     F -> T (rowAtom: fixed)
   --------o-------|-------o----------   (pRowCAtom)
           |       |       |
   --------|-------|-------|----------
 (colA)  F -> T    |     T -> F (diagAtom: fixed)  
   --------o-------|-------o----------   (pCounterRowCAtom)
           |       |       |
   --------|-------|-------|----------
       (pColCAtom)      (pCounterColCAtom)

Case #2:
  
  ---------|-------|-------|----------
 (main)  F -> T    |     T -> F (rowAtom: fixed)
   --------o-------|-------o----------   (pRowCAtom)
           |       |       |
   --------|-------|-------|----------
 (colA:) T -> F    |     F -> T (diagAtom: fixed)
 (fixed)           |
   --------o-------|-------o----------   (pCounterRowCAtom)
           |       |       |
   --------|-------|-------|----------
       (pColCAtom)      (pCounterColCAtom)
       
*/
// given the main atom and the column atom, the other two atoms
// are fixed. this function returns the two fixed atoms.
bool CAssignment::getTheFixed(const CAtom & mainAtom,
			      CAtom * & pColAtom,
			      CAtom * & pRowAtom,
			      CAtom * & pDiagAtom)
{
	if ( mainAtom.getCAtom(0) == NULL ||
	     mainAtom.getCAtom(1) == NULL )
	{
		return false;
	}
	
	pColAtom = findOppositeAtom_SKC(mainAtom, 1);
	if ( pColAtom == NULL )
	{
		return false;
	}

	if ( mainAtom.getValue() ) // case #1
	{
		const CCAtom * pCounterRowCAtom;

		if ( pColAtom->getCAtom(0) == mainAtom.getCAtom(1) )
		{
			pDiagAtom = findOppositeAtom_SKC(*pColAtom, 1);
			pCounterRowCAtom = pColAtom->getCAtom(1);
		}
		else
		{	
			pDiagAtom = findOppositeAtom_SKC(*pColAtom, 0);
			pCounterRowCAtom = pColAtom->getCAtom(0);
		}
	
		if ( pDiagAtom == NULL )
			return false;

		const CCAtom * pCounterColCAtom;
		if ( pDiagAtom->getCAtom(0) == pCounterRowCAtom )
			pCounterColCAtom = pDiagAtom->getCAtom(1);
		else
			pCounterColCAtom = pDiagAtom->getCAtom(0);

		if ( pCounterColCAtom == NULL )
			return false;

		const CCAtom * pRowCAtom = mainAtom.getCAtom(0);

		CAtom ** pCounterColAtoms = pCounterColCAtom->getAtoms();
		int numOfAtomsInCCol = pCounterColCAtom->getAtomLength();
		for ( int i = 0; i < numOfAtomsInCCol; i++ )
		{
			if ( pCounterColAtoms[i]->getCAtom(0) == pRowCAtom ||
			     pCounterColAtoms[i]->getCAtom(1) == pRowCAtom )
			{
				pRowAtom = pCounterColAtoms[i];
				if ( pRowAtom->forced() ||
				     pColAtom->forced() ||
				     pDiagAtom->forced() )
					return false;
				else
					return true;
			}
		}
	}
	else // case #2
	{
		pRowAtom = findOppositeAtom_SKC(mainAtom, 0);
		if ( pRowAtom == NULL )
		{
			return false;
		}
		
		const CCAtom * pCounterColCAtom;
		if ( pRowAtom->getCAtom(0) == mainAtom.getCAtom(0) )
			pCounterColCAtom = pRowAtom->getCAtom(1);
		else
			pCounterColCAtom = pRowAtom->getCAtom(0);
		if ( pCounterColCAtom == NULL )
			return false;

		const CCAtom * pCounterRowCAtom;
		if ( pColAtom->getCAtom(0) == mainAtom.getCAtom(1) )
			pCounterRowCAtom = pColAtom->getCAtom(1);
		else
			pCounterRowCAtom = pColAtom->getCAtom(0);
		if ( pCounterRowCAtom == NULL )
			return false;
		
		CAtom ** pCounterColAtoms = pCounterColCAtom->getAtoms();
		int numOfAtomsInCCol = pCounterColCAtom->getAtomLength();
		for ( int i = 0; i < numOfAtomsInCCol; i++ )
		{
			if ( pCounterColAtoms[i]->getCAtom(0) ==
			     pCounterRowCAtom ||
			     pCounterColAtoms[i]->getCAtom(1) ==
			     pCounterRowCAtom )
			{
				pDiagAtom = pCounterColAtoms[i];
				if ( pRowAtom->forced() ||
				     pColAtom->forced() ||
				     pDiagAtom->forced() )
					return false;
				else
					return true;
			}
		}
	}
	
	return false;
}

CAtom * CAssignment::pickAnAtom(const CRuleList & ruleList)
{
#ifdef __LLN_DEBUG__
	int rnd;

	rnd = random();
	int chosenRuleID = rnd%(ruleList.getUNSATRuleLength());
	const CRule & theRule = ruleList.getUNSATRule(chosenRuleID);
	fprintf(pLog,
		"rand@choseRule = %d\tclause = %d\t# of unsat cli = %d\n",
		rnd, theRule.getOrigIndex(),
		ruleList.getUNSATRuleLength());
#else	
	int chosenRuleID = myRand(ruleList.getUNSATRuleLength());
	const CRule & theRule = ruleList.getUNSATRule(chosenRuleID);
	
	if ( theRule.isUnitRule() )
		glNumPickUnitRule++;
#endif    
	if ( theRule.SAT() )
		return NULL;

	CAtom * bestAtom = heuristic(theRule);

	return bestAtom;
}

CAtom * CAssignment::pickAnAtomFromSameRule(const CAtom * theAtom,
					    const CRule & theRule)
{
	if ( theRule.SAT() )
	{
		return NULL;
	}

	if ( theRule.isUnitRule() )
	{
		return findOppositeAtom_SKC(*theAtom, 0);
	}
	else
	{
		return heuristic(theRule, theAtom);
	}
}

bool CAssignment::flip(CAtom * theAtom, CRuleList & ruleList)
{
	if ( theAtom == NULL )
		return false;

	theAtom->negateSelf(&ruleList);
	if ( theAtom->getValue() )
		m_iTrueAtomNum++;
	else
		m_iTrueAtomNum--;
	
	return true;
}

bool CAssignment::doubleFlip(CAtom * theAtom, CRuleList & ruleList)
{
	if ( theAtom == NULL )
		return false;

	if ( theAtom->getType() == CAtom::C_ATOM ||
	     theAtom->getType() == CAtom::W_ATOM )
	{
		theAtom->negateSelf(&ruleList);

		return false;
	}

	if ( m_iAuxAtomNum[m_iChosenAtomIndex] != 1 )
	{
		theAtom->negateSelf(&ruleList);
		
		if ( theAtom->getValue() )
			m_iTrueAtomNum++;
		else
			m_iTrueAtomNum--;

		return false;
	}

	theAtom->negateSelf(&ruleList);
	m_pAuxAtoms[m_iChosenAtomIndex][0]->negateSelf(&ruleList);
	
	m_iDoubleFlipNum++;
	
	return true;
}

bool CAssignment::multiFlip(CAtom * theAtom, CRuleList & ruleList)
{
	if ( theAtom == NULL )
		return false;

	if ( theAtom->getType() == CAtom::C_ATOM ||
	     theAtom->getType() == CAtom::W_ATOM )
	{
		theAtom->negateSelf(&ruleList);
		
		return false;
	}

	if ( theAtom->getUnitWAtomNum() != 1 ||
	     theAtom->getUnitWAtom(0)->getType() != CAtom::W_ATOM )
	{
		return false;
	}

	CWeightAtom * theWAtom = (CWeightAtom *)
		(theAtom->getUnitWAtom(0));

	m_iTrueAtomNum += theWAtom->multiFlip(&ruleList, theAtom);

	theAtom->negateSelf(&ruleList);
	if ( theAtom->getValue() )
		m_iTrueAtomNum++;
	else
		m_iTrueAtomNum--;
		
	
	m_iDoubleFlipNum++;

	return true;
}

bool CAssignment::permutationFlip(CAtom * theAtom, CRuleList & ruleList)
{
	if ( theAtom == NULL )
	{
		return false;
	}

	if ( theAtom->getType() == CAtom::C_ATOM ||
	     theAtom->getType() == CAtom::W_ATOM )
	{
		theAtom->negateSelf(&ruleList);

		return false;
	}

	if ( m_iAuxAtomNum[m_iChosenAtomIndex] != 3 )
	{
		theAtom->negateSelf(&ruleList);

		if ( theAtom->getValue() )
			m_iTrueAtomNum++;
		else
			m_iTrueAtomNum--;

		return false;
	}

	theAtom->negateSelf(&ruleList);
	m_pAuxAtoms[m_iChosenAtomIndex][0]->negateSelf(&ruleList);
	m_pAuxAtoms[m_iChosenAtomIndex][1]->negateSelf(&ruleList);
	m_pAuxAtoms[m_iChosenAtomIndex][2]->negateSelf(&ruleList);

	m_iPermutationFlipNum++;

	return true;
}

bool CAssignment::generalDoubleFlip(CAtom * theAtom, CRuleList & ruleList)
{
	if ( theAtom == NULL )
	{
		return false;
	}

	bool result = true;
	
	result = (result && flip(theAtom, ruleList));

	CArray<CRule *> theRules = theAtom->getGDBFRules();
	int ruleLength = theRules.getLength();
	
	for ( int i = 0; i < ruleLength; i++ )
	{
		if ( !theRules[i]->SAT() )
		{
			CAtom * pChosenAtom;
			
			pChosenAtom = pickAnAtomFromSameRule(theAtom,
							     *theRules[i]);
			result = (result && flip(pChosenAtom, ruleList));
		}
	}

	if ( result )
		m_iDoubleFlipNum++;

	return result;
}

bool CAssignment::fixIt(CRuleList & ruleList)
{
	CAtom * pChosenAtom;

	pChosenAtom = pickAnAtom(ruleList);
	
#ifdef __LLN_DEBUG__
	if ( pLog != NULL )
		fprintf(pLog, "%d:%d\n", pChosenAtom->getIndex(),
			(pChosenAtom->getValue()?1:0));
#endif
	if ( glPermutationCatoms )
	{
		return permutationFlip(pChosenAtom, ruleList);
	}
	else if ( glDisjointCatoms &&
		  (pChosenAtom->isCriticalInCAtom() ||
		   (myRand(glDBFlipProbabilityDenominator)
		    < glDBFlipProbabilityNumerator)) )
	{
		return doubleFlip(pChosenAtom, ruleList);
	}
	else if ( glDisjointWatoms &&
		  (pChosenAtom->isCriticalInWAtom() ||
		   (myRand(glDBFlipProbabilityDenominator)
		    < glDBFlipProbabilityNumerator)) )
	{
		return multiFlip(pChosenAtom, ruleList);
	}
	else if ( glGeneralDoubleFlip )
	{
		return generalDoubleFlip(pChosenAtom, ruleList);
	}
	else
	{
		return flip(pChosenAtom, ruleList);
	}
}

bool CAssignment::setForcedAtom(const char * text)
{
	char * pAtomText = new char[strlen(text)+1];

	strcpy(pAtomText, text);
	m_iForcedAtomLength++;
	return m_pForcedAtoms.add(pAtomText);
}

bool CAssignment::setAtomText(int id, const char * text)
{
	int pos = hasTheAtom(id);
	if ( pos == NONE )
		return false;

	m_pAtoms[pos]->setText(text);
	return true;
}

bool CAssignment::printPositiveAtoms(bool smodelsOutput,
				     int modelNum,
				     bool showPredicate,
				     const char * showAtom) const
{
	int numSAT = 0;

	if ( smodelsOutput )
	{
		cout<<"\nAnswer: "<<modelNum<<endl;
		cout<<"Stable Model: ";
	}
	else
	{
		cout<<"\nAnswer-set:\n";
		cout<<"Atoms determined during grounding:\n";
	}
	
	for ( int i = 0; i < m_iForcedAtomLength; i++ )
	{
		if ( showPredicate &&
		     showAtom != NULL &&
		     strstr(m_pForcedAtoms[i], showAtom) == NULL )
			continue;
		if ( smodelsOutput &&
		     strstr(m_pForcedAtoms[i], "smodels_internal_var") !=
		     NULL )
			continue;
		if ( smodelsOutput &&
		     strstr(m_pForcedAtoms[i], "th_bodyof") != NULL )
			continue;
		
		if ( glShowPositiveAtomsInPBFormat )
			continue;
		
		cout<<m_pForcedAtoms[i];
		if ( smodelsOutput )
			cout<<" ";
		else
			cout<<"."<<endl;
	}

	if ( !smodelsOutput && !glShowPositiveAtomsInPBFormat )
		cout<<"Other atoms:\n";
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		if ( m_pAtoms[i] == NULL )
		{
			continue;
		}

		if ( showPredicate &&
		     showAtom != NULL &&
		     strstr(m_pAtoms[i]->getText(), showAtom) == NULL )
			continue;
		if ( smodelsOutput &&
		     strstr(m_pAtoms[i]->getText(), "smodels_internal_var")
		     != NULL )
			continue;
		if ( smodelsOutput &&
		     strstr(m_pAtoms[i]->getText(), "th_bodyof") != NULL )
			continue;

		if ( glShowPositiveAtomsInPBFormat )
		{
			if ( m_pAtoms[i]->getValue() )
				cout<<" +1*v"<<m_pAtoms[i]->getIndex()
				    <<" >= +1; "<<endl;
			else
				cout<<" +1*v"<<m_pAtoms[i]->getIndex()
				    <<" <= 0; "<<endl;
			continue;
		}
		
		if ( m_pAtoms[i]->getType() == CAtom::N_ATOM &&
		     m_pAtoms[i]->getValue() )
		{
			numSAT++;
			cout<<m_pAtoms[i]->getText();
			if ( smodelsOutput )
				cout<<" ";
			else
				cout<<"."<<endl;
		}
	}

	if ( smodelsOutput )
	{
		cout<<endl;
		for ( int i = 0; i < m_iCAtomLength; i++ )
		{
			m_pCAtoms[i]->printWeight();
		}
	} else if ( !glShowPositiveAtomsInPBFormat )
	{	
		cout<<"Total Number of Positive Atoms : "<<numSAT<<endl;
	}
	
	return numSAT > 0;
}

void CAssignment::updateBestValues()
{
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		m_baBestValues[i] = m_pAtoms[i]->getValue();
	}
}

void CAssignment::printBestValues()
{
	int numSAT = 0;

	cout<<"\nAnswer-set: (not yet, just an assignment)\n";
	cout<<"Atoms determined during grounding:\n";
	for ( int i = 0; i < m_iForcedAtomLength; i++ )
	{
		cout<<m_pForcedAtoms[i]<<endl;
	}

	cout<<"Other atoms:\n";
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		if ( m_pAtoms[i] == NULL )
		{
			cout<<"The "<<i<<"th atom is NULL"<<endl;
			continue;
		}

		if ( m_pAtoms[i]->getType() == CAtom::N_ATOM &&
		     m_baBestValues[i] )
		{
			numSAT++;
			cout<<m_pAtoms[i]->getText()<<endl;
		}
	}
	
	cout<<"Total Number of Positive Atoms : "<<numSAT<<endl;
}

void CAssignment::setInitValues(const char * fileName)
{
	if ( !glUseInitValues )
		return;
	
	FILE * file = fopen(fileName, "r");

	if ( file == NULL )
	{
		glUseInitValues = false;
		return;
	}

	m_bInitValues.empty();
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		m_bInitValues.add(-1);
	}
	
	char * endOfFile = NULL;
	char aline[MAX_LINE_LENGTH];
	char atomText[MAX_LINE_LENGTH];
	bool trueAtoms = true;
	
	endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
	while ( endOfFile != NULL )
	{
		int i = 0;
		int length = strlen(aline);
		
		while ( i < length && isspace(aline[i]) )
			i++;

		if ( i >= length ||
		     (i < length && aline[i] == '%') )
		{
			trueAtoms = false;// from this point on, we will see false atoms
			endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
			continue;
		}
		
		int j = 0;
		while ( i < length && !isspace(aline[i]) &&
			aline[i] != '.' )
		{
			atomText[j] = aline[i];
			i++;
			j++;
		}
		// modified on 01/02/2006: atom name does not need to
		//                         have )
		//atomText[j] = ')';
		//atomText[j+1] = '\0';
		atomText[j] = '\0';

		int k;
		for ( k = 0; k < m_iAtomLength; k++ )
		{
			if ( m_pAtoms[k] == NULL )
				continue;
			
			if ( strstr(m_pAtoms[k]->getText(), atomText) != NULL )
				break;
		}

		if ( k < m_iAtomLength )
		{
			if ( trueAtoms )
				m_bInitValues[k] = 1;
			else
				m_bInitValues[k] = 0;
		}
		
		endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
	}
	if ( trueAtoms == true ) // no "%" occurs in the data
	{
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( m_bInitValues[i] == -1 )
				m_bInitValues[i] = 0;
		}
	}
}

void CAssignment::reinitialize(CRuleList & ruleList)
{
	bool value;

	glNumPickUnitRule = 0;
	glOverFlow = 0;
	glAllBestAtomsOverflow = 0;
	
// initialization using make count of an atom.	
	if ( !glUseInitValues &&
	     glPermutationCatoms &&
	     glInitialMethod == 3 )
	{
                bool * settled = new bool [m_iAtomLength];
		int unsettledNum = m_iAtomLength;

                for ( int i = 0; i < m_iAtomLength; i++ )
                {
                        if ( m_pAtoms[i] == NULL )
                        {
                                settled[i] = true;
				unsettledNum--;
                                continue;
                        }
                        settled[i] = false;
                        if ( m_pAtoms[i]->getValue() )
                                m_pAtoms[i]->negateSelf(&ruleList);
                }

		while ( unsettledNum > 0 )
		{
			int weight = 0;
			int bestAtoms[HEU_ATOM_NUM], numOfBestAtoms = 0;
	
			for ( int i = 0;
			      i < m_iAtomLength &&
				      numOfBestAtoms < HEU_ATOM_NUM;
			      i++ )
			{
				if ( settled[i] )
					continue;
				int n;
				
				n = m_pAtoms[i]->getMKCT();
		
				if ( n >= weight )
				{
					if ( n > weight )
					{
						numOfBestAtoms = 0;
						weight = n;
					}
					
					bestAtoms[numOfBestAtoms++] = i;
				}
			}
			
			int realPos;
			if ( numOfBestAtoms == 0 )
			{
				int vpos = myRand(unsettledNum);
				int id = 0;
				for ( realPos = 0; realPos < m_iAtomLength;
				      realPos++ )
				{
					if ( settled[realPos] )
						continue;
					if ( id == vpos )
						break;
					id++;
				}
			}
			else
				realPos = bestAtoms[myRand(numOfBestAtoms)];

			unsettledNum--;
			settled[realPos] = true;
			m_pAtoms[realPos]->negateSelf(&ruleList);
			CAtom ** allForcedAtoms;
			int forcedAtomNum;
			for ( int j = 0; j < m_pAtoms[realPos]->getCAtomNum();
			      j++ )
			{
				allForcedAtoms = m_pAtoms[realPos]->getCAtom(j)
					->getAtoms();
				forcedAtomNum = m_pAtoms[realPos]->getCAtom(j)
					->getAtomLength();
				for ( int k = 0; k < forcedAtomNum; k++ )
				{
					int id = allForcedAtoms[k]->getIndex()
						-1;
					
					if ( !settled[id] )
					{
						settled[id] = true;
						unsettledNum--;
					}
				}
			}
		}
		
                delete [] settled;
	}

// use random initialization, which seems to be not so good	
	else if ( !glUseInitValues &&
		  glPermutationCatoms &&
		  glInitialMethod == 2 )
	{
		bool * settled = new bool [m_iAtomLength];
		int unsettledNum = m_iAtomLength;
		
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( m_pAtoms[i] == NULL )
			{
				settled[i] = true;
				unsettledNum--;
				continue;
			}
			settled[i] = false;
			if ( m_pAtoms[i]->getValue() )
				m_pAtoms[i]->negateSelf(&ruleList);
		}

		while ( unsettledNum > 0 )
		{
			int vpos = myRand(unsettledNum);

			int id = 0, realPos;
			for ( realPos = 0; realPos < m_iAtomLength; realPos++ )
			{
				if ( settled[realPos] )
					continue;
				if ( id == vpos )
					break;
				id++;
			}

			settled[realPos] = true;
			unsettledNum--;
			m_pAtoms[realPos]->negateSelf(&ruleList);

			CAtom ** allForcedAtoms;
			int forcedAtomNum;
			for ( int j = 0; j < m_pAtoms[realPos]->getCAtomNum();
			      j++ )
			{
				allForcedAtoms = m_pAtoms[realPos]->getCAtom(j)
					->getAtoms();
				forcedAtomNum = m_pAtoms[realPos]->getCAtom(j)
					->getAtomLength();
				for ( int k = 0; k < forcedAtomNum; k++ )
				{
					int id = allForcedAtoms[k]->getIndex()
						-1;

					if ( !settled[id] )
					{
						settled[id] = true;
						unsettledNum--;
					}
				}
			}
		}
		
		delete [] settled;
	}

// use a random choice on the first atom to set to true, then follows the
// order of the atoms in the list.  Performs better than fully randomized.
        else if ( !glUseInitValues &&
		  glPermutationCatoms )
        {
                bool * settled = new bool [m_iAtomLength];

                for ( int i = 0; i < m_iAtomLength; i++ )
                {
                        if ( m_pAtoms[i] == NULL )
                        {
                                settled[i] = true;
                                continue;
                        }
                        settled[i] = false;
                        if ( m_pAtoms[i]->getValue() )
                                m_pAtoms[i]->negateSelf(&ruleList);
                }

                int pos = myRand(m_iAtomLength);
                for ( int i = 0;  i < m_iAtomLength; i++ )
                {
                        if ( pos >= m_iAtomLength )
                                pos = 0;
                        if ( settled[pos] )
                        {
                                pos++;
                                continue;
                        }

                        settled[pos] = true;
                        m_pAtoms[pos]->negateSelf(&ruleList);
                        CAtom ** allForcedAtoms;
                        int forcedAtomNum;
                        for ( int j = 0; j < m_pAtoms[pos]->getCAtomNum();
                              j++ )
                        {
                                allForcedAtoms = m_pAtoms[pos]->getCAtom(j)
                                        ->getAtoms();
                                forcedAtomNum = m_pAtoms[pos]->getCAtom(j)
                                        ->getAtomLength();
                                for ( int k = 0; k < forcedAtomNum; k++ )
                                {
                                        int id = allForcedAtoms[k]->getIndex()
                                                -1;

                                        settled[id] = true;
                                }
                        }
                        pos++;
                }

                delete [] settled;
        }

#ifdef __LLN_DEBUG__
	else if ( !glUseInitValues )
	{
		int rnd;
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( m_pAtoms[i] == NULL )
				continue;
		
			rnd = random();
			value = (rnd%glInitDenominator < glInitNumerator );
			if ( m_pAtoms[i]->getValue() != value )
				m_pAtoms[i]->negateSelf(&ruleList);
			fprintf(pLog, "init -- %d:%d\n", i+1, value);
		}	
		if ( glDisjointCatoms )
		{
			for ( int i = 0; i < m_iCAtomLength; i++ )
				if ( !m_pCAtoms[i]->SAT()  &&
				     m_pCAtoms[i]->getType() == CAtom::C_ATOM &&
				     m_pCAtoms[i]->isInUnitRule() )
					m_pCAtoms[i]->negateSelf(&ruleList);
		}
		if ( glDisjointWatoms )
		{
			for ( int i = 0; i < m_iCAtomLength; i++ )
				if ( !m_pCAtoms[i]->SAT()  &&
				     m_pCAtoms[i]->getType() == CAtom::W_ATOM &&
				     m_pCAtoms[i]->isInUnitRule() )
					m_pCAtoms[i]->negateSelf(&ruleList);
		}
		if ( glGeneralDoubleFlip && !glHasSetDBFRules )
		{
			glHasSetDBFRules = true;
			ruleList.setDBFRules(m_iAtomLength,
					     glGeneralDBFRuleFile);
			for ( int i = 0;
			      i < glRulesUsedInDBF.getLength();
			      i++ )
			{
				if ( !glRulesUsedInDBF[i]->SAT() )
				{
					CAtom * pChosenAtom;

					pChosenAtom =
						pickAnAtomFromSameRule
						(NULL,
						 *glRulesUsedInDBF[i]);
					pChosenAtom->negateSelf(&ruleList);
				}
			}
		}
	}

#else	
	else if ( !glUseInitValues )
	{
		for ( int i = 0; i < m_iAtomLength; i++ )
		{
			if ( m_pAtoms[i] == NULL )
				continue;

			value = (myRand(glInitDenominator) < glInitNumerator);
			if ( m_pAtoms[i]->getValue() != value &&
			     !m_pAtoms[i]->forced() )
				flip(m_pAtoms[i], ruleList);
			else if ( m_pAtoms[i]->forced() )
			{
				bool val = (m_pAtoms[i]->forced() == 1 ?
					    true :
					    false);
				if ( m_pAtoms[i]->getValue() != val )
					flip(m_pAtoms[i], ruleList);
			}
		}
		
		if ( glDisjointCatoms )
		{
			for ( int i = 0; i < m_iCAtomLength; i++ )
				if ( !m_pCAtoms[i]->SAT() &&
				     m_pCAtoms[i]->getType() ==
				     CAtom::C_ATOM &&
				     m_pCAtoms[i]->isInUnitRule() )
					m_pCAtoms[i]->negateSelf(&ruleList);
		}
		if ( glDisjointWatoms )
		{
			for ( int i = 0; i < m_iCAtomLength; i++ )
				if ( !m_pCAtoms[i]->SAT() &&
				     m_pCAtoms[i]->getType() ==
				     CAtom::W_ATOM &&
				     m_pCAtoms[i]->isInUnitRule() )
					m_pCAtoms[i]->negateSelf(&ruleList);
		}

		if ( glGeneralDoubleFlip && !glHasSetDBFRules )
		{
			glHasSetDBFRules = true;
			ruleList.setDBFRules(m_iAtomLength,
					     glGeneralDBFRuleFile);
			for ( int i = 0;
			      i < glRulesUsedInDBF.getLength();
			      i++ )
			{
				if ( !glRulesUsedInDBF[i]->SAT() )
				{
					if (glRulesUsedInDBF[i]->isUnitRule()) 
					{
						glRulesUsedInDBF[i]->
							getCAtoms()[0]->
							negateSelf(&ruleList);
					}
					else
					{
						CAtom * pChosenAtom;

						pChosenAtom =
							pickAnAtomFromSameRule
							(NULL,
							 *glRulesUsedInDBF[i]
							 );
						pChosenAtom->
							negateSelf(&ruleList);
					}
				}
			}
		}
	}
#endif
	else
	{
		for ( int i = 0; i < m_bInitValues.getLength(); i++ )
		{
			if ( m_pAtoms[i] == NULL )
				continue;

			if ( m_pAtoms[i]->forced() )
			{
				bool val = (m_pAtoms[i]->forced() == 1 ?
					    true :
					    false);
				if ( m_pAtoms[i]->getValue() != val )
					flip(m_pAtoms[i], ruleList);
				continue;
			}
			
			if ( m_bInitValues[i] != -1 ) // an assigned atom
			{
				bool val = (m_bInitValues[i] == 0 ?
					    false :
					    true);
				if ( m_pAtoms[i]->getValue() != val )
					flip(m_pAtoms[i], ruleList);
			}
			else
			{
				value = (myRand(glInitDenominator) <
					 glInitNumerator);
				if ( m_pAtoms[i]->getValue() != value )
					flip(m_pAtoms[i], ruleList);
			}
		}

		if ( !glInitValuesError &&
		     (glDisjointCatoms || glPermutationCatoms) )
		{
			for ( int i = 0;
			      i < m_iCAtomLength && !glInitValuesError;
			      i++ )
				if ( !m_pCAtoms[i]->SAT() &&
				     m_pCAtoms[i]->getType() ==
				     CAtom::C_ATOM &&
				     m_pCAtoms[i]->isInUnitRule() )
					glInitValuesError = true;
		}
		if ( !glInitValuesError &&
		     glDisjointWatoms )
		{
			for ( int i = 0;
			      i < m_iCAtomLength && !glInitValuesError;
			      i++ )
				if ( !m_pCAtoms[i]->SAT() &&
				     m_pCAtoms[i]->getType() ==
				     CAtom::W_ATOM &&
				     m_pCAtoms[i]->isInUnitRule() )
					glInitValuesError = true;
		}

		if ( glGeneralDoubleFlip && !glHasSetDBFRules )
		{
			glHasSetDBFRules = true;
			ruleList.setDBFRules(m_iAtomLength,
					     glGeneralDBFRuleFile);
			for ( int i = 0;
			      i < glRulesUsedInDBF.getLength();
			      i++ )
			{
				if ( !glRulesUsedInDBF[i]->SAT() )
				{
					if (glRulesUsedInDBF[i]->isUnitRule())
					{
						glRulesUsedInDBF[i]->
							getCAtoms()[0]->
							negateSelf(&ruleList);
					}
					else
					{
						CAtom * pChosenAtom;

						pChosenAtom =
							pickAnAtomFromSameRule
							(NULL,
							 *glRulesUsedInDBF[i]
							 );
						pChosenAtom->
							negateSelf(&ruleList);
					}
				}
			}
		}
	}
	
	m_iTrueAtomNum = 0;
	for ( int i = 0; i < m_iAtomLength; i++ )
	{
		if ( m_pAtoms[i] == NULL )
			continue;

		if ( m_pAtoms[i]->getValue() )
			m_iTrueAtomNum++;

		if ( glUseTimeStamp )
			m_pAtoms[i]->resetTimeStamp();
	}

	if ( glUseTimeStamp )
	{
		glLastTimeStamp = 0;
		ruleList.resetLastRecentlyFlipped();
	}
	
	m_iDoubleFlipNum = 0;
	m_iPermutationFlipNum = 0;
}

////////////////////////////
// class CRuleList
////////////////////////////

CRuleList::CRuleList(int ruleSize)
{
	m_iSATRuleNum = 0;
	m_iLastUNSATIndex = -1;

	m_iRuleLength = 0;
	m_pRules.newArray(ruleSize);
}

CRuleList::~CRuleList()
{
}

bool CRuleList::newRuleArray(int size)
{
	if ( size > 0 )
	{
		m_iRuleLength = 0;
		m_pRules.newArray(size);

		return true;
	}
	else
	{	
		m_iRuleLength = 0;
		m_pRules.empty();
		return false;
	}
}

void CRuleList::empty()
{
	for ( int i = 0; i < m_iRuleLength; i++ )
		delete m_pRules[i];
	
	m_iRuleLength = 0;
	m_pRules.empty();
}

void CRuleList::dump() const
{
	cout<<"The RuleList =>\n";
	cout<<"  m_iRuleLength = "<<m_iRuleLength<<endl;
	cout<<"  m_iRuleNum = "<<m_pRules.getSize()<<endl;
	cout<<"  m_iSATRuleNum = "<<m_iSATRuleNum<<endl;
	cout<<"  m_iLastUNSATIndex = "<<m_iLastUNSATIndex<<endl;
	
	for ( int i = 0; i < m_iRuleLength; i++ )
	{
		cout<<"  ";
		m_pRules[i]->dump();
	}
}

bool CRuleList::verify()
{
	for ( int i = 0; i < m_iRuleLength; i++ )
		if ( !m_pRules[i]->reeval() )
		{
			return false;
		}

	return true;
}

int CRuleList::addRule(CRule & rule)
{
	int rtrnPos;
	
	if ( rule.SAT() )
	{
		m_iSATRuleNum++;
		m_pRules.add(&rule);
		rtrnPos = m_iRuleLength;
	}
	else
	{
		m_iLastUNSATIndex++;
		if ( m_iLastUNSATIndex < m_iRuleLength )
		{
//			m_pRules.put(m_iRuleLength,
//				     m_pRules[m_iLastUNSATIndex]);
// Strange! add has some bug?
			m_pRules.add(m_pRules[m_iLastUNSATIndex]);
			m_pRules[m_iRuleLength]->setIndex(m_iRuleLength);
		}
		m_pRules.put(m_iLastUNSATIndex, &rule);
		rtrnPos = m_iLastUNSATIndex;
	}
	m_iRuleLength++;
	m_pRules.setLength(m_iRuleLength);
	rule.setIndex(rtrnPos);
	return rtrnPos;
}

void CRuleList::shrinkArrays()
{
	m_pRules.shrinkArray();
	for ( int i = 0; i < m_iRuleLength; i++ )
		m_pRules[i]->shrinkArrays();
}

inline bool CRuleList::allSAT() const
{
	if ( m_iSATRuleNum == m_iRuleLength )
		return true;
	else
		return false;
}

void CRuleList::swap(int index, bool currValue)
{
	if ( index < 0 || index > m_iRuleLength )
		return;
	if ( (currValue && index > m_iLastUNSATIndex) ||
	     (!currValue && index <= m_iLastUNSATIndex) ) // doesn't change
		return;

	CRule * temp;
	if ( currValue ) // the rule just changed from unsat to sat
	{
		temp = m_pRules[m_iLastUNSATIndex];
		m_pRules.put(m_iLastUNSATIndex, m_pRules[index]);
		m_pRules.put(index, temp);
		m_pRules[index]->setIndex(index);
		m_pRules[m_iLastUNSATIndex]->setIndex(m_iLastUNSATIndex);
		m_iLastUNSATIndex--;
		m_iSATRuleNum++;
	}
	else // the rule just changed from sat to unsat
	{
		temp = m_pRules[m_iLastUNSATIndex+1];
		m_pRules.put(m_iLastUNSATIndex+1, m_pRules[index]);
		m_pRules.put(index, temp);
		m_pRules[index]->setIndex(index);
		m_pRules[m_iLastUNSATIndex+1]->setIndex(m_iLastUNSATIndex+1);
		m_iLastUNSATIndex++;
		m_iSATRuleNum--;
	}
}

int CRuleList::getUNSATRuleLength() const
{
	return m_iLastUNSATIndex+1;
}

const CRule & CRuleList::getUNSATRule(int id) const
{
	if ( glPlotTouchedTimes && !glPlotTouchedTimesStart &&
	     getUNSATRuleLength() <= glPlotTouchedTimesThreshold )
	{
		glPlotTouchedTimesStart = true;
	}
	
	if ( id < 0 || id > m_iLastUNSATIndex )
	{
		if ( !m_pRules[0]->SAT() )
		{
			m_pRules[0]->touched(glPlotTouchedTimesStart);
		}
		
		return *m_pRules[0];
	}
	else
	{
		m_pRules[id]->touched(glPlotTouchedTimesStart);
		return *m_pRules[id];
	}
}

void CRuleList::resetLastRecentlyFlipped()
{
	for ( int i = 0; i < m_iRuleLength; i++ )
		m_pRules[i]->resetLastRecentlyFlipped();
}

void CRuleList::resetTouched()
{
	glPlotTouchedTimesStart = false;
	
	for ( int i = 0; i < m_iRuleLength; i++ )
		m_pRules[i]->resetTouched();
}

void CRuleList::plotTouched(FILE * pPlotFile) const
{
	int highest = 0;
	
	for ( int i = 0; i < m_iRuleLength; i++ )
	{
		if ( m_pRules[i]->getTouchedTimes() < glPlotLowerBound )
			continue;

		if ( m_pRules[i]->getTouchedTimes() > highest )
			highest = m_pRules[i]->getTouchedTimes();
		
		fprintf(pPlotFile, "%d %d\n",
			m_pRules[i]->getOrigIndex(),
			m_pRules[i]->getTouchedTimes());
	}

	glPlotHighest = highest;
}

bool CRuleList::setDBFRules(int atomNum, const char * fileName)
{
	if ( !glGeneralDoubleFlip )
	{
		return false;
	}

	CArray<CRule *> *candidates;
	if ( fileName != NULL && strlen(fileName) != 0 )
	{
		FILE *file = fopen(fileName, "r");
		
		if ( file == NULL )
		{
			glGeneralDoubleFlip = false;
			return false;
		}

		candidates = new CArray<CRule *>;

		char * endOfFile = NULL;
		char aline[MAX_LINE_LENGTH];
	
		endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
		while ( endOfFile != NULL )
		{
			int i = 0;
			int length = strlen(aline);
		
			while ( i < length && isspace(aline[i]) )
				i++;

			if ( i >= length ||
			     (i < length && aline[i] == '%') )
			{
				endOfFile = fgets(aline, MAX_LINE_LENGTH,
						  file);
				continue;
			}
		
			int index = getInt(aline, i);
		
			if ( index == NONE || index < 0 ||
			     index >= m_iRuleLength )
			{
				endOfFile = fgets(aline, MAX_LINE_LENGTH,
						  file);
				continue;
			}

			int j;
			for ( j = 0; j < m_iRuleLength; j++ )
				if ( m_pRules[j]->getOrigIndex() == index )
					break;

			if ( j >= m_iRuleLength )
				continue;
		
			candidates->add(m_pRules[j]);
		
			endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
		}
		fclose(file);
	}
	else
		candidates = &m_pRules;

	CArray<bool> coveredAtoms(atomNum);
	coveredAtoms.setLength(atomNum);
	
	for ( int i = 0; i < atomNum; i++ )
		coveredAtoms[i] = false;

	int len = candidates->getLength();
	int offset = myRand(len);
	
	glRulesUsedInDBF.empty();

	for ( int i = 0; i < len; i++ )
	{
		CRule * thisRule = (*candidates)[(i+offset)%len];
		CAtom ** allAtoms = thisRule->getAtoms();
		int n = thisRule->getAtomLength();

		int j;
		for ( j = 0; j < n; j++ )
		{
			// note: atom's index starts from 1, not 0
			if ( coveredAtoms[allAtoms[j]->getIndex()-1] )
				break;
		}

		if ( j < n )
		{
			continue;
		}
		else
		{
			for ( int k = 0; k < n; k++ )
				coveredAtoms[allAtoms[k]->getIndex()-1] = true;
			
			glRulesUsedInDBF.add(thisRule);
			CAtom ** pRuleAtoms = thisRule->getAtoms();
			int ruleAtomLength = thisRule->getAtomLength();
			for ( int i = 0; i < ruleAtomLength; i++ )
				pRuleAtoms[i]->addGDBFRule(*thisRule);
		}
	}

	glCandidateNum = len;
	
	return true;
}

void CRuleList::reinitialize()
{
	for ( int i = 0; i < m_iRuleLength; i++ )
		m_pRules[i]->init();
}

////////////////////////////////
// class CWalkASPPS
////////////////////////////////

void CWalkASPPS::dump() const
{
	m_Assignment.dump();
	m_RuleList.dump();
}

bool CWalkASPPS::toDIMACS(const char * filename)
{
	FILE * file, * dimacsFile;
	char aline[MAX_LINE_LENGTH], dimacsFileName[MAX_LINE_LENGTH];
	char buffer[MAX_BUFFER_LENGTH];
	char * endOfFile;
	int count = 0;
	int numOfRules = 1, processBarUnit = 10000;
	char processBar[16];
	
	file = fopen(filename, "r");
	if ( file == NULL )
	{
		cerr<<"Error: can't open the aspps program file\n";
		printf("==================================== END ");
		printf("=======================================\n\n");
		exit(5);
	}

	char * surfix = strstr(filename, ".aspps");
	if ( surfix == NULL )
		sprintf(dimacsFileName, "%s.cnf", filename);
	else
	{
		int i = 0;
		for ( const char * ptr = filename; ptr != surfix; ptr++ )
		{
			dimacsFileName[i] = *ptr;
			i++;
		}
		dimacsFileName[i] = '\0';
		strcat(dimacsFileName, ".cnf");
	}
	
	dimacsFile = fopen(dimacsFileName, "r");
	if ( dimacsFile != NULL )
	{
		fclose(dimacsFile);
		cerr<<"Warning: "
		    <<"do you want to replace the existing file? (y/n) ";
		char reply[MAX_LINE_LENGTH];
		fgets(reply, MAX_LINE_LENGTH, stdin);
		if ( reply[0] != 'y' && reply[0] != 'Y' )
		{
			printf("==================================== END ");
			printf("=======================================\n\n");
			exit(5);
		}
	}

	dimacsFile = fopen(dimacsFileName, "w");
	if ( dimacsFile == NULL )
	{
		cerr<<"Error: can't create the output file\n";
		printf("==================================== END ");
		printf("=======================================\n\n");
		exit(5);
	}
	
	cout<<"Starting to convert program file =>     ";
	endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
	while ( endOfFile != NULL )
	{
		if ( aline[0] == 'p' )
		{
			int i = 1;
			
			int numOfAtoms = getInt(aline, i);
			if ( numOfAtoms == NONE )
				exit(3);

			numOfRules = getInt(aline, i);
			if ( numOfRules == NONE )
				exit(4);

			if ( numOfRules > INIT_BAR_LENGTH )
				processBarUnit = numOfRules / INIT_BAR_LENGTH;
			else
				processBarUnit = 1;


			sprintf(buffer, "p cnf %d %d\n",
				numOfAtoms, numOfRules);
			if ( fputs(buffer, dimacsFile) == EOF )
				exit(6);
		}
		else if ( aline[0] == 'c' )
		{
			if ( fputs(aline, dimacsFile) == EOF )
				exit(6);			
		}
		else
		{
			int i = 0;
			bool isBody = true;
			int id;
			char sid[MAX_NAME_LENGTH];
			
			buffer[0] = '\0';
			while ( (size_t)i < strlen(aline) && aline[i] != '\n' )
			{
				if ( aline[i] == '{' || aline[i] == '}' )
					exit(6);

				if ( aline[i] == ',' )
				{
					isBody = false;
					i++;
					continue;
				}

				id = getInt(aline, i);
				if ( id == NONE )
				{
					i++;
					continue;
				}
				
				if ( isBody )
					id = 0 - id;
				sprintf(sid, " %d ", id);
				strcat(buffer, sid);
			}

			strcat(buffer, " 0\n");
			if ( fputs(buffer, dimacsFile) == EOF )
				exit(6);
			
			if ( count % processBarUnit == 0 )
			{
				int percentage = (int)((double)count * 100.0 /
						       (double)numOfRules);
				sprintf(processBar, "\b\b\b\b. %2d%%",
					percentage);
				cout<<processBar;
				flush(cout);
			}
			count++;
		}
		endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
	}

	sprintf(processBar, "\b\b\b\b\b. %3d%%", 100);
	cout<<processBar;
	cout<<"\n";
	flush(cout);
	
	fclose(file);
	fclose(dimacsFile);
	return true;
}

bool CWalkASPPS::toASPPS(const char * filename)
{
	FILE * file, * asppsFile;
	char aline[MAX_LINE_LENGTH], asppsFileName[MAX_LINE_LENGTH];
	char buffer[MAX_BUFFER_LENGTH];
	char * endOfFile;
	int numOfRules = 1, processBarUnit = 10000;
	char processBar[16];
	int count = 0;
	
	file = fopen(filename, "r");
	if ( file == NULL )
	{
		cerr<<"Error: can't open the aspps program file\n";
		printf("==================================== END ");
		printf("=======================================\n\n");
		exit(5);
	}

	char * surfix = strstr(filename, ".cnf");
	if ( surfix == NULL )
		sprintf(asppsFileName, "%s.aspps", filename);
	else
	{
		int i = 0;
		for ( const char * ptr = filename; ptr != surfix; ptr++ )
		{
			asppsFileName[i] = *ptr;
			i++;
		}
		asppsFileName[i] = '\0';
		strcat(asppsFileName, ".aspps");
	}
	
	asppsFile = fopen(asppsFileName, "r");
	if ( asppsFile != NULL )
	{
		fclose(asppsFile);
		cerr<<"Warning: "
		    <<"do you want to replace the existing file? (y/n) ";
		char reply[MAX_LINE_LENGTH];
		fgets(reply, MAX_LINE_LENGTH, stdin);
		if ( reply[0] != 'y' && reply[0] != 'Y' )
		{
			printf("==================================== END ");
			printf("=======================================\n\n");
			exit(5);
		}
	}

	asppsFile = fopen(asppsFileName, "w");
	if ( asppsFile == NULL )
	{
		cerr<<"Error: can't create the output file\n";
		printf("==================================== END ");
		printf("=======================================\n\n");
		exit(5);
	}
	
	cout<<"Starting to convert program file =>     ";
	endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
	while ( endOfFile != NULL )
	{
		if ( aline[0] == 'p' )
		{
			int i = 1;
			char junk[16];

			int len = getToken(aline, junk, 16, i);
			if ( len == NONE || strcmp(junk, "cnf") != 0 )
				exit(5);
			
			int numOfAtoms = getInt(aline, i);
			if ( numOfAtoms == NONE )
				exit(3);

			numOfRules = getInt(aline, i);
			if ( numOfRules == NONE )
				exit(4);

			if ( numOfRules > INIT_BAR_LENGTH )
				processBarUnit = numOfRules / INIT_BAR_LENGTH;
			else
				processBarUnit = 1;

			sprintf(buffer, "p %d 0 %d\n", numOfAtoms, numOfRules);
			if ( fputs(buffer, asppsFile) == EOF )
				exit(6);			
		}
		else if ( aline[0] == 'c' )
		{
			if ( fputs(buffer, asppsFile) == EOF )
				exit(6);
		}
		else if ( aline[0] == '%' )
		{	
			break;
		}
		else
		{
			int i = 0;
			int id;
			char sid[MAX_NAME_LENGTH];
			
			buffer[0] = '\0';
			while ( (size_t)i < strlen(aline) && aline[i] != '\n' )
			{
				id = getInt(aline, i);
				if ( id == NONE )
				{
					i++;
					continue;
				}
				
				if ( id < 0 )
				{
					id = 0 - id;
					sprintf(sid, " %d ", id);
					strcat(buffer, sid);
				}
			}

			strcat(buffer, " , ");
			
			i = 0;
			while ( (size_t)i < strlen(aline) && aline[i] != '\n' )
			{
				id = getInt(aline, i);
				if ( id == NONE )
				{
					i++;
					continue;
				}

				if ( id == 0 )
					break;
				if ( id > 0 )
				{
					sprintf(sid, " %d ", id);
					strcat(buffer, sid);
				}
			}
			
			strcat(buffer, " \n");
			if ( fputs(buffer, asppsFile) == EOF )
				exit(6);
			
			if ( count % processBarUnit == 0 )
			{
				int percentage = (int)((double)count * 100.0 /
						       (double)numOfRules);
				sprintf(processBar, "\b\b\b\b. %2d%%",
					percentage);
				cout<<processBar;
				flush(cout);
			}
			count++;
		}
		endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
	}

	sprintf(processBar, "\b\b\b\b\b. %3d%%", 100);
	cout<<processBar;
	cout<<"\n";
	flush(cout);
	
	fclose(file);
	fclose(asppsFile);
	return true;
}

bool CWalkASPPS::initializeByASPPSFile(const char * filename, bool beQuiet)
{
	FILE * file;
	char aline[MAX_LINE_LENGTH];
	char * endOfFile;
	int count = 0;
	int numOfRules = 1, processBarUnit = 10000;
	char processBar[16];
	int ruleIndex = 0;

	if ( strcmp(filename, "") != 0 )
	{
		file = fopen(filename, "r");
		if ( file == NULL )
		{
			cerr<<"Error: can't open the aspps program file\n";
			printf("==================================== END ");
			printf("=======================================\n\n");
			exit(5);
		}
	}
	else
		file = stdin;

	endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
	while ( endOfFile != NULL )
	{
		if ( aline[0] == 'p' )
		{
			int i = 1;
			
			int numOfAtoms = getInt(aline, i);
			if ( numOfAtoms == NONE )
				exit(3);

			numOfRules = getInt(aline, i);
			if ( numOfRules == NONE )
				exit(4);

			int numOfRulesinNewVersion = getInt(aline, i);
			if ( numOfRulesinNewVersion != NONE )
				numOfRules = numOfRulesinNewVersion;

			if ( numOfRules > INIT_BAR_LENGTH )
				processBarUnit = numOfRules / INIT_BAR_LENGTH;
			else
				processBarUnit = 1;

			if ( !glSmodelsOutput )
			{
				cout<<"Allocating the atom array of size "
				    <<numOfAtoms<<endl;
			}
			
			m_Assignment.newAtomArray(numOfAtoms);

			if ( !glSmodelsOutput )
			{
				cout<<"Allocating the rule array of size "
				    <<numOfRules<<endl;
			}
			
			m_RuleList.newRuleArray(numOfRules);

			if ( glSmodelsOutput )
				cout<<"done"<<endl;
			else
			{
				if ( !beQuiet )
					cout<<"Starting to process program file "
					    <<"=>     ";
			}
		}
		else if ( aline[0] == 'c' )
		{
			int i = 1;
			
			int index = getInt(aline, i);
			if ( index == NONE )
				exit(6);

			char buffer[MAX_NAME_LENGTH];
			getAtomText(aline, buffer, MAX_NAME_LENGTH, i);
			
			if ( index == 0 )
				m_Assignment.setForcedAtom(buffer);
			else
				m_Assignment.setAtomText(index, buffer);
		}
		else
		{
			CRule * pNewRule = new CRule;
			if ( pNewRule == NULL )
				exit(2);

			if ( m_Assignment.parseAtoms(aline, *pNewRule) )
			{
				pNewRule->init();
				pNewRule->setOrigIndex(ruleIndex);
				ruleIndex++;
				m_RuleList.addRule(*pNewRule);
			}
			else
			{
				delete pNewRule;
			}

			if ( !glSmodelsOutput )
			{
				if ( !beQuiet && count % processBarUnit == 0 )
				{
					int percentage = (int)((double)count *
							       100.0 /
							       (double)
							       numOfRules);
					sprintf(processBar, "\b\b\b\b. %2d%%",
						percentage);
					cout<<processBar;
					flush(cout);
				}
			}
			count++;
		}
		endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
	}

	if ( !glSmodelsOutput )
	{
		if ( !beQuiet )
		{
			sprintf(processBar, "\b\b\b\b\b. %3d%%", 100);
			cout<<processBar;
			cout<<"\n";
			flush(cout);
		}
	}

	if ( strcmp(filename, "") != 0 )
		fclose(file);
	return true;
}

bool CWalkASPPS::initializeByDimacsFile(const char * filename, bool beQuiet)
{
	FILE * file;
	char aline[MAX_LINE_LENGTH];
	char * endOfFile;
	int count = 0;
	int numOfRules = 1, processBarUnit = 10000;
	char processBar[16];
	int ruleIndex = 0;

	if ( strcmp(filename, "") != 0 )
	{
		file = fopen(filename, "r");
		if ( file == NULL )
		{
			cerr<<"Error: can't open the cnf file\n";
			printf("==================================== END ");
			printf("=======================================\n\n");
			exit(5);
		}
	}
	else
		file = stdin;

	endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
	while ( endOfFile != NULL )
	{
		if ( aline[0] == 'p' )
		{
			int i = 1;
			char junk[16];

			int len = getToken(aline, junk, 16, i);
			if ( len == NONE || strcmp(junk, "cnf") != 0 )
				exit(5);
			
			int numOfAtoms = getInt(aline, i);
			if ( numOfAtoms == NONE )
				exit(3);

			numOfRules = getInt(aline, i);
			if ( numOfRules == NONE )
				exit(4);

			if ( numOfRules > INIT_BAR_LENGTH )
				processBarUnit = numOfRules / INIT_BAR_LENGTH;
			else
				processBarUnit = 1;

			if ( !glSmodelsOutput )
			{
				cout<<"Allocating the atom array of size "
				    <<numOfAtoms<<endl;
			}
			
			m_Assignment.newAtomArray(numOfAtoms);

			if ( !glSmodelsOutput )
			{
				cout<<"Allocating the rule array of size "
				    <<numOfRules<<endl;
			}
			
			m_RuleList.newRuleArray(numOfRules);

			if ( glSmodelsOutput )
				cout<<"done"<<endl;
			else
			{
				if ( !beQuiet )
					cout<<"Starting to process program file "
					    <<"=>      ";
			}
		}
		else if ( aline[0] == 'c' )
		{
		}
		else if ( aline[0] == '%' )
		{	
			break;
		}
		else
		{
			CRule * pNewRule = new CRule;
			if ( pNewRule == NULL )
				exit(2);
			
			if ( m_Assignment.parseAtomsInDimacsForm(aline,
								 *pNewRule) )
			{
				pNewRule->init();
				pNewRule->setOrigIndex(ruleIndex);
				ruleIndex++;
				m_RuleList.addRule(*pNewRule);
			}
			else
			{
				// Bug: if a and not a occur in the clause
				//      at the same time! (fixed)
				// glHasEmptyRules = true;
				delete pNewRule;
			}

			if ( !glSmodelsOutput )
			{
				if ( !beQuiet && count % processBarUnit == 0 )
				{
					int percentage = (int)((double)count *
							       100.0 /
							       (double)
							       numOfRules);
					sprintf(processBar, "\b\b\b\b. %2d%%",
						percentage);
					cout<<processBar;
					flush(cout);
				}
			}
			count++;
		}
		endOfFile = fgets(aline, MAX_LINE_LENGTH, file);
	}

	if ( !glSmodelsOutput )
	{
		if ( !beQuiet )
		{
			sprintf(processBar, "\b\b\b\b\b. %3d%%", 100);
			cout<<processBar;
			cout<<"\n";
			flush(cout);
		}
	}

	if ( strcmp(filename, "") != 0 )
		fclose(file);
	
	return true;
}

bool CWalkASPPS::run(bool beQuiet, bool noProcessBar, bool debugMode,
		     bool isCNFProg,
		     const char * filename, bool verify, long randSeed,
		     bool showAllPositiveAtoms,
		     bool showPredicate, const char * showAtom)
{
	struct timeval tv;
	struct timezone tzp;
	gettimeofday(&tv,&tzp);
	long seed;
	
	if ( randSeed == 0 )
		seed = (( tv.tv_sec & 0177 ) * 1000000) + tv.tv_usec;
	else
		seed = randSeed;
	srandom(seed);
	
	struct timeval startTime, endTime, diff;
	struct timezone timeZoneJunk;
	struct tms cpu;
	double cpuBeginTime = 0, cpuEndTime = 0, cpuDiff;

	cout<<"Random Number Seed: "<<seed<<endl;

	if ( glSmodelsOutput )
	{
		cout<<"wsat(cc) reading ...";
	}
	else
	{
		cout<<"Number of Retry: "<<glMaxRetryNum<<endl;
		cout<<"Number of Flips in Each Try: "<<glMaxFlipNum<<endl;
		cout<<"Noise: "<<glNoiseProbability
		    <<" ("<<glWalkProbabilityNumerator<<" : "
		    <<glWalkProbabilityDenominator<<") "
		    <<endl;
		cout<<"Initialize an atom to true with probability: "
		    <<glInitToTrueProb
		    <<" ("<<glInitNumerator<<" : "
		    <<glInitDenominator<<") "
		    <<endl;
		cout<<"Approximation method: "
		    <<glCombinationMethod
		    <<endl;
		if ( glDisjointCatoms || glDisjointWatoms )
			cout<<"Double Flip Prob. : "<<glDBFlipProbability
			    <<" ("<<glDBFlipProbabilityNumerator<<" : "
			    <<glDBFlipProbabilityDenominator<<") "
			    <<endl;
	}
	
	gettimeofday(&startTime, &timeZoneJunk);

	glDoubleFlipApplicable = 0;
	
	if ( isCNFProg )
	{
		initializeByDimacsFile(filename,
				       (beQuiet||noProcessBar));
	}
	else
	{
		initializeByASPPSFile(filename,
				      (beQuiet||noProcessBar));
	}

	if ( !glSmodelsOutput )
	{
		cout<<"Num of Unit Rules: "<<glUnitRuleNum<<endl;
		cout<<endl;
	}
	
	if ( glPlotTouchedTimesThreshold < 0 )
		glPlotTouchedTimesThreshold = m_RuleList.getRuleLength();
	
	if ( glHasEmptyRules )
	{
		cout<<endl<<"NOTE:\t";
		cout<<"The ground program has empty rule(s), thus "
		    <<"it is unsatisfiable."<<endl;
		
		return false;
	}
	
	if ( m_Assignment.getAtomLength() == 0 ||
	     m_RuleList.getRuleLength() == 0 )
	{
		cout<<endl<<"NOTE:\t";
		cout<<"The ground program contains no atoms/rules."<<endl;
		return true;
	}
	
	if ( glUseInitValues )
	{
		m_Assignment.setInitValues(glInitFileName);
	}
	
	if ( glDisjointCatoms || glDisjointWatoms )
	{
		if ( glDoubleFlipApplicable == 0 )
		{
			cout<<"*************************************"
			    <<"****************************"<<endl;
			cout<<"* Error: no disjoint C-Atoms at all, "
			    <<"DB flip is not applicable! *"<<endl;
			cout<<"*************************************"
			    <<"****************************"<<endl;
		}
		else if ( glDoubleFlipApplicable == 1 )
		{
			cout<<"******************************************"
			    <<"***********************************"<<endl;
			cout<<"* Warning: some C-Atoms are not disjoint, "
			    <<"DB flip couldn't be applied there *"<<endl;
			cout<<"******************************************"
			    <<"***********************************"<<endl;
		}
	}
	
	gettimeofday(&endTime, &timeZoneJunk);

	diff.tv_sec = endTime.tv_sec - startTime.tv_sec;
	diff.tv_usec = endTime.tv_usec - startTime.tv_usec;
	if ( diff.tv_usec < 0 )
	{
		diff.tv_usec += 1000000;
		diff.tv_sec--;
	}

	if ( !glSmodelsOutput )
	{
		cout<<"During initialization: "
		    <<"   "<<(double)(diff.tv_sec) +
			((double)(diff.tv_usec) / (double)1000000)
		    <<" seconds elapsed.\n\n";
	
		if ( !beQuiet)
			cout<<"Starting searching:\n";
	}

	bool retval = false, optval = false;
	
	gettimeofday(&startTime, &timeZoneJunk);
	times(&cpu);
	cpuBeginTime = cpu.tms_utime;

	if ( m_Assignment.isMaxMin() == 0 ) // non-optimization case
	{
		if ( solve(seed, beQuiet, noProcessBar, debugMode,
			   isCNFProg, filename, verify, randSeed,
			   showAllPositiveAtoms,
			   showPredicate, showAtom) )
		{
			cout<<"s SATISFIABLE"<<endl;
			flush(cout);
			gettimeofday(&endTime, &timeZoneJunk);
			times(&cpu);
			retval = true;
		}
		else
		{
			cout<<"s UNKNOWN"<<endl;
			flush(cout);
			gettimeofday(&endTime, &timeZoneJunk);
			times(&cpu);
			retval = false;
		}
	}
	else if ( m_Assignment.isMaxMin() == 1 ) // max case
	{
		int oldMaxFlip = glMaxFlipNum;
		
		glStopOnFirst = true;
		glNumOfSolBeforeStop = 1;
		
		// first phase: use binary search to find an approximation
		cout<<"c Phase 1 ..."<<endl;
		while ( true )
		{
			if ( optval )
				break;

//			if ( finite(glAvgFlipPerSecond) )
//			{
//				glMaxFlipNum = (int)(glAvgFlipPerSecond) *
//					((int)(glBinarySearchJumpTime)+
//					 glBinarySearchTimeLimit);
//			}

			cout<<"c   trying "<<m_Assignment.getOptimalValue()+1
			    <<": "<<endl;
			if ( !solve(seed, beQuiet, noProcessBar, debugMode,
				    isCNFProg, filename, verify, randSeed,
				    false, false, showAtom) ||
			     glUserInterrupted )
				break;
			
//			int oldJumpValue = m_Assignment.getJumpValue();
			
			optval = m_Assignment.improveBound(m_RuleList);
			cout<<"o "<<m_Assignment.getOptimalValue()<<endl;
			flush(cout);

			gettimeofday(&endTime, &timeZoneJunk);
			times(&cpu);

			retval = true;

//			m_Assignment.updateBestValues();
			m_Assignment.tryMiddleAsLB();

//			int newJumpValue = m_Assignment.getNextJumpValue();
			
//			glBinarySearchTimeLimit =
//				(int)((glBinarySearchJumpTime /
//				      (oldJumpValue+1))*
//				      (newJumpValue+1));
			
			if ( !beQuiet )
			{
				diff.tv_sec = endTime.tv_sec -
					startTime.tv_sec;
				diff.tv_usec = endTime.tv_usec -
					startTime.tv_usec;
				if ( diff.tv_usec < 0 )
				{
					diff.tv_usec += 1000000;
					diff.tv_sec--;
				}
	
				cout<<"c \nc Total wall-time:         "
				    <<(double)(diff.tv_sec) +
					((double)(diff.tv_usec) /
					 (double)1000000)
				    <<" seconds.\n";
				
				cpuEndTime = cpu.tms_utime;
				
				cpuDiff = (cpuEndTime-cpuBeginTime)/
					(double)sysconf(_SC_CLK_TCK);
				
				cout<<"c Total CPU time:           "
				    <<cpuDiff<<" seconds.\nc \n";
			}
		
		}

		glMaxFlipNum = oldMaxFlip;

		// second phase: rough linear search
		cout<<"c Phase 2 ..."<<endl;
		m_Assignment.restoreLB();
		while ( true )
		{
			if ( optval )
				break;
			
			cout<<"c   trying "<<m_Assignment.getOptimalValue()+1
			    <<": "<<endl;
			if ( !solve(seed, beQuiet, noProcessBar, debugMode,
				    isCNFProg, filename, verify, randSeed,
				    false, false, showAtom) ||
			     glUserInterrupted )
				break;
			optval = m_Assignment.improveBound(m_RuleList);
			cout<<"o "<<m_Assignment.getOptimalValue()<<endl;
			flush(cout);

			gettimeofday(&endTime, &timeZoneJunk);
			times(&cpu);

//			m_Assignment.updateBestValues();
			retval = true;
			if ( !beQuiet )
			{
				diff.tv_sec = endTime.tv_sec -
					startTime.tv_sec;
				diff.tv_usec = endTime.tv_usec -
					startTime.tv_usec;
				if ( diff.tv_usec < 0 )
				{
					diff.tv_usec += 1000000;
					diff.tv_sec--;
				}
	
				cout<<"c \nc Total wall-time:         "
				    <<(double)(diff.tv_sec) +
					((double)(diff.tv_usec) /
					 (double)1000000)
				    <<" seconds.\n";
				
				cpuEndTime = cpu.tms_utime;
				
				cpuDiff = (cpuEndTime-cpuBeginTime)/
					(double)sysconf(_SC_CLK_TCK);
				
				cout<<"c Total CPU time:           "
				    <<cpuDiff<<" seconds.\nc \n";
			}
		
		}
		
		if ( optval )
			cout<<"s OPTIMUM FOUND: O = "
			    <<m_Assignment.getOptimalValue()<<endl;
		else if ( retval )
			cout<<"s SATISFIABLE: O = "
			    <<m_Assignment.getOptimalValue()<<endl;
		else
			cout<<"s UNKNOWN"<<endl;
		if ( (optval || retval) &&
		     (showAllPositiveAtoms || showPredicate) )
			m_Assignment.printPositiveAtoms(false,
							showAllPositiveAtoms,
							showPredicate,
							showAtom);
	}
	else if ( m_Assignment.isMaxMin() == 2 ) // min case
	{
		int oldMaxFlip = glMaxFlipNum;
		glStopOnFirst = true;
		glNumOfSolBeforeStop = 1;

		// first phase: use binary search to find an approximation
		cout<<"c Phase 1 ..."<<endl;
		while ( true )
		{
			if ( optval )
				break;
			
//			if ( finite(glAvgFlipPerSecond) )
//			{
//				glMaxFlipNum = ((int)(glAvgFlipPerSecond)+1) *
//					((int)(glBinarySearchJumpTime)+
//					 glBinarySearchTimeLimit);
//			}			

			cout<<"c   trying "<<m_Assignment.getOptimalValue()-1
			    <<": "<<endl;
			if ( !solve(seed, beQuiet, noProcessBar, debugMode,
				    isCNFProg, filename, verify, randSeed,
				    false, false, showAtom) ||
			     glUserInterrupted )
				break;
			
//			int oldJumpValue = m_Assignment.getJumpValue();
			//cout<<"www oldjumpvalue = "<<oldJumpValue<<endl;
			
			optval = m_Assignment.improveBound(m_RuleList);
			cout<<"o "<<m_Assignment.getOptimalValue()<<endl;
			flush(cout);

			gettimeofday(&endTime, &timeZoneJunk);
			times(&cpu);

			retval = true;

//			m_Assignment.updateBestValues();
			m_Assignment.tryMiddleAsUB();

//			int newJumpValue = m_Assignment.getNextJumpValue();
			//cout<<"www newjumpvalue = "<<newJumpValue<<endl;
			//cout<<"www jumptime = "<<glBinarySearchJumpTime<<endl;			
//			glBinarySearchTimeLimit =
//				(int)((glBinarySearchJumpTime /
//				       (oldJumpValue+1)) *
//				      (newJumpValue+1));

//			cout<<"www searchtimelimit = "<<glBinarySearchTimeLimit<<endl;
			
			if ( !beQuiet )
			{
				diff.tv_sec = endTime.tv_sec -
					startTime.tv_sec;
				diff.tv_usec = endTime.tv_usec -
					startTime.tv_usec;
				if ( diff.tv_usec < 0 )
				{
					diff.tv_usec += 1000000;
					diff.tv_sec--;
				}
	
				cout<<"c \nc Total wall-time:         "
				    <<(double)(diff.tv_sec) +
					((double)(diff.tv_usec) /
					 (double)1000000)
				    <<" seconds.\n";
				
				cpuEndTime = cpu.tms_utime;
				
				cpuDiff = (cpuEndTime-cpuBeginTime)/
					(double)sysconf(_SC_CLK_TCK);
				
				cout<<"c Total CPU time:           "
				    <<cpuDiff<<" seconds.\nc \n";
			}
		
		}

		glMaxFlipNum = oldMaxFlip;

		// second phase: rough linear search
		cout<<"c Phase 2 ..."<<endl;
		m_Assignment.restoreUB();
		while ( true )
		{
			if ( optval )
				break;

			cout<<"c   trying "<<m_Assignment.getOptimalValue()-1
			    <<": "<<endl;
			if ( !solve(seed, beQuiet, noProcessBar, debugMode,
				    isCNFProg, filename, verify, randSeed,
				    false, false, showAtom) ||
			     glUserInterrupted )
				break;
			optval = m_Assignment.improveBound(m_RuleList);
			cout<<"o "<<m_Assignment.getOptimalValue()<<endl;
			flush(cout);

			gettimeofday(&endTime, &timeZoneJunk);
			times(&cpu);

//			m_Assignment.updateBestValues();
			retval = true;
			if ( !beQuiet )
			{
				diff.tv_sec = endTime.tv_sec -
					startTime.tv_sec;
				diff.tv_usec = endTime.tv_usec -
					startTime.tv_usec;
				if ( diff.tv_usec < 0 )
				{
					diff.tv_usec += 1000000;
					diff.tv_sec--;
				}
	
				cout<<"c \nc Total wall-time:         "
				    <<(double)(diff.tv_sec) +
					((double)(diff.tv_usec) /
					 (double)1000000)
				    <<" seconds.\n";
				
				cpuEndTime = cpu.tms_utime;
				
				cpuDiff = (cpuEndTime-cpuBeginTime)/
					(double)sysconf(_SC_CLK_TCK);
				
				cout<<"c Total CPU time:           "
				    <<cpuDiff<<" seconds.\nc \n";
			}
		
		}

		if ( optval )
			cout<<"s OPTIMUM FOUND: O = "
			    <<m_Assignment.getOptimalValue()<<endl;
		else if ( retval )
			cout<<"s SATISFIABLE: O = "
			    <<m_Assignment.getOptimalValue()<<endl;
		else
			cout<<"s UNKNOWN"<<endl;
		if ( (optval || retval) &&
		     (showAllPositiveAtoms || showPredicate) )
			m_Assignment.printPositiveAtoms(false,
							showAllPositiveAtoms,
							showPredicate,
							showAtom);
	}	
////
	diff.tv_sec = endTime.tv_sec - startTime.tv_sec;
	diff.tv_usec = endTime.tv_usec - startTime.tv_usec;
	if ( diff.tv_usec < 0 )
	{
		diff.tv_usec += 1000000;
		diff.tv_sec--;
	}
	
	cout<<"\nct During the search for the optimum:         "
	    <<(double)(diff.tv_sec) +
		((double)(diff.tv_usec) / (double)1000000)
	    <<" seconds elapsed.\n";

	times(&cpu);
	cpuEndTime = cpu.tms_utime;

	cpuDiff = (cpuEndTime-cpuBeginTime)/(double)sysconf(_SC_CLK_TCK);

	cout<<"ct CPU time spent:           "
	    <<cpuDiff<<" seconds.\n\n";
	
//	cout<<"Min power: "
//	    <<glPower
//	    <<endl;
//	cout<<"Max comb-base: "
//	    <<glCombBase
//	    <<endl<<endl;
	
	if ( glInitValuesError )
	{
		cout<<"c ****************************************************"
		    <<endl;
		cout<<"c * Warning: initial data caused some C-Atoms UNSAT, *"
		    <<endl
		    <<"c * DB/PERM flip has not been applied successfully.  *"
		    <<endl;
		cout<<"c ****************************************************"
		    <<endl;
	}

	return retval;
}

bool CWalkASPPS::solve(long seed,
		       bool beQuiet, bool noProcessBar, bool debugMode,
		       bool isCNFProg,
		       const char * filename, bool verify, long randSeed,
		       bool showAllPositiveAtoms,
		       bool showPredicate, const char * showAtom)
{
	struct timeval startTime, endTime, diff;
	struct timezone timeZoneJunk;
	struct tms cpu;
	double cpuBeginTime = 0, cpuEndTime = 0, cpuDiff;

	int flipNum = 0; 
	int retryNum = 0, totalFlipNum = 0;
	int flipNumInSuccessRuns = 0;
	int processBarUnit = 3000;
	char processBar[16];
	int processBarPrinted = 0;
	int successNum = 0;

	gettimeofday(&startTime, &timeZoneJunk);
	times(&cpu);
	cpuBeginTime = cpu.tms_utime;
	if ( !beQuiet && !noProcessBar )
	{
		if ( glMaxFlipNum > SEARCH_BAR_LENGTH )
			processBarUnit = glMaxFlipNum /
				SEARCH_BAR_LENGTH;
		else
			processBarUnit = 1;		
	}

	int datIndexTrueAtom = 0;
	int datIndexUnsatRule = 0;
	
	int normalWalkNumerator = glWalkProbabilityNumerator;
	int normalWalkDenominator = glWalkProbabilityDenominator;
	int unsatRuleNum = 0;
	
	while ( retryNum < glMaxRetryNum && !glUserInterrupted )
	{
		FILE * datTAN = NULL;

		datIndexTrueAtom = 0;
		datIndexUnsatRule = 0;

		if ( !glSmodelsOutput )
		{
			if ( !beQuiet && !noProcessBar )
				cout<<"c Try #"<<retryNum<<" =>     ";
			else if ( !beQuiet )
				cout<<"c Try #"<<retryNum<<" => ";
		}
		
		if ( glUseInitValues && glSetEscapePara )
		{
			glWalkProbabilityNumerator = glEscapeProbNumerator;
			glWalkProbabilityDenominator = glEscapeProbDenominator;
		}

		m_Assignment.reinitialize(m_RuleList);
		m_RuleList.resetTouched();
		
		if ( glPlotTrueAtomNum )
		{
			char fileName[256];
		
			sprintf(fileName, "TrueAtomNum_%ld_%d.dat",
				seed, retryNum);
			datTAN = fopen(fileName, "w");
			
			if ( datTAN == NULL )
			{
				glPlotTrueAtomNum = false;
			}
			else
			{
				fprintf(datTAN,
					"# Data File for number of true atoms, Run #%d\n", retryNum);
				fprintf(datTAN, "%d %d\n",
					datIndexTrueAtom++,
					m_Assignment.getTrueAtomNum());
			}
		}

		if ( glPlotUnsatRuleNum )
		{
			char fileName[256];
		
			sprintf(fileName, "UnsatRuleNum_%ld_%d.dat",
				seed, retryNum);
			datTAN = fopen(fileName, "w");
			
			if ( datTAN == NULL )
			{
				glPlotUnsatRuleNum = false;
			}
			else
			{
				fprintf(datTAN,
					"# Data File for number of unsat rules, Run #%d\n", retryNum);
				fprintf(datTAN, "%d %d\n",
					datIndexUnsatRule++,
					m_RuleList.getUNSATRuleLength());
			}
		}
	
		flipNum = 0;
		processBarPrinted = 0;
		glMinUnsatClauses = m_RuleList.getUNSATRuleLength();
		if ( glPrintAssignmentAfterFlip >= 0 )
		{
			m_Assignment.updateBestValues();
		}
		
		while ( flipNum < glMaxFlipNum && !m_RuleList.allSAT() &&
			!glUserInterrupted )
		{
			int unsatLen = m_RuleList.getUNSATRuleLength();
			if ( glPrintAssignmentAfterFlip < 0 )
			{
				if ( glMinUnsatClauses > unsatLen )
					glMinUnsatClauses = unsatLen;
			}
			else if ( glPrintAssignmentAfterFlip >= 0 &&
			          glPrintAssignmentAfterFlip <= flipNum+1 )
			{
				if ( glMinUnsatClauses > unsatLen )
				{
					m_Assignment.updateBestValues();
					glMinUnsatClauses = unsatLen;
				}
			}
			
			if ( glUseInitValues && glSetEscapePara &&
			     flipNum >= glMaxFlipNum * glEscapeLimit / 100 )
			{
				glWalkProbabilityNumerator =
					normalWalkNumerator;
				glWalkProbabilityDenominator =
					normalWalkDenominator;
			}
			
			m_Assignment.fixIt(m_RuleList);
			
			if ( glPlotTrueAtomNum &&
			     datIndexTrueAtom % glPlotFrequency ==  0 )
			{
				fprintf(datTAN, "%d %d\n",
					datIndexTrueAtom++,
					m_Assignment.getTrueAtomNum());
			}
			else
			{
				datIndexTrueAtom++;
			}
			if ( glPlotUnsatRuleNum &&
			     datIndexUnsatRule % glPlotFrequency ==  0 )
			{
				fprintf(datTAN, "%d %d\n",
					datIndexUnsatRule++,
					m_RuleList.getUNSATRuleLength());
			}
			else
			{
				datIndexUnsatRule++;
			}

			if ( !glSmodelsOutput )
			{
				if ( verify &&
				     !m_RuleList.allSAT() &&
				     verifyAssignment() )
				{
					cerr<<"c Error: verification passed "
					    <<"but allSAT() is false.\n";
				}

				if ( !beQuiet  && !noProcessBar &&
				     flipNum % processBarUnit == 0 )
				{
					int percentage = (int)((double)
							       flipNum *
							       100.0 /
							       (double)
							       glMaxFlipNum);
			
					sprintf(processBar, "\b\b\b\b. %2d%%",
						percentage);
					cout<<processBar;
					flush(cout);
					processBarPrinted++;
				}
			}
			
			flipNum++;		
		}
		unsatRuleNum += m_RuleList.getUNSATRuleLength();

//		m_Assignment.dump();
//		m_RuleList.dump();
		
		if ( glPlotTrueAtomNum )
		{
			fprintf(datTAN, "%d %d\n",
				datIndexTrueAtom++,
				m_Assignment.getTrueAtomNum());
			fclose(datTAN);
		}

		if ( glPlotUnsatRuleNum )
		{
			fprintf(datTAN, "%d %d\n",
				datIndexUnsatRule++,
				m_RuleList.getUNSATRuleLength());
			fclose(datTAN);
		}

		if ( glPlotTouchedTimes )
		{
			FILE * datTouched = NULL;
			char fileName[256];
		
			sprintf(fileName, "TouchedTimes_%ld_%d.dat",
				seed, retryNum);
			datTouched = fopen(fileName, "w");
			
			if ( datTouched == NULL )
			{
				glPlotTouchedTimes = false;
			}
			else
			{
				fprintf(datTouched,
					"# Data File for touched times of rules, Run #%d\n", retryNum);
				m_RuleList.plotTouched(datTouched);
				fclose(datTouched);
			}
		}

		if ( !glSmodelsOutput )
		{
			if ( !beQuiet && !noProcessBar )
			{
				cout<<"\b\b\b\b";
				for ( int i = processBarPrinted - 4;
				      i <= SEARCH_BAR_LENGTH; i++ )
					cout<<".";
				
				if ( m_RuleList.allSAT() )
					sprintf(processBar,
						"\b\b\b\b\b\b\b\b Succeed!");
				else
					sprintf(processBar,
						"\b\b\b\b\b Fail!");
				
				cout<<processBar;
				cout<<"\n";
				flush(cout);
			}
			else if ( !beQuiet )
			{
				if ( m_RuleList.allSAT() )
					cout<<"Succeed!\n";
				else
					cout<<"Fail!\n";
			}
		}
		
		if ( m_RuleList.allSAT() )
		{
			flipNumInSuccessRuns += flipNum;
			successNum++;
			if ( verify )
			{
				cout<<"c           Verifying Found Assignment "
				    <<"............................... ";
				if ( verifyAssignment() )
					cout<<"PASSED!\n";
				else
					cout<<"NOT PASSED!\n";
			}

			if ( !glSmodelsOutput )
			{
				if ( !beQuiet )
				{
					cout<<"c             Num of flips: "
					    <<flipNum<<endl;
					cout<<"c             Min num of unsat rules before sat: "
					    <<glMinUnsatClauses<<endl;
					if ( debugMode )
					{
						cout<<"c             Pick the unit rule "
						    <<glNumPickUnitRule<<" times"<<endl;
						cout<<"c             "
						    <<"Num of overflow computation: "
						    <<glOverFlow<<endl;
						cout<<"c             "
						    <<"Choose from overflowed atoms "
						    <<glAllBestAtomsOverflow
						    <<" times"<<endl;
					}
				}
			}
			
		}
		else
		{
			if ( !glSmodelsOutput )
			{
				if ( !beQuiet )
				{
					cout<<"c             Num of final unsat rules: "
					    <<m_RuleList.getUNSATRuleLength()<<endl;
					cout<<"c             Min num of unsat rules: "
					    <<glMinUnsatClauses<<endl;
				}
				
				if ( !beQuiet )
				{
					cout<<"c             Num of flips: "
					    <<flipNum<<endl;
					if ( debugMode )
					{
						cout<<"c             Pick the unit rule "
						    <<glNumPickUnitRule<<" times"<<endl;
						cout<<"c             "
						    <<"Num of overflow computation: "
						    <<glOverFlow<<endl;
						cout<<"c             "
						    <<"Choose from overflowed atoms "
						    <<glAllBestAtomsOverflow
						    <<" times"<<endl;
					}
				}
			}	
		}

		if ( !glSmodelsOutput )
		{
			if ( !beQuiet && (glDisjointCatoms ||
					  glDisjointWatoms ||
					  glGeneralDoubleFlip) )
				cout<<"c             Num of double flips: "
				    <<m_Assignment.getDoubleFlipNum()<<endl;
			if ( !beQuiet && glGeneralDoubleFlip )
				cout<<"c             Num of rules used in DBF: "
				    <<glRulesUsedInDBF.getLength()
				    <<" / "
				    <<glCandidateNum<<endl;
			
			if ( !beQuiet && glPermutationCatoms )
				cout<<"c             Num of permutation flips: "
				    <<m_Assignment.getPermutationFlipNum()<<endl;
		}
		
		if ( m_RuleList.allSAT() &&
		     (showAllPositiveAtoms ||
		      showPredicate ||
		      glSmodelsOutput) )
		{
			m_Assignment.printPositiveAtoms(glSmodelsOutput,
							successNum,
							showPredicate,
							showAtom);
		}

		if ( !glSmodelsOutput )
		{
			if ( glPrintAssignmentAfterFlip >= 0 &&
			     !m_RuleList.allSAT() )
			{
				m_Assignment.printBestValues();
			}
		}
		
		totalFlipNum += flipNum;
		retryNum++;

		if ( glStopOnFirst && successNum >= glNumOfSolBeforeStop )
			break;
	}

	if ( glPlotTrueAtomNum )
	{
		char fileName[256];
		
		FILE * plotTAN = NULL;
		sprintf(fileName, "TrueAtomNum_%ld.plot", seed);
		plotTAN = fopen(fileName, "w");
		if ( plotTAN == NULL )
		{
			cout<<"Cannot generate plot file\n";
		}
		else
		{
			fprintf(plotTAN,
				"# Plot File for number of true atoms\n");
			fprintf(plotTAN,
				"set title \"Changes on the number of true atoms\"\n");
			fprintf(plotTAN, "set xlabel \"Flip Index\"\n");
			fprintf(plotTAN, "set ylabel \"Num of TRUE Atoms\"\n");

			fprintf(plotTAN, "plot \\\n");
			for ( int i = 0; i < retryNum-1; i++ )
			{
				sprintf(fileName, "TrueAtomNum_%ld_%d.dat",
					seed, i);
				fprintf(plotTAN,
					"\"%s\" using 1:2 title 'run #%d' with linespoints, \\\n",
					fileName, i);
			}
			sprintf(fileName, "TrueAtomNum_%ld_%d.dat",
				seed, retryNum-1);
			fprintf(plotTAN,
				"\"%s\" using 1:2 title 'run #%d' with linespoints \n",
				fileName, retryNum-1);
			fclose(plotTAN);
		}
	}

	if ( glPlotUnsatRuleNum )
	{
		char fileName[256];
		
		FILE * plotTAN = NULL;
		sprintf(fileName, "UnsatRuleNum_%ld.plot", seed);
		plotTAN = fopen(fileName, "w");
		if ( plotTAN == NULL )
		{
			cout<<"Cannot generate plot file\n";
		}
		else
		{
			fprintf(plotTAN,
				"# Plot File for number of unsat rules\n");
			fprintf(plotTAN,
				"set title \"Changes on the number of unsat rules\"\n");
			fprintf(plotTAN, "set xlabel \"Flip Index\"\n");
			fprintf(plotTAN, "set ylabel \"Num of unsat rules\"\n");

			fprintf(plotTAN, "plot \\\n");
			for ( int i = 0; i < retryNum-1; i++ )
			{
				sprintf(fileName, "UnsatRuleNum_%ld_%d.dat",
					seed, i);
				fprintf(plotTAN,
					"\"%s\" using 1:2 title 'run #%d' with linespoints, \\\n",
					fileName, i);
			}
			sprintf(fileName, "UnsatRuleNum_%ld_%d.dat",
				seed, retryNum-1);
			fprintf(plotTAN,
				"\"%s\" using 1:2 title 'run #%d' with linespoints \n",
				fileName, retryNum-1);
			fclose(plotTAN);
		}
	}

	if ( glPlotTouchedTimes )
	{
		char fileName[256];

		for ( int i = 0; i < retryNum; i++ )
		{
			FILE * plotTouched = NULL;
			sprintf(fileName, "TouchedTimes_%ld_%d.plot",
				seed, i);
			plotTouched = fopen(fileName, "w");
			if ( plotTouched == NULL )
			{
				cout<<"Cannot generate plot file\n";
			}
			else
			{
				fprintf(plotTouched,
					"# Plot File for touched times of rules\n");
				fprintf(plotTouched,
					"set title \"Touched Times of Rules\"\n");
				fprintf(plotTouched,
					"set xlabel \"Rule Index\"\n");
				fprintf(plotTouched,
					"set ylabel \"Touched Times\"\n");
				fprintf(plotTouched, "set yrange [0 : %d]\n",
					glPlotHighest +
					((int)(glPlotHighest * 0.1) > 5 ?
					 (int)(glPlotHighest * 0.1) : 5));

				sprintf(fileName, "TouchedTimes_%ld_%d.dat",
					seed, i);
				fprintf(plotTouched,
					"plot \"%s\" using 1:2 title 'run #%d' with impulses\n",
					fileName, i);
				fclose(plotTouched);
			}
		}
	}
	
	gettimeofday(&endTime, &timeZoneJunk);
	diff.tv_sec = endTime.tv_sec - startTime.tv_sec;
	diff.tv_usec = endTime.tv_usec - startTime.tv_usec;
	if ( diff.tv_usec < 0 )
	{
		diff.tv_usec += 1000000;
		diff.tv_sec--;
	}
	
	cout<<"\nc During searching:         "
	    <<(double)(diff.tv_sec) +
		((double)(diff.tv_usec) / (double)1000000)
	    <<" seconds elapsed.\n";

	times(&cpu);
	cpuEndTime = cpu.tms_utime;

	cpuDiff = (cpuEndTime-cpuBeginTime)/(double)sysconf(_SC_CLK_TCK);

	glBinarySearchJumpTime = cpuDiff;
	
	cout<<"c CPU time spent:           "
	    <<cpuDiff<<" seconds.\n\n";
	
	cout<<"c Total number of flips:    "<<totalFlipNum<<endl;
	cout<<"c That in success tries:    "
	    <<flipNumInSuccessRuns<<endl;
	cout<<"c Average success flips:    "
	    <<(successNum != 0 ? flipNumInSuccessRuns/successNum : 0)
	    <<endl;
	cout<<"c Average unsat rules  :    "
	    <<(retryNum - successNum != 0 ?
	       unsatRuleNum/(retryNum-successNum) : 0)
	    <<endl;

	glAvgFlipPerSecond = ((double)totalFlipNum /
			      (double)cpuDiff + 0.00000001);
	
	cout<<"c Average flips per second: "
	    <<glAvgFlipPerSecond
	    <<endl<<endl;
	cout<<"c Success Retries / Total:  "
	    <<successNum<<" / "<<retryNum<<endl;
	cout<<"c Success rate:             "
	    <<(double)successNum*100.0/(double)retryNum
	    <<"%\n\n";
	
	if ( successNum > 0 )
	{
		cout<<"c ASSIGNMENT FOUND!\n\n";
	}
	else
	{
		cout<<"c ASSIGNMENT NOT FOUND!\n\n";
	}

	if ( glInitValuesError )
	{
		cout<<"c ****************************************************"
		    <<endl;
		cout<<"c * Warning: initial data caused some C-Atoms UNSAT, *"
		    <<endl
		    <<"c * DB/PERM flip has not been applied successfully.  *"
		    <<endl;
		cout<<"c ****************************************************"
		    <<endl;
	}

	if ( glWriteLogs )
	{
		bool newStatusFile = false;
		struct stat junk;
		FILE * statusFile;
	
		if ( stat(STATUS_FILE_NAME,&junk) == -1 )
			newStatusFile = true;

		if ( (statusFile=fopen(STATUS_FILE_NAME, "a")) == NULL )
		{
			cout<<"c Error: Cannot append to the status file"<<endl;
		}
		else
		{
			if ( newStatusFile )
				fprintf(statusFile,
					"Var       Clause    Seed      CPU Time  Tries Cutoff    ASF/AUC   SR  Filename\n");
			fprintf(statusFile,
				"%-10i%-10i%-10li%-10g%-6i%-10i%-8i%3i%%  %s\n",
				m_Assignment.getAtomLength(),
				m_RuleList.getRuleLength(),
				seed,
				cpuDiff,
				glMaxRetryNum,
				glMaxFlipNum,
				(successNum != 0 ?
				 (int)(flipNumInSuccessRuns/successNum) :
				 (retryNum != 0 ?
				  (int)(unsatRuleNum/retryNum) :
				  0)),
				(retryNum != 0 ?
				 (int)(successNum*100/retryNum) :
				 0),
				filename);
			fclose(statusFile);
		}
	}
	
	return (successNum > 0 ? true : false);
}

bool CWalkASPPS::verifyAssignment()
{
	return m_RuleList.verify();
}

/////////////////////////////
// main
/////////////////////////////

void printUsage()
{
	cout<<"Usage: wsatcc -f filename [OPTIONS]"<<endl;
	cout<<"OPTIONS include:"<<endl;
	
	cout<<endl<<"*** Input Control ***"<<endl;
	cout<<"  -D:\t\t"
	    <<"DIMACS format is used in the input file;"
	    <<endl;
	cout<<"  -DT <file>:\t"
	    <<"Use the given data file as initial assignment for each try;"
	    <<endl;
	
	cout<<"*** Computation Control ***"<<endl;
	cout<<"  -DF [nn [mm]]: "
	    <<"Use double-flip procedure on simple theories;"
	    <<endl;
	cout<<"  -RNP:\t\t"
	    <<"Use rnovelty+ heuristic instead of SKC;"
	    <<endl;	
	cout<<"  -c nn:\t"
	    <<"The cutoff (or the MAX number of flips) of each try;"
	    <<endl;
	cout<<"  -t nn:\t"
	    <<"The total number of tries;"
	    <<endl;
	cout<<"  -F [n]:\t"
	    <<"Stop after finding the first n solutions (n=1 by default);"
	    <<endl;
	cout<<"  -s nn:\t"
	    <<"The seed of the random number generator;"
	    <<endl;
	cout<<"  -N nn [mm]:\t"
	    <<"The noise ratio (default is 10 100);"
	    <<endl;
	cout<<"  -T nn [mm]:\t"
	    <<"The probability (nn : mm) of initializing an atom to TRUE;"
	    <<endl<<"\t\t"
	    <<"(default is 50 100);"
	    <<endl;
	cout<<"  -V:\t\t"
	    <<"Verify solutions;"
	    <<endl;

	cout<<"*** Output Control ***"<<endl;
	cout<<"  -SM:\t\t"
	    <<"Output solutions in smodels output format."
	    <<endl;
	cout<<"  -A:\t\t"
	    <<"Output each solution by printing all positive atoms"
	    <<endl<<"\t\t"
	    <<"in the solution;"
	    <<endl;
	cout<<"  -S <pred>:\t"
	    <<"Output each solution by printing only positive atoms that"
	    <<endl<<"\t\t"
	    <<"have the given predicate name <pred> in the solution;"
	    <<endl;
	cout<<"  -Q:\t\t"
	    <<"Turn on the quiet mode. Least info will be printed;"
	    <<endl;
	cout<<"  -q:\t\t"
	    <<"Turn on the quiet mode. Less info will be printed;"
	    <<endl;
	cout<<"  --nolog:\t"
	    <<"Turn off the logs. No wsatcc.stat will be created."
	    <<endl;
	
	cout<<endl;
}

/*
void printUsage()
{
	cout<<"Usage: wsatcc -f filename [OPTIONS]"<<endl;
	cout<<"OPTIONS include:"<<endl;
	
	cout<<endl<<"*** Input Format Control ***"<<endl;
	cout<<"-D, --dimacs:"
	    <<endl<<"\t\t"
	    <<"DIMACS format is used in the input file;"
	    <<endl;
	cout<<"-DT, --data <file name>:"
	    <<endl<<"\t\t"
	    <<"Use the given data file as initial assignment for each try;"
	    <<endl;
	
	cout<<endl<<"*** Computation Control ***"<<endl;
	cout<<"-C nn, --combination-computation nn:"
	    <<endl<<"\t\t"
	    <<"Specify the method to compute the combination number (n, k);"
	    <<endl<<"\t\t"
	    <<"  nn = 1: use linear approximation;"
	    <<endl<<"\t\t"
	    <<"  nn = 2[default]: use quadratic approximation;"
	    <<endl<<"\t\t"
	    <<"  nn = 3: use brute-force computation (may cause overflow);"
	    <<endl;
	cout<<"-DF [nn [mm]], --double-flip [nn [mm]]:"
	    <<endl<<"\t\t"
	    <<"When the given theory is simple, double flip can keep all"
	    <<endl<<"\t\t"
	    <<"the unit catom rules SAT. If flipping a chosen atom will"
	    <<endl<<"\t\t"
	    <<"break a unit catom rule, then double flip method will pick"
	    <<endl<<"\t\t"
	    <<"an opposite atom in the same unit catom rule to flip."
	    <<endl<<"\t\t"
	    <<"Otherwise, with probability nn : mm (default is 0 100) to"
	    <<endl<<"\t\t"
	    <<"perform double flip;"
	    <<endl;
	cout<<"-DF-W, --double-flip-walk [nn [mm]]:"
	    <<endl<<"\t\t"
	    <<"Almost the same as -DBF. The only exception is when the"
	    <<endl<<"\t\t"
	    <<"program performs random \"walk\" step, it will not"
	    <<endl<<"\t\t"
	    <<"try to keep all disjoint catoms SAT.  This works well"
	    <<endl<<"\t\t"
	    <<"in Schur # problem;"
	    <<endl;
	cout<<"-DF-G, --double-flip-general [file]:"
	    <<endl<<"\t\t"
	    <<"Apply DBF on a set of disjoint rules specified in [file]."
	    <<endl<<"\t\t"
	    <<"If no file is given, then find a set of disjoint rules from"
	    <<endl<<"\t\t"
	    <<"all rules in the theory;"
	    <<endl;
	cout<<"-PF, --permutation-flip:"
	    <<endl<<"\t\t"
	    <<"When catoms only occur in orthogonal 1-1 catom rules,"
	    <<endl<<"\t\t"
	    <<"permutation flip can keep all the orthogonal 1-1 catom rules"
	    <<endl<<"\t\t"
	    <<"SAT.  The idea is that, since there could be one and only one"
	    <<endl<<"\t\t"
	    <<"atom TRUE in each orthogonal 1-1 catom rule and each atom"
	    <<endl<<"\t\t"
	    <<"appears in exactly two such catom rules, we can use 4 flips"
	    <<endl<<"\t\t"
	    <<"to maintain the SAT status of all orthogonal 1-1 catom rules."
	    <<endl<<"\t\t"
	    <<"Since the set of orthogonal 1-1 catom rules usually defines"
	    <<endl<<"\t\t"
	    <<"a permutation, we can think of permutation flip as the action"
	    <<endl<<"\t\t"
	    <<"of directly swapping two elements in the permutation:"
	    <<endl<<"\t\t\t"
	    <<"(1, 3, 2, 4) => (1, 4, 2, 3)"
	    <<endl<<"\t\t\t"
	    <<"    -     -         -     -"
	    <<endl<<"\t\t"
	    <<"3 and 4 are swapped and we can use 4 flips to implement"
	    <<endl<<"\t\t"
	    <<"this kind of swapping;"
	    <<endl;
	cout<<"-SW, --single-weight:"
	    <<endl<<"\t\t"
	    <<"When double flip or permutation flip is used, each \"flip\""
	    <<endl<<"\t\t"
	    <<"actually may end up with 2 or 4 flips.  So in choosing the"
	    <<endl<<"\t\t"
	    <<"best candidate atom to flip, we can either use the break"
	    <<endl<<"\t\t"
	    <<"count of the atom alone, or use the sum of 2 or 4 break-counts"
	    <<endl<<"\t\t"
	    <<"of 2 or 4 atoms. However, you can use this parameter to force"
	    <<endl<<"\t\t"
	    <<"the program to use single break-count even if"
	    <<endl<<"\t\t"
	    <<"double or permutation flip is used;"
	    <<endl;
	cout<<"-RNP, --rnovelty+:"
	    <<endl<<"\t\t"
	    <<"Use rnovelty+ heuristic;"
	    <<endl;	
	cout<<"-c, --cutoff nn:"
	    <<endl<<"\t\t"
	    <<"The cutoff (or the MAX number of flips) of each try;"
	    <<endl;
	cout<<"-t, --tries nn:"
	    <<endl<<"\t\t"
	    <<"The total number of tries;"
	    <<endl;
	cout<<"-F, --stop-on-first:"
	    <<endl<<"\t\t"
	    <<"Stop right after finding the first solution;"
	    <<endl;
	cout<<"-s, --seed nn:"
	    <<endl<<"\t\t"
	    <<"The seed of the random number generator;"
	    <<endl;
	cout<<"-N, --noise nn [mm]:"
	    <<endl<<"\t\t"
	    <<"The noise ratio (default is 10 100);"
	    <<endl;
	cout<<"-T, --true nn [mm]:"
	    <<endl<<"\t\t"
	    <<"The probability (nn : mm) of initializing an atom to TRUE;"
	    <<endl<<"\t\t"
	    <<"(default is 50 100);"
	    <<endl;
	cout<<"-V, --verify-solution:"
	    <<endl<<"\t\t"
	    <<"Perform solution verification when solutions are found;"
	    <<endl;

	cout<<endl<<"*** Output Control ***"<<endl;
	cout<<"-I [n], --show-init: [n]"
	    <<endl<<"\t\t"
	    <<"Output assignments after the nth flip;"
	    <<endl;
	cout<<"-SM, --smodels-format:"
	    <<endl<<"\t\t"
	    <<"Output solutions in smodels output format."
	    <<endl;
	cout<<"-A, --show-all:"
	    <<endl<<"\t\t"
	    <<"Output each solution by printing all positive atoms"
	    <<endl<<"\t\t"
	    <<"in the solution;"
	    <<endl;
	cout<<"-S, --show predicate:"
	    <<endl<<"\t\t"
	    <<"Output each solution by printing only positive atoms that"
	    <<endl<<"\t\t"
	    <<"have the given predicate name in the solution;"
	    <<endl;
	cout<<"-B, --debug:"
	    <<endl<<"\t\t"
	    <<"Turn on debug mode by which more info will be dumped;"
	    <<endl;
	cout<<"-Q, --QUIET:"
	    <<endl<<"\t\t"
	    <<"Turn on the quiet mode. No statistical info will be dumped;"
	    <<endl;
	cout<<"--nolog:"
	    <<endl<<"\t\t"
	    <<"Turn off the logs. No wsatcc.stat will be created."
	    <<endl;
	cout<<"-q, --no-process-bar:"
	    <<endl<<"\t\t"
	    <<"Turn on the mode in which no process bar will be shown. This"
	    <<endl<<"\t\t"
	    <<"is useful when you want to redirect the output to a file and"
	    <<endl<<"\t\t"
	    <<"want to skip all the messy characters used to print the"
	    <<endl<<"\t\t"
	    <<"process bar;"
	    <<endl;
	cout<<"-pta, --plot-true-atom-number nn:"
	    <<endl<<"\t\t"
	    <<"Generate GNUPLOT files for the # of true atoms after nn flips;"
	    <<endl;
	cout<<"-ptu, --plot-unsat-rule-number nn:"
	    <<endl<<"\t\t"
	    <<"Generate GNUPLOT files for the # of unsat rules after nn flips;"
	    <<endl;
	cout<<"-ptt, --plot-touched-times nn:"
	    <<endl<<"\t\t"
	    <<"Generate GNUPLOT files for the clauses that are touched for at"
	    <<endl<<"\t\t"
	    <<"least nn times;"
	    <<endl;

	cout<<endl<<"*** Utilities ***"<<endl;
	cout<<"-h, --help:     Print out help message."<<endl;
	
	cout<<endl;
}
*/

void handleKBInterrupt(int sig)
{
	if ( glUserInterrupted ) exit(-1);
	glUserInterrupted = true;
}

int main(int argc, char * argv[])
{
	printf("\nwsat(cc): a stochastic local search solver for propositional theories with generalized pseudo Boolean constraints.\n");
#ifdef __WSATCC_VERSION__
	printf("VERSION: %s\n", __WSATCC_VERSION__);
#endif	

#ifdef __AUTHOR__
	printf("Developed by %s\n", __AUTHOR__);
#endif
	
#ifdef __DATE_MODIFIED__	
	printf("Last built on ");
	printf(__DATE_MODIFIED__);
#endif	
	printf("\n\n");

	for ( int i = 0; i < argc; i++ )
		cout<<argv[i]<<" ";
	cout<<endl;

	bool hasFileName = false;
	char fileName[MAX_FILENAME_LENGTH];
	bool isCNFProg = false;
	long randSeed = 0;
	bool verify = false;
	bool showAllPositiveAtoms = false;
	bool showPredicate = false;
	char * predName = NULL;
	bool debugMode = false;
	bool beQuiet = false;
	bool noProcessBar = false;
	bool toDIMACS = false;
	bool toASPPS = false;

	int i = 1;
	strcpy(fileName, "");
	while ( i < argc )
	{
		if ( !strcmp(argv[i], "-f") )
		{
			if ( i+1 < argc && argv[i+1][0] != '-' )
			{
				hasFileName = true;
				strcpy(fileName, argv[++i]);
			}
		}
		else if ( !strcmp(argv[i], "-D") ||
		     !strcmp(argv[i], "--dimacs") )
		{
			isCNFProg = true;
		}
		else if ( !strcmp(argv[i], "-DT") ||
			  !strcmp(argv[i], "--data") )
		{
			if ( i+1 < argc && argv[i+1][0] != '-' )
			{
				glUseInitValues = true;
				strcpy(glInitFileName, argv[++i]);
			}
			if ( i+1 < argc && argv[i+1][0] != '-' )
			{
				glSetEscapePara = true;
				glEscapeLimit = atoi(argv[++i]);
				if ( glEscapeLimit < 0 ||
				     glEscapeLimit > 100 )
					glEscapeLimit = 25;
			}
			if ( i+1 < argc && argv[i+1][0] != '-' )
			{
				glEscapeProbNumerator = atoi(argv[++i]);
			
				if ( i+1 < argc && argv[i+1][0] != '-' )
				{
					glEscapeProbDenominator =
						atoi(argv[++i]);
				}
			}
		}
		else if ( !strcmp(argv[i], "-C") ||
			  !strcmp(argv[i], "--combination-computation") )
		{
			if ( i+1 < argc && argv[i+1][0] != '-' )
				 glCombinationMethod = atoi(argv[++i]);
		}
		else if ( !strcmp(argv[i], "-c") ||
			  !strcmp(argv[i], "--cutoff") )
		{
			if ( i+1 < argc && argv[i+1][0] != '-' )
				 glMaxFlipNum = atol(argv[++i]);
		}
		else if ( !strcmp(argv[i], "-t") ||
			  !strcmp(argv[i], "--tries") )
		{
			if ( i+1 < argc && argv[i+1][0] != '-' )
				glMaxRetryNum = atol(argv[++i]);
		}
		else if ( !strcmp(argv[i], "-F") ||
			  !strcmp(argv[i], "--stop-on-first") )
		{
			glStopOnFirst = true;
			if ( i+1 < argc && argv[i+1][0] != '-' )
				glNumOfSolBeforeStop = atoi(argv[++i]);
		}
		else if ( !strcmp(argv[i], "-s") ||
			  !strcmp(argv[i], "--seed") )
		{
			if ( i+1 < argc && argv[i+1][0] != '-' )
				randSeed = atol(argv[++i]);
		}
		else if ( !strcmp(argv[i], "-N") ||
			  !strcmp(argv[i], "--noise") )
		{
			if ( i+1 < argc && argv[i+1][0] != '-' )
			{
				glWalkProbabilityNumerator =
					atoi(argv[++i]);
			
				if ( i+1 < argc && argv[i+1][0] != '-' )
				{
					glWalkProbabilityDenominator =
						atoi(argv[++i]);
				}
						
				glNoiseProbability =
					(double)glWalkProbabilityNumerator /
					(double)glWalkProbabilityDenominator;
			}
		}		
		else if ( !strcmp(argv[i], "-T") ||
			  !strcmp(argv[i], "--true") )
		{
			if ( i+1 < argc && argv[i+1][0] != '-' )
			{
				glInitNumerator = atoi(argv[++i]);
			
				if ( i+1 < argc && argv[i+1][0] != '-' )
				{
					glInitDenominator = atoi(argv[++i]);
				}
						
				glInitToTrueProb =
					(double)glInitNumerator /
					(double)glInitDenominator;
			}
		}		
		else if ( !strcmp(argv[i], "-MF") ||
			  !strcmp(argv[i], "--multi-flip") )
		{
			glPureDoubleFlip = true;
			glDisjointCatoms = false;
			glDisjointWatoms = true;
			glUseTimeStamp = true;
			if ( i+1 < argc && argv[i+1][0] != '-' )
			{
				glDBFlipProbabilityNumerator =
					atoi(argv[++i]);
				
				if ( i+1 < argc && argv[i+1][0] != '-' )
				{
					glDBFlipProbabilityDenominator =
						atoi(argv[++i]);
				}
				glDBFlipProbability =
					(double)glDBFlipProbabilityNumerator /
					(double)glDBFlipProbabilityDenominator;
			}
		}		
		else if ( !strcmp(argv[i], "-DF") ||
			  !strcmp(argv[i], "--double-flip") )
		{
			glPureDoubleFlip = true;
			glDisjointCatoms = true;
			glDisjointWatoms = false;
			if ( i+1 < argc && argv[i+1][0] != '-' )
			{
				glDBFlipProbabilityNumerator =
					atoi(argv[++i]);
				
				if ( i+1 < argc && argv[i+1][0] != '-' )
				{
					glDBFlipProbabilityDenominator =
						atoi(argv[++i]);
				}
				glDBFlipProbability =
					(double)glDBFlipProbabilityNumerator /
					(double)glDBFlipProbabilityDenominator;
			}
		}		
		else if ( !strcmp(argv[i], "-DF-W") ||
			  !strcmp(argv[i], "--double-flip-walk") )
		{
			glPureDoubleFlip = false;
			glDisjointCatoms = true;
			glDisjointWatoms = false;
			if ( i+1 < argc && argv[i+1][0] != '-' )
			{
				glDBFlipProbabilityNumerator =
					atoi(argv[++i]);
				
				if ( i+1 < argc && argv[i+1][0] != '-' )
				{
					glDBFlipProbabilityDenominator =
						atoi(argv[++i]);
				}
				glDBFlipProbability =
					(double)glDBFlipProbabilityNumerator /
					(double)glDBFlipProbabilityDenominator;
			}
		}
		else if ( !strcmp(argv[i], "-DF-G") ||
			  !strcmp(argv[i], "--double-flip-general") )
		{
			glGeneralDoubleFlip = true;
			glDisjointWatoms = false;
			if ( i+1 < argc && argv[i+1][0] != '-' )
			{
				strcpy(glGeneralDBFRuleFile, argv[++i]);
			}
			else
			{
				glGeneralDBFRuleFile[0] = '\0';
			}
		}		
		else if ( !strcmp(argv[i], "-PF") ||
			  !strcmp(argv[i], "--permutation-flip") )
		{
			glPermutationCatoms = true;
			glInitialMethod = 1;
		}		
		else if ( !strcmp(argv[i], "-PF") ||
			  !strcmp(argv[i], "--permutation-flip-R") )
		{
			glPermutationCatoms = true;
			glInitialMethod = 2;
		}		
		else if ( !strcmp(argv[i], "-PF") ||
			  !strcmp(argv[i], "--permutation-flip-M") )
		{
			glPermutationCatoms = true;
			glUseMakeCount = true;
			glInitialMethod = 3;
		}		
		else if ( !strcmp(argv[i], "-VC") ||
			  !strcmp(argv[i], "--virtual-counts") )
		{
			glUseVirtualCounts = true;
		}		
		else if ( !strcmp(argv[i], "-SKC") ||
			  !strcmp(argv[i], "--break-count") )
		{
			glHeuristics = HEU_SKC;
		}		
		else if ( !strcmp(argv[i], "-RNP") ||
			  !strcmp(argv[i], "--rnovelty+") )
		{
			glUseTimeStamp = true;
			glUseMakeCount = true;
			glHeuristics = HEU_RNOVELTY_PLUS;
		}		
		else if ( !strcmp(argv[i], "-SW") ||
			  !strcmp(argv[i], "--single-weight") )
		{
			glComboWeight = false;
		}		
		else if ( !strcmp(argv[i], "-TB") ||
			  !strcmp(argv[i], "--tie-breaking") )
		{
			glTieBreaking = true;
		}		
		else if ( !strcmp(argv[i], "-WC") ||
			  !strcmp(argv[i], "--wrong-case") )
		{
			glUseCorrectCase = false;
		}		
		else if ( !strcmp(argv[i], "-V") ||
			  !strcmp(argv[i], "--verify-solution") )
		{
			verify = true;
		}
		else if ( !strcmp(argv[i], "-I") ||
			  !strcmp(argv[i], "--show-init") )
		{
			if ( i+1 < argc && argv[i+1][0] != '-' )
			{
				glPrintAssignmentAfterFlip = atoi(argv[++i]);
			}
		}
		else if ( !strcmp(argv[i], "-SM") ||
			  !strcmp(argv[i], "--smodels-format") )
		{
			glSmodelsOutput = true;
		}
		else if ( !strcmp(argv[i], "-APB") ||
			  !strcmp(argv[i], "--show-all-pbformat") )
		{
			showAllPositiveAtoms = true;
			glShowPositiveAtomsInPBFormat = true;
		}
		else if ( !strcmp(argv[i], "-A") ||
			  !strcmp(argv[i], "--show-all") )
		{
			showAllPositiveAtoms = true;
		}
		else if ( !strcmp(argv[i], "-S") ||
			  !strcmp(argv[i], "--show") )
		{
			if ( i+1 < argc && argv[i+1][0] != '-' )
			{
				showPredicate = true;
				predName = new char [strlen(argv[i])+1];
				if ( predName == NULL )
					return 2;
				
				strcpy(predName, argv[++i]);
			}

		}
		else if ( !strcmp(argv[i], "-B") ||
			  !strcmp(argv[i], "--debug") )
		{
			debugMode = true;
		}
		else if ( !strcmp(argv[i], "-Q") ||
			  !strcmp(argv[i], "--QUIET") )
		{
			beQuiet = true;
		}
		else if ( !strcmp(argv[i], "--nolog") )
		{
			glWriteLogs = false;
		}
		else if ( !strcmp(argv[i], "-q") ||
			  !strcmp(argv[i], "--no-process-bar") )
		{
			noProcessBar = true;
		}
		else if ( !strcmp(argv[i], "-pta") ||
			  !strcmp(argv[i], "--plot-true-atom-num") )
		{
			int freq = -1;
			
			glPlotTrueAtomNum = true;

			if ( i+1 < argc && argv[i+1][0] != '-' )
				freq = atoi(argv[++i]);

			if ( freq > 0 )
				glPlotFrequency = freq;
		}
		else if ( !strcmp(argv[i], "-ptu") ||
			  !strcmp(argv[i], "--plot-unsat-rule-num") )
		{
			int freq = -1;
			
			glPlotUnsatRuleNum = true;

			if ( i+1 < argc && argv[i+1][0] != '-' )
				freq = atoi(argv[++i]);

			if ( freq > 0 )
				glPlotFrequency = freq;
		}
		else if ( !strcmp(argv[i], "-ptt") ||
			  !strcmp(argv[i], "--plot-touched-times") )
		{
			int lb = -1;
			int threshold = -1;

			glPlotTouchedTimes = true;

			if ( i+1 < argc && argv[i+1][0] != '-' )
				lb = atoi(argv[++i]);

			if ( lb >= 0 )
				glPlotLowerBound = lb;

			if ( i+1 < argc && argv[i+1][0] != '-' )
				threshold = atoi(argv[++i]);

			if ( threshold >= 0 )
			{
				glPlotTouchedTimesThreshold = threshold;
			}
		}
		else if ( !strcmp(argv[i], "-d") ||
			  !strcmp(argv[i], "--to-dimacs") )
		{
			toDIMACS = true;
		}
		else if ( !strcmp(argv[i], "-a") ||
			  !strcmp(argv[i], "--to-aspps") )
		{
			toASPPS = true;
		}
		else if ( !strcmp(argv[i], "--clean") )
		{
			system("rm -i ./*.dat");
			system("rm -i ./*.plot");
			return 1;
		}
		else if ( !strcmp(argv[i], "--force-clean") )
		{
			system("rm -f ./*.dat");
			system("rm -f ./*.plot");
			return 1;
		}
		else if ( !strcmp(argv[i], "-h") ||
			  !strcmp(argv[i], "--help") )
		{
			printUsage();
			return 1;
		}
		else
		{
			printf("Unknown command-line argument: %s\n",
			       argv[i]);
			printUsage();
			return 1;
		}
		i++;
	}

	if ( !glDisjointCatoms && !glPermutationCatoms && !glDisjointWatoms )
	{
		glUseVirtualCounts = true;
	}
	
	if ( glPermutationCatoms &&
	     (glDisjointCatoms || glDisjointWatoms) )
	{
		printf("You cannot have permutation flip or double flip at the same time.\n");
		return 1;
	}

	if ( glDisjointCatoms && glGeneralDoubleFlip )
	{
		printf("You cannot have double flip or general double flip at the same time.\n");
	}

	if ( glPermutationCatoms && glGeneralDoubleFlip )
	{
		printf("You cannot have permutation flip or general double flip at the same time.\n");
	}
	
#ifdef __LLN_DEBUG__	
	pLog = fopen("aspps.log","w");
#endif
	
	glUserInterrupted = false;
	signal(SIGINT, handleKBInterrupt);

	CWalkASPPS asppsSolver;
	int rtval;
	
	printf("==================================== BEGIN ");
	printf("====================================\n\n");
	if ( toDIMACS && toASPPS )
	{
		cout<<"Error: You can't choose -a and -d at the same time.\n";
		rtval = -2;
	}
	else if ( toDIMACS )
	{
		asppsSolver.toDIMACS(fileName);
		rtval = -1;
	}
	else if ( toASPPS )
	{
		asppsSolver.toASPPS(fileName);
		rtval = -1;
	}
	else
	{
		asppsSolver.run(beQuiet, noProcessBar, debugMode,
				isCNFProg,
				fileName, verify, randSeed,
				showAllPositiveAtoms,
				showPredicate, predName);
		rtval = 0;
	}

	printf("==================================== END ");
	printf("======================================\n\n");

#ifdef __LLN_DEBUG__
	if ( pLog != NULL )
		fclose(pLog);
#endif
	
	return rtval;
}



